@@ -1,6 +1,6 @@
 ;;;; batch-driver.scm - Driver procedure for the compiler
 ;
-; Copyright (c) 2008-2011, The Chicken Team
+; Copyright (c) 2008-2012, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
@@ -75,6 +75,8 @@
 	(time0 #f)
 	(time-breakdown #f)
 	(forms '())
+	(inline-output-file #f)
+	(type-output-file #f)
 	(cleanup-forms '(((##sys#implicit-exit-handler))))
 	(profile (or (memq 'profile options)
 		     (memq 'accumulate-profile options) 
@@ -82,15 +84,13 @@
 	(profile-name 
 	 (and-let* ((pn (memq 'profile-name options))) (cadr pn)))
 	(hsize (memq 'heap-size options))
-	(hisize (memq 'heap-initial-size options))
-	(hgrowth (memq 'heap-growth options))
-	(hshrink (memq 'heap-shrinkage options))
 	(kwstyle (memq 'keyword-style options))
+        (loop/dispatch (memq 'clustering options))
 	(uses-units '())
 	(uunit (memq 'unit options))
 	(a-only (memq 'analyze-only options))
 	(dynamic (memq 'dynamic options))
-	(unbox (memq 'unboxing options))
+	(do-scrutinize (memq 'scrutinize options))
 	(dumpnodes #f)
 	(start-time #f)
 	(upap #f)
@@ -100,10 +100,10 @@
     (define (cputime) (current-milliseconds))
 
     (define (dribble fstr . args)
-      (when verbose (printf "~?~%~!" fstr args)))
+      (debugging 'p (apply sprintf fstr args)))
 
     (define (print-header mode dbgmode)
-      (dribble "pass: ~a" mode)
+      (debugging 'p "pass" mode)
       (and (memq dbgmode debugging-chicken)
 	   (begin
 	     (printf "[~a]~%" mode)
@@ -149,7 +149,9 @@
 
     (define (end-time pass)
       (when time-breakdown
-	(printf "milliseconds needed for ~a: \t~s~%" pass (- (cputime) time0)) ) )
+	(printf "milliseconds needed for ~a: \t~s~%" 
+	  pass
+	  (inexact->exact (round (- (cputime) time0)) ) )))
 
     (define (analyze pass node . args)
       (let-optionals args ((no 0) (contf #t))
@@ -168,12 +170,16 @@
     (when (memq 'ignore-repository options)
       (set! ##sys#dload-disabled #t)
       (repository-path #f))
+    (set! enable-specialization (memq 'specialize options))
     (set! debugging-chicken 
       (append-map
        (lambda (do)
 	 (map (lambda (c) (string->symbol (string c)))
 	      (string->list do) ) )
        (collect-options 'debug) ) )
+    (when (memq 'h debugging-chicken)
+      (print-debug-options)
+      (exit))
     (set! dumpnodes (memq '|D| debugging-chicken))
     (set! import-libraries
       (map (lambda (il)
@@ -184,7 +190,8 @@
 	       (not a-only))
       (set! all-import-libraries #t))
     (set! enable-module-registration (not (memq 'no-module-registration options)))
-    (when (memq 'scrutinize options)
+    (when (or enable-specialization
+	      (memq 'scrutinize options))
       (set! do-scrutinize #t))
     (when (memq 't debugging-chicken) (##sys#start-timer))
     (when (memq 'b debugging-chicken) (set! time-breakdown #t))
@@ -192,6 +199,8 @@
       (warning "obsolete compiler option: -emit-exports") )
     (when (memq 'lambda-lift options)	; OBSOLETE
       (warning "obsolete compiler option: -lambda-lift") )
+    (when (memq 'unboxing options)	; OBSOLETE
+      (warning "obsolete compiler option: -unboxing") )
     (when (memq 'raw options)
       (set! explicit-use-flag #t)
       (set! cleanup-forms '())
@@ -205,11 +214,16 @@
     (when (memq 'inline-global options)
       (set! enable-inline-files #t)
       (set! inline-locally #t))
-    (when (or verbose do-scrutinize)
+    (when verbose
       (set! ##sys#notices-enabled #t))
+    (when (memq 'strict-types options)
+      (set! strict-variable-types #t)
+      (set! enable-specialization #t)
+      (set! do-scrutinize #t))
     (when (memq 'no-warnings options) 
       (dribble "Warnings are disabled")
-      (set! ##sys#warnings-enabled #f) )
+      (set! ##sys#warnings-enabled #f) 
+      (set! do-scrutinize #f))		; saves some processing time
     (when (memq 'optimize-leaf-routines options) (set! optimize-leaf-routines #t))
     (when (memq 'unsafe options) 
       (set! unsafe #t) )
@@ -225,6 +239,8 @@
       (set! inline-locally #t)		; otherwise this option makes no sense
       (set! local-definitions #t)
       (set! inline-output-file (option-arg ifile)))
+    (and-let* ((tfile (memq 'emit-type-file options)))
+      (set! type-output-file (option-arg tfile)))
     (and-let* ([inlimit (memq 'inline-limit options)])
       (set! inline-max-size 
 	(let ([arg (option-arg inlimit)])
@@ -260,7 +276,10 @@
 	      ipath) )
     (when (and outfile filename (string=? outfile filename))
       (quit "source- and output-filename are the same") )
-    (set! uses-units (map string->symbol (collect-options 'uses)))
+    (set! uses-units
+      (append-map
+       (lambda (u) (map string->symbol (string-split u ", ")))
+       (collect-options 'uses)))
     (when (memq 'keep-shadowed-macros options)
       (set! undefine-shadowed-macros #f) )
     (when (memq 'no-argc-checks options)
@@ -282,11 +301,12 @@
          (mark-variable v '##compiler#always-bound-to-procedure)
          (mark-variable v '##compiler#always-bound) )
        default-extended-bindings) )
+    (when (memq 'p debugging-chicken) (load-verbose #t))
 
     ;; Handle feature options:
     (for-each 
      register-feature!
-     (append-map (cut string-split <> ",") (collect-options 'feature)))
+     (append-map (cut string-split <> ", ") (collect-options 'feature)))
     (for-each 
      unregister-feature!
      (append-map (cut string-split <> ",") (collect-options 'no-feature)))
@@ -295,7 +315,6 @@
     (set! ##sys#features (cons #:compiler-extension ##sys#features))
     (let ([extends (collect-options 'extend)])
       (dribble "Loading compiler extensions...")
-      (when verbose (load-verbose #t))
       (for-each
        (lambda (f) (load (##sys#resolve-include-filename f #f #t))) 
        extends) )
@@ -306,7 +325,7 @@
     ;; Insert postponed initforms:
     (set! initforms (append initforms postponed-initforms))
 
-    (let ((se (map string->symbol (collect-options 'static-extension))))
+    (let ((se (map string->symbol (collect-options 'static-extension)))) ; DEPRECATED
       ;; Append required extensions to initforms:
       (set! initforms
 	(append 
@@ -321,20 +340,11 @@
     (when (memq 'compile-syntax options)
       (set! ##sys#enable-runtime-macros #t) )
     (set! target-heap-size
-      (if hsize
-	  (arg-val (option-arg hsize))
-	  (and-let* ([hsize default-default-target-heap-size]
-		     [(not (zero? hsize))] )
-	    hsize) ) )
-    (set! target-initial-heap-size (and hisize (arg-val (option-arg hisize))))
-    (set! target-heap-growth (and hgrowth (arg-val (option-arg hgrowth))))
-    (set! target-heap-shrinkage (and hshrink (arg-val (option-arg hshrink))))
+      (and hsize
+	   (arg-val (option-arg hsize))))
     (set! target-stack-size
-      (if ssize
-	  (arg-val (option-arg ssize))
-	  (and-let* ([ssize default-default-target-stack-size]
-		     [(not (zero? ssize))] )
-	    ssize) ) )
+      (and ssize
+	   (arg-val (option-arg ssize))))
     (set! emit-trace-info (not (memq 'no-trace options)))
     (set! disable-stack-overflow-checking (memq 'disable-stack-overflow-checks options))
     (set! bootstrap-mode (feature? #:chicken-bootstrap))
@@ -362,7 +372,7 @@
 	       '() ) ) )
 	(dribble "generating ~aprofiled code" (if acc "accumulative " "")) ))
 
-    ;;*** hardcoded "modules.db" is bad (also used in chicken-install.scm)
+    ;;XXX hardcoded "modules.db" is bad (also used in chicken-install.scm)
     (load-identifier-database "modules.db")
 
     (cond ((memq 'version options)
@@ -463,11 +473,14 @@
 			     '() )
 			 '((##core#undefined))) ] )
 
-	     (when (and (pair? compiler-syntax-statistics)
-			(debugging 'S "applied compiler syntax:"))
-	       (for-each 
-		(lambda (cs) (printf "  ~a\t\t~a~%" (car cs) (cdr cs)))
-		compiler-syntax-statistics))
+	     (when (pair? compiler-syntax-statistics)
+	       (with-debugging-output
+		'S
+		(lambda ()
+		  (print "applied compiler syntax:")
+		  (for-each 
+		   (lambda (cs) (printf "  ~a\t\t~a~%" (car cs) (cdr cs)))
+		   compiler-syntax-statistics))))
    	     (when (debugging '|N| "real name table:")
 	       (display-real-name-table) )
 	     (when (debugging 'n "line number database:")
@@ -497,11 +510,10 @@
 			   (list (build-node-graph
 				  (canonicalize-begin-body exps) ) ) ) ) 
 		   (db #f))
-
 	       (print-node "initial node tree" '|T| node0)
 	       (initialize-analysis-database)
 
-	       ;; collect requirements and load inline and types files
+	       ;; collect requirements and load inline files
 	       (let* ((req (concatenate (vector->list file-requirements)))
 		      (mreq (concatenate (map cdr req))))
 		 (when (debugging 'M "; requirements:")
@@ -524,11 +536,16 @@
 			(dribble "Loading inline file ~a ..." ilf)
 			(load-inline-file ilf) )
 		      ifs)))
-		 (when do-scrutinize
-		   ;;*** hardcoded database file name
+		 (when (or do-scrutinize enable-specialization)
+		   ;;XXX hardcoded database file name
 		   (unless (memq 'ignore-repository options)
-		     (load-type-database "types.db"))
-		   (for-each (cut load-type-database <> #f) (collect-options 'types))
+		     (unless (load-type-database "types.db")
+		       (quit "default type-database `types.db' not found")))
+		   (for-each 
+		    (lambda (fn)
+		      (or (load-type-database fn #f)
+			  (quit "type-database `~a' not found" fn)))
+		    (collect-options 'types))
 		   (for-each
 		    (lambda (id)
 		      (load-type-database (make-pathname #f (symbol->string id) "types")))
@@ -537,11 +554,13 @@
 		   (set! first-analysis #f)
 		   (set! db (analyze 'scrutiny node0))
 		   (print-db "analysis" '|0| db 0)
-		   (end-time "pre-analysis")
+		   (end-time "pre-analysis (scrutiny)")
 		   (begin-time)
 		   (debugging 'p "performing scrutiny")
-		   (scrutinize node0 db)
+		   (scrutinize node0 db do-scrutinize enable-specialization)
 		   (end-time "scrutiny")
+		   (when enable-specialization
+		     (print-node "specialization" '|P| node0))
 		   (set! first-analysis #t) ) )
 
 	       (set! ##sys#line-number-database #f)
@@ -556,8 +575,11 @@
 		 (print-node "cps" '|3| node1)
 
 		 ;; Optimization loop:
-		 (let loop ([i 1] [node2 node1] [progress #t])
-
+		 (let loop ((i 1)
+			    (node2 node1)
+			    (progress #t)
+			    (l/d #f)
+			    (l/d-done #f))
 		   (begin-time)
 		   (let ([db (analyze 'opt node2 i progress)])
 		     (when first-analysis
@@ -566,34 +588,51 @@
 		       (when (memq 'd debugging-chicken)
 			 (dump-defined-globals db))
 		       (when (memq 'v debugging-chicken)
-			 (dump-global-refs db)) )
+			 (dump-global-refs db))
+		       ;; do this here, because we must make sure we have a db
+		       (when type-output-file
+			 (dribble "generating type file `~a' ..." type-output-file)
+			 (emit-type-file type-output-file db)))
 		     (set! first-analysis #f)
 		     (end-time "analysis")
 		     (print-db "analysis" '|4| db i)
 
-		     (when (memq 's debugging-chicken) (print-program-statistics db))
+		     (when (memq 's debugging-chicken) 
+		       (print-program-statistics db))
 
 		     (cond (progress
 			    (debugging 'p "optimization pass" i)
 			    (begin-time)
 			    (receive (node2 progress-flag)
-				(perform-high-level-optimizations node2 db)
+				(if l/d
+				    (determine-loop-and-dispatch node2 db)
+				    (perform-high-level-optimizations node2 db))
 			      (end-time "optimization")
 			      (print-node "optimized-iteration" '|5| node2)
-			      (cond [progress-flag (loop (add1 i) node2 #t)]
-				    [(not inline-substitutions-enabled)
-				     (debugging 'p "rewritings enabled...")
+			      (cond (progress-flag
+				     (loop (add1 i) node2 #t #f l/d))
+				    ((and (not l/d-done) loop/dispatch)
+				     (debugging 'p "clustering enabled")
+				     (loop (add1 i) node2 #t #t #t))
+				    ((not inline-substitutions-enabled)
+				     (debugging 'p "rewritings enabled")
 				     (set! inline-substitutions-enabled #t)
-				     (loop (add1 i) node2 #t) ]
-				    [optimize-leaf-routines
+				     (loop (add1 i) node2 #t #f l/d-done) )
+				    (optimize-leaf-routines
 				     (begin-time)
 				     (let ([db (analyze 'leaf node2)])
 				       (end-time "analysis")
 				       (begin-time)
-				       (let ([progress (transform-direct-lambdas! node2 db)])
+				       (let ((progress
+					      (transform-direct-lambdas! node2 db)))
 					 (end-time "leaf routine optimization")
-					 (loop (add1 i) node2 progress) ) ) ]
-				    [else (loop (add1 i) node2 #f)] ) ) )
+					 (loop (add1 i) 
+					       node2
+					       progress
+					       #f
+					       l/d-done) ) ) )
+				    (else
+				     (loop (add1 i) node2 #f #f l/d-done)) ) ) )
 			   
 			   (else
 			    (print-node "optimized" '|7| node2)
@@ -611,23 +650,18 @@
 				       (> (- (cputime) start-time) funny-message-timeout))
 			      (display "(don't worry - still compiling...)\n") )
 			    (print-node "closure-converted" '|9| node2)
-			    (when (and unbox (or unsafe unchecked-specialized-arithmetic))
-			      (debugging 'p "performing unboxing")
-			      (begin-time)
-			      (perform-unboxing! node2)
-			      (end-time "unboxing")
-			      (print-node "unboxing" '|U| node2) )
 			    (when a-only (exit 0))
 			    (begin-time)
 			    (receive 
-			     (node literals lliterals lambdas)
+			     (node literals lliterals lambda-table)
 			     (prepare-for-code-generation node2 db)
 			     (end-time "preparation")
 			     (begin-time)
 			     (let ((out (if outfile (open-output-file outfile) (current-output-port))) )
 			       (dribble "generating `~A' ..." outfile)
-			       (generate-code literals lliterals lambdas out filename dynamic db)
-			       (when outfile (close-output-port out)))
+			       (generate-code literals lliterals lambda-table out filename dynamic db)
+			       (when outfile
+				 (close-output-port out)))
 			     (end-time "code generation")
 			     (when (memq 't debugging-chicken)
 			       (##sys#display-times (##sys#stop-timer)))
