@@ -1,6 +1,6 @@
 ;;; c-backend.scm - C-generating backend for the CHICKEN compiler
 ;
-; Copyright (c) 2008-2011, The Chicken Team
+; Copyright (c) 2008-2012, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
@@ -59,18 +59,17 @@
 
 ;;; Generate target code:
 
-(define (generate-code literals lliterals lambdas out source-file dynamic db)
+(define (generate-code literals lliterals lambda-table out source-file dynamic db)
+  ;; Don't truncate floating-point precision!
+  (flonum-print-precision (+ flonum-maximum-decimal-exponent 1))
   (let ()
 
     ;; Some helper procedures
 
     (define (find-lambda id)
-      (or (find (lambda (ll) (eq? id (lambda-literal-id ll))) lambdas)
+      (or (##sys#hash-table-ref lambda-table id)
 	  (bomb "can't find lambda" id) ) )
 
-    (define (slashify s) (string-translate (->string s) "\\" "/"))
-    (define (uncommentify s) (string-translate* (->string s) '(("*/" . "* /"))))
-  
     ;; Compile a single expression
     (define (expression node temps ll)
 
@@ -189,7 +188,9 @@
 	       (cond [block
 		      (if safe
 			  (gen "lf[" index "]")
-			  (gen "C_retrieve2(lf[" index "]," (c-ify-string (symbol->string (fourth params))) #\)) ) ]
+			  (gen "C_retrieve2(lf[" index "]," 
+			       (c-ify-string (##sys#symbol->qualified-string
+					      (fourth params))) #\)) ) ]
 		     [safe (gen "*((C_word*)lf[" index "]+1)")]
 		     [else (gen "C_fast_retrieve(lf[" index "])")] ) ) )
 
@@ -210,12 +211,12 @@
 		   (var (third params)) )
 	       (cond [block
 		      (gen "lf[" index "] /* "
-			   (uncommentify (symbol->string var)) " */ =")
+			   (uncommentify (##sys#symbol->qualified-string var)) " */ =")
 		      (expr (car subs) i)
 		      (gen #\;) ]
 		     [else
 		      (gen "C_set_block_item(lf[" index "] /* "
-			   (uncommentify (symbol->string var)) " */,0,")
+			   (uncommentify (##sys#symbol->qualified-string var)) " */,0,")
 		      (expr (car subs) i)
 		      (gen #\)) ] ) ) )
 
@@ -289,7 +290,8 @@
 			       (if safe
 				   (gen "C_fast_retrieve_proc(" carg ")")
 				   (gen "C_retrieve2_symbol_proc(" carg "," 
-					(c-ify-string (symbol->string (fourth gparams))) #\)) ) )
+					(c-ify-string (##sys#symbol->qualified-string
+						       (fourth gparams))) #\)) ) )
 			      (safe 
 			       (set! carg 
 				 (string-append "*((C_word*)lf[" (number->string index) "]+1)"))
@@ -360,7 +362,7 @@
 
 	    ((##core#callunit)
 	     ;; The code generated here does not use the extra temporary needed for standard calls, so we have
-	     ;;  one unused varable:
+	     ;;  one unused variable:
 	     (let* ((n (length subs))
 		    (nf (+ n 1)) )
 	       (gen #t "C_" (first params) "_toplevel(" nf ",C_SCHEME_UNDEFINED,")
@@ -468,7 +470,7 @@
 	     (mon (vector-ref tm 4))
 	     (year (vector-ref tm 5)) )
 	(gen "/* Generated from " source-file " by the CHICKEN compiler" #t
-	     "   http://www.call-with-current-continuation.org" #t
+	     "   http://www.call-cc.org" #t
 	     "   " (+ 1900 year) #\- (pad0 (add1 mon)) #\- (pad0 mday) #\space (pad0 hour) #\: (pad0 min) #t
 	     (string-intersperse
 	      (map (cut string-append "   " <> "\n") 
@@ -491,7 +493,12 @@
 	  (generate-foreign-callback-stub-prototypes foreign-callback-stubs) ) ) )
   
     (define (trailer)
-      (gen #t "/* end of file */" #t) )
+      (gen #t #t "/*" #t 
+	   (uncommentify
+	    (get-output-string
+	     collected-debugging-output))
+	   "*/"
+	   #t "/* end of file */" #t))
   
     (define (declarations)
       (let ((n (length literals)))
@@ -525,13 +532,12 @@
     (define (prototypes)
       (let ([large-signatures '()])
 	(gen #t)
-	(for-each
-	 (lambda (ll)
+	(##sys#hash-table-for-each
+	 (lambda (id ll)
 	   (let* ([n (lambda-literal-argument-count ll)]
 		  [customizable (lambda-literal-customizable ll)] 
 		  [empty-closure (and customizable (zero? (lambda-literal-closure-size ll)))]
 		  [varlist (intersperse (make-variable-list (if empty-closure (sub1 n) n) "t") #\,)]
-		  [id (lambda-literal-id ll)]
 		  [rest (lambda-literal-rest-argument ll)]
 		  [rest-mode (lambda-literal-rest-argument-mode ll)]
 		  [direct (lambda-literal-direct ll)] 
@@ -576,7 +582,7 @@
 		    ;;(when customizable (gen " C_c_regparm"))
 		    (unless direct (gen " C_noret"))
 		    (gen #\;) ] ) ) )
-	 lambdas) 
+	 lambda-table) 
 	(for-each
 	 (lambda (s)
 	   (gen #t "typedef void (*C_proc" s ")(C_word")
@@ -618,12 +624,11 @@
 	    (apply gen (intersperse (make-argument-list (+ n 1) "t") #\,))
 	    (gen ");}") ) )
 
-	(for-each
-	 (lambda (ll)
+	(##sys#hash-table-for-each
+	 (lambda (id ll)
 	   (let* ([argc (lambda-literal-argument-count ll)]
 		  [rest (lambda-literal-rest-argument ll)]
 		  [rest-mode (lambda-literal-rest-argument-mode ll)]
-		  [id (lambda-literal-id ll)]
 		  [customizable (lambda-literal-customizable ll)]
 		  [empty-closure (and customizable (zero? (lambda-literal-closure-size ll)))] )
 	     (when empty-closure (set! argc (sub1 argc)))
@@ -641,7 +646,7 @@
 		      (if (and rest (not (eq? rest-mode 'none)))
 			  (set! nsr (lset-adjoin = nsr argc)) 
 			  (set! ns (lset-adjoin = ns argc)) ) ] ) ) ) )
-	 lambdas)
+	 lambda-table)
 	(for-each
 	 (lambda (n)
 	   (gen #t #t "C_noret_decl(tr" n ")"
@@ -738,10 +743,9 @@
 	(else (bomb "invalid unboxed type" t))))
 
     (define (procedures)
-      (for-each
-       (lambda (ll)
+      (##sys#hash-table-for-each
+       (lambda (id ll)
 	 (let* ((n (lambda-literal-argument-count ll))
-		(id (lambda-literal-id ll))
 		(rname (real-name id db))
 		(demand (lambda-literal-allocated ll))
 		(rest (lambda-literal-rest-argument ll))
@@ -810,15 +814,9 @@
 		    (when disable-stack-overflow-checking
 		      (gen #t "C_disable_overflow_check=1;") )
 		    (unless unit-name
-		      (cond [target-initial-heap-size
-			     (gen #t "C_set_or_change_heap_size(" target-initial-heap-size ",1);") ]
-			    [target-heap-size
-			     (gen #t "C_set_or_change_heap_size(" target-heap-size ",1);"
-				  #t "C_heap_size_is_fixed=1;") ] )
-		      (when target-heap-growth
-			(gen #t "C_heap_growth=" target-heap-growth #\;) )
-		      (when target-heap-shrinkage
-			(gen #t "C_heap_shrinkage=" target-heap-shrinkage #\;) )
+		      (when target-heap-size
+			(gen #t "C_set_or_change_heap_size(" target-heap-size ",1);"
+			     #t "C_heap_size_is_fixed=1;"))
 		      (when target-stack-size
 			(gen #t "C_resize_stack(" target-stack-size ");") ) )
 		    (gen #t "C_check_nursery_minimum(" demand ");"
@@ -857,7 +855,7 @@
 			[else
 			 (unless direct (gen #t "C_word *a;"))
 			 (when (and direct (not unsafe) (not disable-stack-overflow-checking))
-			   (gen #t "C_stack_check;") )
+			   (gen #t "C_stack_overflow_check;") )
 			 (when looping (gen #t "loop:")) ] )
 		  (when (and external (not unsafe) (not no-argc-checks) (not customizable))
 		    ;; (not customizable) implies empty-closure
@@ -911,7 +909,7 @@
 		n)
 	    ll)
 	   (gen #\}) ) )
-       lambdas) )
+       lambda-table) )
 
     (debugging 'p "code generation phase...")
     (set! output out)
@@ -923,25 +921,25 @@
     (generate-foreign-callback-stubs foreign-callback-stubs db)
     (trampolines)
     (procedures)
-    (emit-procedure-table-info lambdas source-file)
+    (emit-procedure-table-info lambda-table source-file)
     (trailer) ) )
 
 
 ;;; Emit procedure table:
 
-(define (emit-procedure-table-info lambdas sf)
+(define (emit-procedure-table-info lambda-table sf)
   (gen #t #t "#ifdef C_ENABLE_PTABLES"
-       #t "static C_PTABLE_ENTRY ptable[" (add1 (length lambdas)) "] = {")
-  (do ((ll lambdas (cdr ll)))
-      ((null? ll)
-       (gen #t "{NULL,NULL}};") )
-    (let ((id (lambda-literal-id (car ll))))
-      (gen #t "{\"" id #\: (string->c-identifier sf) "\",(void*)")
-      (if (eq? 'toplevel id)
-	  (if unit-name
-	      (gen "C_" unit-name "_toplevel},")
-	      (gen "C_toplevel},") )
-	  (gen id "},") ) ) )
+       #t "static C_PTABLE_ENTRY ptable[" (add1 (##sys#hash-table-size lambda-table)) "] = {")
+  (##sys#hash-table-for-each
+   (lambda (id ll)
+     (gen #t "{\"" id #\: (string->c-identifier sf) "\",(void*)")
+     (if (eq? 'toplevel id)
+         (if unit-name
+             (gen "C_" unit-name "_toplevel},")
+             (gen "C_toplevel},") )
+         (gen id "},") ) )
+   lambda-table)
+  (gen #t "{NULL,NULL}};")
   (gen #t "#endif")
   (gen #t #t "static C_PTABLE_ENTRY *create_ptable(void)")
   (gen "{" #t "#ifdef C_ENABLE_PTABLES"
@@ -1116,7 +1114,8 @@
 	 (when rname
 	   (gen #t "/* from " (cleanup rname) " */") )
 	 (generate-foreign-callback-header "" stub)
-	 (gen #\{ #t "C_word x,s=" sizestr ",*a=C_alloc(s);")
+	 (gen #\{ #t "C_word x,s=" sizestr ",*a="
+	      (if (string=? "0" sizestr) "C_stack_pointer;" "C_alloc(s);"))
 	 (gen #t "C_callback_adjust_stack(a,s);") ; make sure content is below stack_bottom as well
 	 (for-each
 	  (lambda (v t)
@@ -1249,7 +1248,7 @@
     (case type
       ((scheme-object) "(")
       ((char unsigned-char) "C_character_code((C_word)")
-      ((byte int unsigned-int unsigned-int32 unsigned-byte) "C_unfix(")
+      ((byte int int32 unsigned-int unsigned-int32 unsigned-byte) "C_unfix(")
       ((short) "C_unfix(")
       ((unsigned-short) "(unsigned short)C_unfix(")
       ((unsigned-long) "C_num_to_unsigned_long(")
@@ -1322,7 +1321,7 @@
       ((unsigned-short) "C_fix(0xffff&(C_word)")
       ((byte) "C_fix((char)")
       ((unsigned-byte) "C_fix(0xff&(C_word)")
-      ((float double) (sprintf "C_flonum(&~a," dest))	;*** suboptimal for int64
+      ((float double) (sprintf "C_flonum(&~a," dest))	;XXX suboptimal for int64
       ((number) (sprintf "C_number(&~a," dest))
       ((nonnull-c-string c-string nonnull-c-pointer c-string* nonnull-c-string* 
 			 unsigned-c-string unsigned-c-string* nonnull-unsigned-c-string
