@@ -1,6 +1,6 @@
 ;;;; csc.scm - Driver program for the CHICKEN compiler - felix -*- Scheme -*-
 ;
-; Copyright (c) 2008-2011, The Chicken Team
+; Copyright (c) 2008-2012, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
@@ -119,8 +119,13 @@
 (define windows-shell WINDOWS_SHELL)
 (define generate-manifest #f)
 
+(define libchicken
+  (if cygwin
+      (string-append "cyg" INSTALL_LIB_NAME "-0")
+      (string-append "lib" INSTALL_LIB_NAME)))
+
 (define default-library
-  (string-append "libchicken." library-extension))
+  (string-append libchicken "." library-extension))
 
 (define default-compilation-optimization-options (string-split (if host-mode INSTALL_CFLAGS TARGET_CFLAGS)))
 (define best-compilation-optimization-options default-compilation-optimization-options)
@@ -137,17 +142,20 @@
     -analyze-only -keep-shadowed-macros -inline-global -ignore-repository
     -no-symbol-escape -no-parentheses-synonyms -r5rs-syntax
     -no-argc-checks -no-bound-checks -no-procedure-checks -no-compiler-syntax
-    -emit-all-import-libraries -setup-mode -unboxing -no-elevation -no-module-registration
+    -emit-all-import-libraries -setup-mode -no-elevation -no-module-registration
     -no-procedure-checks-for-usual-bindings -module
-    -lambda-lift			; OBSOLETE
+    -specialize -strict-types -clustering
+    -lambda-lift -unboxing		; OBSOLETE
     -no-procedure-checks-for-toplevel-bindings))
 
 (define-constant complex-options
-  '(-debug -output-file -heap-size -nursery -stack-size -compiler -unit -uses -keyword-style
+  '(-debug -heap-size -nursery -stack-size -compiler -unit -uses -keyword-style
     -optimize-level -include-path -database-size -extend -prelude -postlude -prologue -epilogue 
     -inline-limit -profile-name
-    -emit-inline-file -types
-    -feature -debug-level -heap-growth -heap-shrinkage -heap-initial-size -consult-inline-file
+    -emit-inline-file -types -emit-type-file
+    -feature -debug-level 
+    -heap-growth -heap-shrinkage -heap-initial-size ; DEPRECATED
+    -consult-inline-file
     -emit-import-library
     -no-feature))
 
@@ -157,7 +165,7 @@
     (-S "-scrutinize")
     (-M "-module")
     (|-P| "-check-syntax")
-    (|-V| "-version")
+    (|-V| "-version")			; DEPRECATED
     (-f "-fixnum-arithmetic")
     (|-D| "-feature")
     (-i "-case-insensitive")
@@ -224,7 +232,8 @@
 (define translate-options '())
 
 (define include-dir
-  (let ((id (prefix "" "include" 
+  (let ((id (prefix ""
+                    (make-pathname "include" "chicken")
 		    (if host-mode INSTALL_INCLUDE_HOME TARGET_INCLUDE_HOME))))
     (and (not (member id '("/usr/include" "")))
 	 id) ) )
@@ -232,7 +241,12 @@
 (define compile-options '())
 
 (define builtin-compile-options
-  (if include-dir (list (conc "-I\"" include-dir "\"")) '()))
+  (append
+   (if include-dir (list (conc "-I\"" include-dir "\"")) '())
+   (cond ((get-environment-variable "CHICKEN_C_INCLUDE_PATH") => 
+	  (lambda (path) 
+	    (map (cut string-append "-I\"" <> "\"") (string-split path ":;"))))
+	 (else '()))))
 
 (define compile-only-flag "-c")
 (define translation-optimization-options default-translation-optimization-options)
@@ -248,19 +262,24 @@
 (define link-options '())
 
 (define (builtin-link-options)
-  (cond (elf
-	 (list
-	  (conc "-L\"" library-dir "\"")
-	  (conc " -Wl,-R\""
-		(if (and deployed (not netbsd))
-		    "\\$ORIGIN"
-		    (prefix "" "lib"
-			    (if host-mode
-				INSTALL_LIB_HOME
-				TARGET_RUN_LIB_HOME)))
-		"\"")) )
-	(else
-	 (list (conc "-L\"" library-dir "\"")))))
+  (append
+   (cond (elf
+	  (list
+	   (conc "-L\"" library-dir "\"")
+	   (conc " -Wl,-R\""
+		 (if (and deployed (not netbsd))
+		     "\\$ORIGIN"
+		     (prefix "" "lib"
+			     (if host-mode
+				 INSTALL_LIB_HOME
+				 TARGET_RUN_LIB_HOME)))
+		 "\"")) )
+	 (else
+	  (list (conc "-L\"" library-dir "\""))))
+   (cond ((get-environment-variable "CHICKEN_C_LIBRARY_PATH") => 
+	  (lambda (path) 
+	    (map (cut string-append "-L\"" <> "\"") (string-split path ":;"))))
+	 (else '()))))
 	
 (define target-filename #f)
 (define verbose #f)
@@ -291,12 +310,12 @@ Usage: #{csc} FILENAME | OPTION ...
   General options:
 
     -h  -help                      display this text and exit
-    -v                             show intermediate compilation stages
-    -vv  -verbose                  display information about translation
+    -v  -verbose                   show compiler notes and tool-invocations
+    -vv                            display information about translation
                                     progress
     -vvv                           display information about all compilation
                                     stages
-    -V  -version                   display Scheme compiler version and exit
+    -version                       display Scheme compiler version and exit
     -release                       display release number and exit
 
   File and pathname options:
@@ -369,10 +388,11 @@ Usage: #{csc} FILENAME | OPTION ...
     -inline                        enable inlining
     -inline-limit LIMIT            set inlining threshold
     -inline-global                 enable cross-module inlining
-    -unboxing                      use unboxed temporaries if possible
+    -specialize                    perform type-based specialization of primitive calls
     -n -emit-inline-file FILENAME  generate file with globally inlinable
                                     procedures (implies -inline -local)
     -consult-inline-file FILENAME  explicitly load inline file
+    -emit-type-file FILENAME       write type-declaration information into file
     -no-argc-checks                disable argument count checks
     -no-bound-checks               disable bound variable checks
     -no-procedure-checks           disable procedure call checks
@@ -382,15 +402,15 @@ Usage: #{csc} FILENAME | OPTION ...
     -no-procedure-checks-for-toplevel-bindings
                                    disable procedure call checks for toplevel
                                     bindings
+    -strict-types                  assume variable do not change their type
+    -clustering                    combine groups of local procedures into dispatch
+                                     loop
 
   Configuration options:
 
     -unit NAME                     compile file as a library unit
     -uses NAME                     declare library unit as used.
     -heap-size NUMBER              specifies heap-size of compiled executable
-    -heap-initial-size NUMBER      specifies heap-size at startup time
-    -heap-growth PERCENTAGE        specifies growth-rate of expanding heap
-    -heap-shrinkage PERCENTAGE     specifies shrink-rate of contracting heap
     -nursery NUMBER  -stack-size NUMBER
                                    specifies nursery size of compiled
                                    executable
@@ -594,14 +614,14 @@ EOF
 	       [(-libs)
 		(set! inquiry-only #t)
 		(set! show-libs #t) ]
-	       [(-v -verbose)
+	       ((-v -verbose)
 		(when (number? verbose)
 		  (set! compile-options (cons* "-v" "-Q" compile-options))
 		  (set! link-options (cons "-v" link-options)) )
-		(cond (verbose
-		       (t-options "-verbose") 
-		       (set! verbose 2)) 
-		      (else (set! verbose #t))) ]
+		(t-options "-verbose")
+		(if verbose
+		    (set! verbose 2)
+		    (set! verbose #t)) )
 	       [(-w -no-warnings)
 		(set! compile-options (cons "-w" compile-options))
 		(t-options "-no-warnings") ]
@@ -648,7 +668,7 @@ EOF
 		(when osx 
 		  (set! link-options (cons* "-framework" (car rest) link-options)) )
 		(set! rest (cdr rest)) ]
-	       [(-o)
+	       [(-o -output-file)
 		(check s rest)
 		(let ([fn (car rest)])
 		  (set! rest (cdr rest))
@@ -659,13 +679,10 @@ EOF
 	       [(|-O3|) (set! rest (cons* "-optimize-level" "3" rest))]
 	       [(|-O4|) (set! rest (cons* "-optimize-level" "4" rest))]
 	       [(|-O5|)
-		(set! rest (cons* "-optimize-level" "5" rest))
-		(when (memq (build-platform) '(mingw32 cygwin gnu clang))
-		  (set! compile-options 
-		    (cons* "-O3" "-fomit-frame-pointer" compile-options)) ) ]
-	       [(-d0) (set! rest (cons* "-debug-level" "0" rest))]
-	       [(-d1) (set! rest (cons* "-debug-level" "1" rest))]
-	       [(-d2) (set! rest (cons* "-debug-level" "2" rest))]
+		(set! rest (cons* "-optimize-level" "5" rest))]
+	       [(|-d0|) (set! rest (cons* "-debug-level" "0" rest))]
+	       [(|-d1|) (set! rest (cons* "-debug-level" "1" rest))]
+	       [(|-d2|) (set! rest (cons* "-debug-level" "2" rest))]
 	       [(-dry-run) 
 		(set! verbose #t)
 		(set! dry-run #t)]
@@ -706,9 +723,9 @@ EOF
 		(check s rest)
 		(set! rpath (car rest))
 		(when (and (memq (build-platform) '(gnu clang))
-			   (not mingw))
-		  (set! link-options (append link-options (list (string-append "-Wl,-R" rpath))))
-		  (set! rest (cdr rest)) ) ]
+			   (not mingw) (not osx))
+		  (set! link-options (append link-options (list (string-append "-Wl,-R" rpath)))) )
+	  	(set! rest (cdr rest)) ]
 	       [(-host) #f]
 	       [(-) 
 		(set! target-filename (make-pathname #f "a" executable-extension))
@@ -900,9 +917,9 @@ EOF
     (when (and osx (or (not cross-chicken) host-mode))
       (command
        (string-append
-	"install_name_tool -change libchicken.dylib "
+	"install_name_tool -change " libchicken ".dylib "
 	(quotewrap 
-	 (let ((lib "libchicken.dylib"))
+	 (let ((lib (string-append libchicken ".dylib")))
 	   (if deployed
 	       (make-pathname "@executable_path" lib)
 	       (make-pathname
@@ -943,9 +960,9 @@ EOF
 (define (copy-libraries targetdir)
   (let ((lib (make-pathname
 	      (target-lib-path) 
-	      "libchicken"
+	      libchicken
 	      (cond (osx "dylib")
-		    (win "dll")
+		    ((or win cygwin) "dll")
 		    (else (string-append
                            "so."
                            (number->string BINARY_VERSION)))))))
