@@ -1,6 +1,6 @@
 ;;;; csi.scm - Interpreter stub for CHICKEN
 ;
-; Copyright (c) 2008-2011, The Chicken Team
+; Copyright (c) 2008-2012, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
@@ -31,6 +31,8 @@
   (disable-interrupts)
   (compile-syntax)
   (foreign-declare #<<EOF
+#include <signal.h>
+
 #if defined(HAVE_DIRECT_H)
 # include <direct.h>
 #else
@@ -43,10 +45,10 @@ EOF
 
 (private csi 
   print-usage print-banner
-  run hexdump del 
+  run hexdump
   parse-option-string chop-separator lookup-script-file
   report describe dump hexdump bytevector-data get-config
-  deldups tty-input?
+  tty-input?
   history-list history-count history-add history-ref history-clear history-show)
 
 (declare
@@ -55,7 +57,7 @@ EOF
   (hide parse-option-string bytevector-data member* canonicalize-args 
 	describer-table dirseparator? circular-list? improper-pairs?
 	show-frameinfo selected-frame select-frame copy-from-frame
-	findall command-table default-editor) )
+	findall command-table default-editor csi-eval) )
 
 
 ;;; Parameters:
@@ -89,7 +91,7 @@ usage: csi [FILENAME | OPTION ...]
   one of the following:
 
     -h  -help  --help             display this text and exit
-    -v  -version                  display version and exit
+        -version                  display version and exit
         -release                  print release number and exit
     -i  -case-insensitive         enable case-insensitive reading
     -e  -eval EXPRESSION          evaluate given expression
@@ -130,7 +132,7 @@ EOF
 
 (define (print-banner)
   (newline)
-  ;;UNUSED
+  ;;UNUSED BECAUSE IT IS STUPID
   #;(when (and (tty-input?) (##sys#fudge 11))
     (let* ((t (string-copy +product+))
 	   (len (string-length t))
@@ -205,7 +207,7 @@ EOF
 	(and (> (##sys#size name) 0)
 	     (cond [(dirseparator? (string-ref name 0)) (addext name)]
 		   [(string-index dirseparator? name)
-		    (and-let* ([p (_getcwd buf 256)])
+		    (let ((p (_getcwd buf 256)))
 		      (addext (string-append (chop-separator p) "/" name)) ) ]
 		   [(addext name)]
 		   [else
@@ -286,7 +288,7 @@ EOF
 	 (set! command-table (cons (list name proc help) command-table))))
   (##sys#void))
 
-(set! ##sys#repl-eval-hook
+(set! csi-eval
   (let ((eval eval)
 	(load-noisily load-noisily)
 	(read read)
@@ -336,7 +338,7 @@ EOF
 				[xn (eval n)] )
 			   (dump xe xn) ) )
 			((r) (report))
-			((q) (exit))
+			((q) (##sys#quit-hook))
 			((l)
 			 (let ((fns (string-split (read-line))))
 			   (for-each load fns)
@@ -427,16 +429,16 @@ EOF
  (let ((printf printf))
    (lambda ()
      (let ((name (read)))
+       (when (string? name)
+	 (set! name (##sys#string->symbol name)))
        (cond ((not name)
-	      (##sys#current-module #f)
+	      (##sys#switch-module #f)
 	      (printf "; resetting current module to toplevel~%"))
-	     ((string? name)
-	      (set! name (##sys#string->symbol name)))
 	     ((not (symbol? name))
 	      (printf "invalid module name `~a'~%" name))
-	     ((##sys#find-module name #f) =>
+	     ((##sys#find-module (##sys#resolve-module-name name #f) #f) =>
 	      (lambda (m)
-		(##sys#current-module m)
+		(##sys#switch-module m)
 		(printf "; switching current module to `~a'~%" name)))
 	     (else
 	      (printf "undefined module `~a'~%" name))))))
@@ -583,7 +585,10 @@ EOF
 		   (let ((v (pref x (fx+ start i))))
 		     (let loop2 ((n 1) (j (fx+ i (fx+ start 1))))
 		       (cond ((fx>= j len)
-			      (fprintf out " ~S: ~S" i v)
+			      (##sys#with-print-length-limit
+			       1000
+			       (lambda ()
+				 (fprintf out " ~S: ~S" i v)))
 			      (if (fx> n 1)
 				  (fprintf out "\t(followed by ~A identical instance~a)~% ...~%" 
 					   (fx- n 1)
@@ -727,7 +732,7 @@ EOF
       (##sys#void) ) ) )
 
 (define (set-describer! tag proc)
-  (##sys#check-symbol tag 'symbol 'set-describer!)
+  (##sys#check-symbol tag 'set-describer!)
   (##sys#hash-table-set! describer-table tag proc) )
 
 
@@ -903,26 +908,22 @@ EOF
 	    (##sys#void))))))
 
 
+;;; Handle some signals:
+
+(define-foreign-variable _sigint int "SIGINT")
+
+(define-syntax defhandler 
+  (syntax-rules ()
+    ((_ sig handler)
+     (begin
+       (##core#inline "C_establish_signal_handler" sig sig)
+       (##sys#setslot ##sys#signal-vector sig handler)))))
+
+(defhandler _sigint (lambda (n) (##sys#user-interrupt-hook)))
+
+
 ;;; Start interpreting:
 
-(define (del x lst tst)
-  (let loop ([lst lst])
-    (if (null? lst)
-	'()
-	(let ([y (car lst)])
-	  (if (tst x y)
-	      (cdr lst)
-	      (cons y (loop (cdr lst))) ) ) ) ) )
-
-(define (deldups lis . maybe-=)
-  (let ((elt= (optional maybe-= equal?)))
-    (let recur ((lis lis))
-      (if (null? lis) lis
-	  (let* ((x (car lis))
-		 (tail (cdr lis))
-		 (new-tail (recur (del x tail elt=))))
-	    (if (eq? tail new-tail) lis (cons x new-tail)))))))
-
 (define (member* keys set)
   (let loop ((set set))
     (and (pair? set)
@@ -1029,7 +1030,9 @@ EOF
       (when (member* '("-h" "-help" "--help") args)
 	(print-usage)
 	(exit 0) )
-      (when (member* '("-v" "-version") args)
+      (when (member* '("-v"		; DEPRECATED
+		       "-version") 
+		     args)
 	(print-banner)
 	(exit 0) )
       (when (member "-setup-mode" args)
@@ -1051,7 +1054,7 @@ EOF
       (for-each register-feature! (collect-options "-D"))
       (for-each unregister-feature! (collect-options "-no-feature"))
       (set! ##sys#include-pathnames 
-	(deldups
+	(##sys#nodups
 	 (append (map chop-separator (collect-options "-include-path"))
 		 (map chop-separator (collect-options "-I"))
 		 ##sys#include-pathnames
@@ -1084,7 +1087,10 @@ EOF
       (do ([args args (cdr args)])
 	  ((null? args)
 	   (unless batch 
-	     (repl)
+	     (call/cc
+	      (lambda (k)
+		(set! ##sys#quit-hook (lambda _ (k #f)))
+		(repl csi-eval)))
 	     (##sys#write-char-0 #\newline ##sys#standard-output) ) )
 	(let* ((arg (car args)))
 	  (cond ((member arg simple-options))
