@@ -1,6 +1,6 @@
 ;;; data-structures.scm - Optional data structures extensions
 ;
-; Copyright (c) 2008-2011, The Chicken Team
+; Copyright (c) 2008-2012, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without
@@ -26,11 +26,7 @@
 
 
 (declare
- (unit data-structures)
- (foreign-declare #<<EOF
-#define C_mem_compare(to, from, n)   C_fix(C_memcmp(C_c_string(to), C_c_string(from), C_unfix(n)))
-EOF
-) )
+ (unit data-structures))
 
 (include "common-declarations.scm")
 
@@ -114,11 +110,9 @@ EOF
 
 (define (any? x) #t)
 
-(define (none? x) #f)
-
-(define (always? . _) #t)
-
-(define (never? . _) #f)
+(define (none? x) #f)			; DEPRECATED
+(define (always? . _) #t)		; DEPRECATED
+(define (never? . _) #f)		; DEPRECATED
 
 
 ;;; List operators:
@@ -204,7 +198,7 @@ EOF
 	       (cons (##sys#slot lst 0) (loop (##sys#slot blst 1) (##sys#slot lst 1)))]
 	      [else (loop (##sys#slot blst 1) (##sys#slot lst 1))] ) ) ) ) )
 
-(define shuffle
+(define shuffle				; DEPRECATED
   (lambda (l random)
     (let ((len (length l)))
       (map cdr
@@ -233,6 +227,19 @@ EOF
 	  lst)
 	(cons (cons x y) lst) ) ) )
 
+(define (alist-update k v lst #!optional (cmp eqv?))
+  (let loop ((lst lst))
+    (if (null? lst)
+        (list (cons k v))
+        (let ((a (##sys#slot lst 0)))
+          (cond ((not (pair? a))
+                 (error 'alist-update "bad argument type" a))
+                ((cmp (##sys#slot a 0) k)
+                 (cons (cons k v) (##sys#slot lst 1)))
+                (else
+                 (cons (cons (##sys#slot a 0) (##sys#slot a 1))
+                       (loop (##sys#slot lst 1)))))))))
+
 (define (alist-ref x lst #!optional (cmp eqv?) (default #f))
   (let* ([aq (cond [(eq? eq? cmp) assq]
 		   [(eq? eqv? cmp) assv]
@@ -266,20 +273,18 @@ EOF
 ; (reverse-string-append l) = (apply string-append (reverse l))
 
 (define (reverse-string-append l)
-
   (define (rev-string-append l i)
     (if (pair? l)
       (let* ((str (car l))
 	     (len (string-length str))
-	     (result (rev-string-append (cdr l) (+ i len))))
-	(let loop ((j 0) (k (- (- (string-length result) i) len)))
-	  (if (< j len)
+	     (result (rev-string-append (cdr l) (fx+ i len))))
+	(let loop ((j 0) (k (fx- (fx- (string-length result) i) len)))
+	  (if (fx< j len)
 	    (begin
 	      (string-set! result k (string-ref str j))
-	      (loop (+ j 1) (+ k 1)))
+	      (loop (fx+ j 1) (fx+ k 1)))
 	    result)))
       (make-string i)))
-
   (rev-string-append l 0))
 
 ;;; Anything->string conversion:
@@ -343,7 +348,7 @@ EOF
   (let ((len1 (##sys#size s1))
 	(len2 (##sys#size s2)) )
     (let* ((len-diff (fx- len1 len2)) 
-	   (cmp (##core#inline "C_mem_compare" s1 s2 (if (fx< len-diff 0) len1 len2))))
+	   (cmp (##core#inline "C_string_compare" s1 s2 (if (fx< len-diff 0) len1 len2))))
       (if (fx= cmp 0) 
 	  len-diff 
 	  cmp))))
@@ -512,7 +517,7 @@ EOF
       (if (fx>= i len)
 	  (##sys#fragments->string
 	   total
-	   (reverse 
+	   (##sys#fast-reverse 
 	    (if (fx> i from) 
 		(cons (##sys#substring str from i) fs)
 		fs) ) )
@@ -795,14 +800,18 @@ EOF
 ; is stored in the queue type so that datums can be added in constant
 ; time.
 
-(define (make-queue) (##sys#make-structure 'queue '() '()))
+(define (make-queue) (##sys#make-structure 'queue '() '() 0))
 (define (queue? x) (##sys#structure? x 'queue))
 
-(define (queue-empty? q)
+(define (queue-length q)		; thread-safe
+  (##sys#check-structure q 'queue 'queue-length)
+  (##sys#slot q 3))
+
+(define (queue-empty? q)		; thread-safe
   (##sys#check-structure q 'queue 'queue-empty?)
   (eq? '() (##sys#slot q 1)) )
 
-(define queue-first
+(define queue-first			; thread-safe
   (lambda (q)
     (##sys#check-structure q 'queue 'queue-first)
     (let ((first-pair (##sys#slot q 1)))
@@ -810,7 +819,7 @@ EOF
 	(##sys#error 'queue-first "queue is empty" q))
       (##sys#slot first-pair 0) ) ) )
 
-(define queue-last
+(define queue-last			; thread-safe
   (lambda (q)
     (##sys#check-structure q 'queue 'queue-last)
     (let ((last-pair (##sys#slot q 2)))
@@ -818,15 +827,16 @@ EOF
 	(##sys#error 'queue-last "queue is empty" q))
       (##sys#slot last-pair 0) ) ) )
 
-(define (queue-add! q datum)
+(define (queue-add! q datum)		; thread-safe
   (##sys#check-structure q 'queue 'queue-add!)
   (let ((new-pair (cons datum '())))
     (cond ((eq? '() (##sys#slot q 1)) (##sys#setslot q 1 new-pair))
 	  (else (##sys#setslot (##sys#slot q 2) 1 new-pair)) )
     (##sys#setslot q 2 new-pair) 
+    (##sys#setislot q 3 (fx+ (##sys#slot q 3) 1))
     (##core#undefined) ) )
 
-(define queue-remove!
+(define queue-remove!			; thread-safe
   (lambda (q)
     (##sys#check-structure q 'queue 'queue-remove!)
     (let ((first-pair (##sys#slot q 1)))
@@ -836,13 +846,17 @@ EOF
 	(##sys#setslot q 1 first-cdr)
 	(if (eq? '() first-cdr)
 	    (##sys#setslot q 2 '()) )
+	(##sys#setislot q 3 (fx- (##sys#slot q 3) 1))
 	(##sys#slot first-pair 0) ) ) ) )
 
 (define (queue->list q)
   (##sys#check-structure q 'queue 'queue->list)
-  (##sys#slot q 1) )
+  (let loop ((lst (##sys#slot q 1)) (lst2 '()))
+    (if (null? lst)
+	(##sys#fast-reverse lst2)
+	(loop (##sys#slot lst 1) (cons (##sys#slot lst 0) lst2)))))
 
-(define (list->queue lst0)
+(define (list->queue lst0)		
   (##sys#check-list lst0 'list->queue)
   (##sys#make-structure 
    'queue lst0
@@ -858,12 +872,13 @@ EOF
 ; (queue-push-back! queue item)
 ; Pushes an item into the first position of a queue.
 
-(define (queue-push-back! q item)
+(define (queue-push-back! q item)	; thread-safe
   (##sys#check-structure q 'queue 'queue-push-back!)
   (let ((newlist (cons item (##sys#slot q 1))))
     (##sys#setslot q 1 newlist)
     (if (eq? '() (##sys#slot q 2))
-	(##sys#setslot q 2 newlist))))
+	(##sys#setslot q 2 newlist))
+    (##sys#setislot q 3 (fx+ (##sys#slot q 3) 1))))
 
 ; (queue-push-back-list! queue item-list)
 ; Pushes the items in item-list back onto the queue,
@@ -881,4 +896,5 @@ EOF
 		       '()
 		       (last-pair newlist))))
     (##sys#setslot q 1 newlist)
-    (##sys#setslot q 2 newtail)))
+    (##sys#setslot q 2 newtail)
+    (##sys#setislot q 3 (fx+ (##sys#slot q 3) (##core#inline "C_i_length" itemlist)))))
