@@ -1,6 +1,6 @@
 ;;;; eval.scm - Interpreter for CHICKEN
 ;
-; Copyright (c) 2008-2011, The Chicken Team
+; Copyright (c) 2008-2012, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
@@ -27,10 +27,9 @@
 
 (declare
   (unit eval)
-  (uses expand)
-  (hide ##sys#r4rs-environment ##sys#r5rs-environment 
-	##sys#interaction-environment pds pdss pxss d) 
-  (not inline ##sys#repl-eval-hook ##sys#repl-read-hook ##sys#repl-print-hook 
+  (uses expand modules)
+  (hide pds pdss pxss d) 
+  (not inline ##sys#repl-read-hook ##sys#repl-print-hook 
        ##sys#read-prompt-hook ##sys#alias-global-hook ##sys#user-read-hook
        ##sys#syntax-error-hook))
 
@@ -46,22 +45,31 @@
 #ifndef C_BINARY_VERSION
 # define C_BINARY_VERSION      0
 #endif
+
+#define C_rnd_fix()		(C_fix(rand()))
 <#
 
 (include "common-declarations.scm")
 
+(define-syntax d (syntax-rules () ((_ . _) (void))))
+
+
 (define-foreign-variable install-egg-home c-string "C_INSTALL_EGG_HOME")
 (define-foreign-variable installation-home c-string "C_INSTALL_SHARE_HOME")
 (define-foreign-variable binary-version int "C_BINARY_VERSION")
+(define-foreign-variable install-lib-name c-string "C_INSTALL_LIB_NAME")
 
 (define ##sys#core-library-modules
   '(extras lolevel utils files tcp irregex posix srfi-1 srfi-4 srfi-13 
-	   srfi-14 srfi-18 srfi-69 data-structures ports chicken-syntax
-	   chicken-ffi-syntax))
+	   srfi-14 srfi-18 srfi-69 data-structures ports))
+
+(define ##sys#core-syntax-modules
+  '(chicken-syntax chicken-ffi-syntax))
 
 (define ##sys#explicit-library-modules '())
 
-(define-constant default-dynamic-load-libraries '("libchicken"))
+(define default-dynamic-load-libraries
+  `(,(string-append "lib" install-lib-name)))
 (define-constant cygwin-default-dynamic-load-libraries '("cygchicken-0"))
 (define-constant macosx-load-library-extension ".dylib")
 (define-constant windows-load-library-extension ".dll")
@@ -78,11 +86,11 @@
 ; srfi-98 partially in unit posix
 
 (define-constant builtin-features
-  '(chicken srfi-2 srfi-6 srfi-10 srfi-12 srfi-23 srfi-28 srfi-30 srfi-31 srfi-39 
+  '(chicken srfi-2 srfi-6 srfi-10 srfi-12 srfi-23 srfi-28 srfi-30 srfi-39 
 	    srfi-55 srfi-88 srfi-98) )
 
 (define-constant builtin-features/compiled
-  '(srfi-6 srfi-8 srfi-9 srfi-11 srfi-15 srfi-16 srfi-17 srfi-26) )
+  '(srfi-8 srfi-9 srfi-11 srfi-15 srfi-16 srfi-17 srfi-26) )
 
 (define ##sys#chicken-prefix
   (let ((prefix (and-let* ((p (get-environment-variable prefix-environment-variable)))
@@ -105,13 +113,15 @@
 
 (define ##sys#hash-symbol
   (let ([cache-s #f]
-	[cache-h #f] )
+	[cache-h #f]
+        ;; NOTE: All low-level hash tables share the same randomization factor
+        [rand (##core#inline "C_rnd_fix")] )
     (lambda (s n)
       (if (eq? s cache-s)
 	  (##core#inline "C_fixnum_modulo" cache-h n)
           (begin
             (set! cache-s s)
-            (set! cache-h (##core#inline "C_hash_string" (##sys#slot s 1)))
+            (set! cache-h (##core#inline "C_u_i_string_hash" (##sys#slot s 1) rand))
             (##core#inline "C_fixnum_modulo" cache-h n))))))
 
 (define (##sys#hash-table-ref ht key)
@@ -157,12 +167,14 @@
 		    b
 		    (loop (##sys#slot bucket 1)) ) ) ) ) ) ) ) )
 
+(define (##sys#hash-table-size ht)
+  (let loop ((len (##sys#size ht)) (bkt 0) (size 0))
+    (if (fx= bkt len)
+        size
+        (loop len (fx+ bkt 1) (fx+ size (##sys#length (##sys#slot ht bkt)))))))
 
 ;;; Compile lambda to closure:
 
-(define ##sys#eval-environment #f)
-(define ##sys#environment-is-mutable #f)
-
 (define (##sys#eval-decorator p ll h cntr)
   (##sys#decorate-lambda
    p 
@@ -177,7 +189,6 @@
      p) ) )
 
 (define ##sys#unbound-in-eval #f)
-(define ##sys#unsafe-eval #f)
 (define ##sys#eval-debug-level (make-parameter 1))
 
 (define ##sys#compile-to-closure
@@ -186,7 +197,7 @@
 	[with-input-from-file with-input-from-file]
 	[unbound (##sys#slot '##sys#arbitrary-unbound-symbol 0)]
 	[display display] )
-    (lambda (exp env se #!optional cntr)
+    (lambda (exp env se #!optional cntr evalenv static)
 
       (define (find-id id se)		; ignores macro bindings
 	(cond ((null? se) #f)
@@ -200,7 +211,7 @@
 
       (define (lookup var0 e se)
 	(let ((var (rename var0 se)))
-	  (d `(LOOKUP/EVAL: ,var0 ,var ,e ,(map car se)))
+	  (d `(LOOKUP/EVAL: ,var0 ,var ,e ,(map (lambda (x) (car x)) se)))
 	  (let loop ((envs e) (ei 0))
 	    (cond ((null? envs) (values #f var))
 		  ((posq var (##sys#slot envs 0)) => (lambda (p) (values ei p)))
@@ -231,59 +242,26 @@
       (define (decorate p ll h cntr)
 	(##sys#eval-decorator p ll h cntr) )
 
-      (define (eval/meta form)
-	(let ((oldcm (##sys#current-module))
-	      (oldme (##sys#macro-environment))
-	      (mme (##sys#meta-macro-environment)))
-	  (dynamic-wind
-	      (lambda () 
-		(##sys#current-module #f)
-		(##sys#macro-environment mme))
-	      (lambda ()
-		((##sys#compile-to-closure
-		  form
-		  '() 
-		  (##sys#current-meta-environment))
-		 '() ) )
-	      (lambda ()
-		(##sys#current-module oldcm)
-		(##sys#meta-macro-environment (##sys#macro-environment))
-		(##sys#macro-environment oldme)))))
-
-      (define (eval/elab form)
-	((##sys#compile-to-closure
-	  form
-	  '() 
-	  (##sys#current-environment))
-	 '() ) )
-
       (define (compile x e h tf cntr se)
 	(cond ((keyword? x) (lambda v x))
 	      ((symbol? x)
 	       (receive (i j) (lookup x e se)
 		 (cond ((not i)
-			(let ((var (if (not (assq x se)) ; global?
-				       (##sys#alias-global-hook j #f cntr)
-				       (or (##sys#get j '##core#primitive) j))))
-			  (if ##sys#eval-environment
-			      (let ([loc (##sys#hash-table-location ##sys#eval-environment var #t)])
-				(unless loc
-				  (##sys#syntax-error-hook "reference to undefined identifier" var))
-				(if ##sys#unsafe-eval 
-				    (lambda v (##sys#slot loc 1))
-				    (lambda v 
-				      (let ([val (##sys#slot loc 1)])
-					(if (eq? unbound val)
-					    (##sys#error "unbound variable" var)
-					    val) ) ) ))
-			      (cond (##sys#unsafe-eval
-				     (lambda v (##core#inline "C_slot" var 0)))
-				    (else
-				     (when (and ##sys#unbound-in-eval
-						(not (##sys#symbol-has-toplevel-binding? var)))
-				       (set! ##sys#unbound-in-eval
-					 (cons (cons var cntr) ##sys#unbound-in-eval)) )
-				     (lambda v (##core#inline "C_retrieve" var)))))))
+			(let ((var (cond ((not (symbol? j)) x) ; syntax?
+                                         ((not (assq x se))
+                                          (and (not static)
+                                               (##sys#alias-global-hook j #f cntr)))
+                                         (else (or (##sys#get j '##core#primitive) j)))))
+			  (when (and ##sys#unbound-in-eval
+				     (or (not var)
+					 (not (##sys#symbol-has-toplevel-binding? var))))
+			    (set! ##sys#unbound-in-eval
+			      (cons (cons var cntr) ##sys#unbound-in-eval)) )
+			  (cond ((not var)
+				 (lambda (v)
+				   (##sys#error "unbound variable" x)))
+				(else
+				 (lambda v (##core#inline "C_retrieve" var))))))
                       (else
                        (case i
                          ((0) (lambda (v) 
@@ -326,7 +304,7 @@
 		 (d `(EVAL/EXPANDED: ,x2))
 		 (if (not (eq? x2 x))
 		     (compile x2 e h tf cntr se)
-		     (let ((head (rename (##sys#slot x 0) se))) 
+		     (let ((head (rename (##sys#slot x 0) se)))
 		       ;; here we did't resolve ##core#primitive, but that is done in compile-call (via 
 		       ;; a normal walking of the operator)
 		       (case head
@@ -347,13 +325,6 @@
 			  (let ((c (cadr x)))
 			    (lambda v c)))
 
-			 [(##core#global-ref)
-			  (let ([var (cadr x)]) ;XXX broken - should alias (see above)
-			    (if ##sys#eval-environment
-				(let ([loc (##sys#hash-table-location ##sys#eval-environment var #t)])
-				  (lambda v (##sys#slot loc 1)) )
-				(lambda v (##core#inline "C_slot" var 0)) ) ) ]
-
 			 [(##core#check)
 			  (compile (cadr x) e h tf cntr se) ]
 
@@ -394,17 +365,14 @@
 					 (and-let* ((a (assq var (##sys#current-environment)))
 						    ((symbol? (cdr a))))
 					   (##sys#notice "assignment to imported value binding" var)))
-				       (let ((var (##sys#alias-global-hook j #t cntr)))
-					 (if ##sys#eval-environment
-					     (let ([loc (##sys#hash-table-location
-							 ##sys#eval-environment 
-							 var
-							 ##sys#environment-is-mutable) ] )
-					       (unless loc
-						 (##sys#error "assignment to undefined identifier" var))
-					       (if (##sys#slot loc 2)
-						   (lambda (v) (##sys#setslot loc 1 (##core#app val v)))
-						   (lambda v (##sys#error "assignment to immutable variable" var)) ) )
+				       (let ((var
+					      (if (not (assq x se))
+						  (and (not static)
+						       (##sys#alias-global-hook j #t cntr))
+						  (or (##sys#get j '##core#primitive) j))))
+					 (if (not var) ; static
+					     (lambda (v)
+					       (##sys#error 'eval "environment is not mutable" evalenv var))
 					     (lambda (v)
 					       (##sys#setslot var 0 (##core#app val v))) ) ) ]
 				      [(zero? i) (lambda (v) (##sys#setslot (##sys#slot v 0) j (##core#app val v)))]
@@ -416,15 +384,13 @@
 			 [(##core#let)
 			  (let* ([bindings (cadr x)]
 				 [n (length bindings)] 
-				 [vars (map (lambda (x) (car x)) bindings)] 
+				 [vars (map (lambda (x) (car x)) bindings)]
 				 (aliases (map gensym vars))
 				 [e2 (cons aliases e)]
 				 (se2 (##sys#extend-se se vars aliases))
 				 [body (##sys#compile-to-closure
 					(##sys#canonicalize-body (cddr x) se2 #f)
-					e2
-					se2
-					cntr) ] )
+					e2 se2 cntr evalenv static) ] )
 			    (case n
 			      [(1) (let ([val (compile (cadar bindings) e (car vars) tf cntr se)])
 				     (lambda (v)
@@ -497,9 +463,7 @@
 				      (body 
 				       (##sys#compile-to-closure
 					(##sys#canonicalize-body body se2 #f)
-					e2
-					se2
-					(or h cntr) ) ) )
+					e2 se2 (or h cntr) evalenv static) ) )
 				 (case argc
 				   [(0) (if rest
 					    (lambda (v)
@@ -579,8 +543,9 @@
 					     (list
 					      (car b)
 					      se
-					      (##sys#er-transformer
-					       (eval/meta (cadr b)))))
+					      (##sys#ensure-transformer
+					       (##sys#eval/meta (cadr b))
+					       (##sys#strip-syntax (car b)))))
 					   (cadr x) ) 
 				      se) ) )
 			    (compile
@@ -592,8 +557,9 @@
 					    (list
 					     (car b)
 					     #f
-					     (##sys#er-transformer
-					      (eval/meta (cadr b)))))
+					     (##sys#ensure-transformer
+					      (##sys#eval/meta (cadr b))
+					      (##sys#strip-syntax (car b)))))
 					  (cadr x) ) )
 				 (se2 (append ms se)) )
 			    (for-each 
@@ -608,13 +574,15 @@
 			  (let* ((var (cadr x))
 				 (body (caddr x))
 				 (name (rename var se)))
+			    (when (and static (not (assq var se)))
+			      (##sys#error 'eval "environment is not mutable" evalenv var))
 			    (##sys#register-syntax-export 
 			     name (##sys#current-module)
-			     body)	;*** not really necessary, it only shouldn't be #f
+			     body)	; not really necessary, it only shouldn't be #f
 			    (##sys#extend-macro-environment
 			     name
 			     (##sys#current-environment)
-			     (##sys#er-transformer (eval/meta body)))
+			     (##sys#eval/meta body))
 			    (compile '(##core#undefined) e #f tf cntr se) ) )
 
 			 ((##core#define-compiler-syntax)
@@ -631,9 +599,18 @@
 			     ,@(##sys#include-forms-from-file (cadr x)))
 			   e #f tf cntr se))
 
+			 ((##core#let-module-alias)
+			  (##sys#with-module-aliases
+			   (map (lambda (b)
+				  (##sys#check-syntax 'functor b '(symbol symbol))
+				  (##sys#strip-syntax b))
+				(cadr x))
+			   (lambda ()
+			     (compile `(##core#begin ,@(cddr x)) e #f tf cntr se))))
+
 			 ((##core#module)
 			  (let* ((x (##sys#strip-syntax x))
-				 (name (##sys#strip-syntax (cadr x)))
+				 (name (cadr x))
 				 (exports 
 				  (or (eq? #t (caddr x))
 				      (map (lambda (exp)
@@ -648,41 +625,45 @@
 						    (##sys#syntax-error-hook
 						     'module
 						     "invalid export syntax" exp name))))
-					   (##sys#strip-syntax (caddr x))))))
+					   (caddr x)))))
 			    (when (##sys#current-module)
 			      (##sys#syntax-error-hook 'module "modules may not be nested" name))
 			    (parameterize ((##sys#current-module 
 					    (##sys#register-module name exports))
 					   (##sys#current-environment '())
-					   (##sys#macro-environment ##sys#initial-macro-environment))
-				(let loop ((body (cdddr x)) (xs '()))
-				  (if (null? body)
-				      (let ((xs (reverse xs)))
-					(##sys#finalize-module (##sys#current-module))
-					(lambda (v)
-					  (let loop2 ((xs xs))
-					    (if (null? xs)
-						(##sys#void)
-						(let ((n (##sys#slot xs 1)))
-						  (cond ((pair? n)
-							 ((##sys#slot xs 0) v)
-							 (loop2 n))
-							(else
-							 ((##sys#slot xs 0) v))))))))
-				      (loop 
-				       (cdr body)
-				       (cons (compile 
-					      (car body) 
-					      '() #f tf cntr 
-					      (##sys#current-environment))
-					     xs))))) ) )
+					   (##sys#macro-environment 
+					    ##sys#initial-macro-environment)
+					   (##sys#module-alias-environment
+					    (##sys#module-alias-environment)))
+			      (##sys#with-property-restore
+			       (lambda ()
+				 (let loop ((body (cdddr x)) (xs '()))
+				   (if (null? body)
+				       (let ((xs (reverse xs)))
+					 (##sys#finalize-module (##sys#current-module))
+					 (lambda (v)
+					   (let loop2 ((xs xs))
+					     (if (null? xs)
+						 (##sys#void)
+						 (let ((n (cdr xs)))
+						   (cond ((pair? n)
+							  ((car xs) v)
+							  (loop2 n))
+							 (else
+							  ((car xs) v))))))))
+				       (loop 
+					(cdr body)
+					(cons (compile 
+					       (car body) 
+					       '() #f tf cntr 
+					       (##sys#current-environment))
+					      xs))))) ) )))
 
 			 [(##core#loop-lambda)
 			  (compile `(,(rename 'lambda se) ,@(cdr x)) e #f tf cntr se) ]
 
 			 [(##core#require-for-syntax)
-			  (let ([ids (map (lambda (x)
-					    (eval/meta x))
+			  (let ([ids (map (lambda (x) (##sys#eval/meta x))
 					  (cdr x))])
 			    (apply ##sys#require ids)
 			    (let ([rs (##sys#lookup-runtime-requirements ids)])
@@ -698,14 +679,13 @@
 			     (let loop ([ids (##sys#strip-syntax (cadr x))])
 			       (if (null? ids)
 				   '(##core#undefined)
-				   (let-values ([(exp _) 
-						 (##sys#do-the-right-thing
-						  (car ids) #f imp?)])
+				   (let-values (((exp f real-id) 
+						 (##sys#do-the-right-thing (car ids) #f imp?)))
 				     `(##core#begin ,exp ,(loop (cdr ids))) ) ) )
 			     e #f tf cntr se) ) ]
 
 			 [(##core#elaborationtimeonly ##core#elaborationtimetoo) ; <- Note this!
-			  (eval/meta (cadr x))
+			  (##sys#eval/meta (cadr x))
 			  (compile '(##core#undefined) e #f tf cntr se) ]
 
 			 [(##core#compiletimetoo)
@@ -735,6 +715,20 @@
 			 [(##core#app)
 			  (compile-call (cdr x) e tf cntr se) ]
 
+			 ((##core#the)
+			  (compile (cadddr x) e h tf cntr se))
+			 
+			 ((##core#typecase)
+			  ;; drops exp and requires "else" clause
+			  (cond ((assq 'else (##sys#strip-syntax (cdddr x))) =>
+				 (lambda (cl)
+				   (compile (cadr cl) e h tf cntr se)))
+				(else
+				 (##sys#syntax-error-hook
+				  'compiler-typecase
+				  "no `else-clause' in unresolved `compiler-typecase' form"
+				  x))))
+
 			 (else
 			  (fluid-let ((##sys#syntax-context (cons head ##sys#syntax-context)))
 			    (compile-call x e tf cntr se)))))))]
@@ -764,10 +758,13 @@
 		[else #f] ) ) )
 
       (define (compile-call x e tf cntr se)
-	(let* ([fn (compile (##sys#slot x 0) e #f tf cntr se)]
-	       [args (##sys#slot x 1)]
-	       [argc (checked-length args)]
-	       [info x] )
+	(let* ((head (##sys#slot x 0))
+	       (fn (if (procedure? head) 
+		       (lambda _ head)
+		       (compile (##sys#slot x 0) e #f tf cntr se)))
+	       (args (##sys#slot x 1))
+	       (argc (checked-length args))
+	       (info x) )
 	  (case argc
 	    [(#f) (##sys#syntax-error/context "malformed expression" x)]
 	    [(0) (lambda (v)
@@ -802,21 +799,44 @@
 
       (compile exp env #f (fx> (##sys#eval-debug-level) 0) cntr se) ) ) )
 
+
+;;; evaluate in the macro-expansion/compile-time environment
+(define (##sys#eval/meta form)
+  (let ((oldcm (##sys#current-module))
+	(oldme (##sys#macro-environment))
+	(mme (##sys#meta-macro-environment))
+	(aee (##sys#active-eval-environment)))
+    (dynamic-wind
+	(lambda () 
+	  (##sys#current-module #f)
+	  (##sys#macro-environment mme)
+	  (##sys#active-eval-environment ##sys#current-meta-environment))
+	(lambda ()
+	  ((##sys#compile-to-closure
+	    form
+	    '() 
+	    (##sys#current-meta-environment)) ;XXX evalenv? static?
+	   '() ) )
+	(lambda ()
+	  (##sys#active-eval-environment aee)
+	  (##sys#current-module oldcm)
+	  (##sys#meta-macro-environment (##sys#macro-environment))
+	  (##sys#macro-environment oldme)))))
+
 (define ##sys#eval-handler 
   (make-parameter
-   (lambda (x . env)
-     (let ([mut ##sys#environment-is-mutable]
-	   [e #f] )
-       (when (pair? env)
-	 (let ([env (car env)])
-	   (when env
-	     (##sys#check-structure env 'environment)
-	     (set! e (##sys#slot env 1)) 
-	     (set! mut (##sys#slot env 2)) ) ) )
-       ((fluid-let ((##sys#environment-is-mutable mut)
-		    (##sys#eval-environment e) )
-	  (##sys#compile-to-closure x '() (##sys#current-environment)) )
-	'() ) ) ) ) )
+   (lambda (x #!optional env)
+     (let ((se (##sys#current-environment)))
+       (cond (env
+	      (##sys#check-structure env 'environment 'eval)
+	      (let ((se2 (##sys#slot env 2)))
+		((if se2		; not interaction-environment?
+		     (parameterize ((##sys#macro-environment '()))
+		       (##sys#compile-to-closure x '() se2 #f env (##sys#slot env 3)))
+		     (##sys#compile-to-closure x '() se #f env #f))
+		 '() ) ) )
+	     (else
+	      ((##sys#compile-to-closure x '() se #f #f #f) '() ) ) ) ) )))
 
 (define eval-handler ##sys#eval-handler)
 
@@ -825,6 +845,33 @@
 	 x
 	 env) )
 
+
+;;; Setting properties dynamically scoped
+
+(define-values (##sys#put/restore! ##sys#with-property-restore)
+  (let ((trail '())
+	(restoring #f))
+    (values
+     (lambda (sym prop val)
+       (when restoring
+	 (set! trail (cons (list sym prop (##sys#get sym prop)) trail)))
+       (##sys#put! sym prop val)
+       val)
+     (lambda (thunk)
+       (let ((t0 #f)
+	     (r0 restoring))
+	 (dynamic-wind
+	     (lambda ()
+	       (set! t0 trail)
+	       (set! restoring #t))
+	     thunk
+	     (lambda ()
+	       (do () ((eq? t0 trail))
+		 (apply ##sys#put! (car trail))
+		 (set! trail (cdr trail)))
+	       (set! restoring r0))))))))
+
+
 ;;; Split lambda-list into its parts:
 
 (define ##sys#decompose-lambda-list
@@ -892,25 +939,20 @@
     (lambda (input evaluator pf #!optional timer printer)
       (when (string? input) 
 	(set! input (##sys#expand-home-path input)) )
-      (let* ([isdir #f]
-	     [fname 
+      (let* ((fname 
 	      (cond [(port? input) #f]
 		    [(not (string? input)) (badfile input)]
-		    [(and-let* ([info (##sys#file-info input)]
-				[id (##sys#slot info 4)] )
-		       (set! isdir (eq? 1 id)) 
-		       (not isdir) )
-		     input]
-		    [else
+		    ((##sys#file-exists? input #t #f 'load) input)
+		    (else
 		     (let ([fname2 (##sys#string-append input ##sys#load-dynamic-extension)])
 		       (if (and (not ##sys#dload-disabled)
 				(##sys#fudge 24) ; dload?
-				(##sys#file-info fname2))
+				(##sys#file-exists? fname2 #t #f 'load))
 			   fname2
 			   (let ([fname3 (##sys#string-append input source-file-extension)])
-			     (if (##sys#file-info fname3)
+			     (if (##sys#file-exists? fname3 #t #f 'load)
 				 fname3
-				 (and (not isdir) input) ) ) ) ) ] ) ]
+				 input) ) ) ) )))
 	     [evproc (or evaluator eval)] )
 	(cond [(and (string? input) (not fname))
 	       (##sys#signal-hook #:file-error 'load "cannot open file" input) ]
@@ -920,13 +962,13 @@
 	       (display " ...\n") 
 	       (flush-output)] )
 	(or (and fname
-		 (or (##sys#dload (##sys#make-c-string fname 'load) topentry #t) 
+		 (or (##sys#dload (##sys#make-c-string fname 'load) topentry) 
 		     (and (not (has-sep? fname))
 			  (##sys#dload 
 			   (##sys#make-c-string
 			    (##sys#string-append "./" fname) 
 			    'load) 
-			   topentry #t) ) ) )
+			   topentry) ) ) )
 	    (call-with-current-continuation
 	     (lambda (abrt)
 	       (fluid-let ((##sys#read-error-with-line-number #t)
@@ -1034,7 +1076,7 @@
 		(display " ...\n") )
 	      (let loop ([libs libs])
 		(cond [(null? libs) #f]
-		      [(##sys#dload (##sys#make-c-string (##sys#slot libs 0) 'load-library) top #f)
+		      [(##sys#dload (##sys#make-c-string (##sys#slot libs 0) 'load-library) top)
 		       (unless (memq id ##sys#features) (set! ##sys#features (cons id ##sys#features)))
 		       #t]
 		      [else (loop (##sys#slot libs 1))] ) ) ) ) ) ) ) )
@@ -1125,15 +1167,15 @@
 			  (file-exists? (##sys#string-append p0 ##sys#load-dynamic-extension)))
 		     (file-exists? (##sys#string-append p0 source-file-extension)) )
 		 p0) ) )
-	  (let loop ((paths (##sys#append
-			     (if ##sys#setup-mode '(".") '())
-			     (if rp (list rp) '())
-			     (if inc? ##sys#include-pathnames '())
-			     (if ##sys#setup-mode '() '("."))) ))
-	    (and (pair? paths)
-		 (let ((pa (##sys#slot paths 0)))
-		   (or (check pa)
-		       (loop (##sys#slot paths 1)) ) ) ) ) ) ) ))
+	(let loop ((paths (##sys#append
+			   (if ##sys#setup-mode '(".") '())
+			   (if rp (list rp) '())
+			   (if inc? ##sys#include-pathnames '())
+			   (if ##sys#setup-mode '() '("."))) ))
+	  (and (pair? paths)
+	       (let ((pa (##sys#slot paths 0)))
+		 (or (check pa)
+		     (loop (##sys#slot paths 1)) ) ) ) ) ) ) ))
 
 (define ##sys#loaded-extensions '())
 
@@ -1144,7 +1186,8 @@
 	    (else (##sys#check-symbol id loc)) )
       (let ([p (##sys#canonicalize-extension-path id loc)])
 	(cond ((member p ##sys#loaded-extensions))
-	      ((memq id ##sys#core-library-modules)
+	      ((or (memq id ##sys#core-library-modules)
+		   (memq id ##sys#core-syntax-modules))
 	       (or (##sys#load-library-0 id #f)
 		   (and err?
 			(##sys#error loc "cannot load core library" id))))
@@ -1236,15 +1279,23 @@
 		   (if comp?
 		       (memq id builtin-features/compiled)
 		       (##sys#feature? id) ) )
-	       (values (impform '(##core#undefined) impid #t) #t) )
+	       (values (impform '(##core#undefined) impid #t) #t id) )
 	      ((memq id ##sys#core-library-modules)
+	       (values
+		(impform
+		 (if comp?
+		     `(##core#declare (uses ,id))
+		     `(##sys#load-library ',id #f) )
+		 impid #f)
+		#t id) )
+	      ((memq id ##sys#core-syntax-modules)
 	       (values
 		(impform
 		 (if comp?
 		     `(##core#declare (uses ,id))
 		     `(##sys#load-library ',id #f) )
 		 impid #t)
-		#t) )
+		#t id) )
 	      ((memq id ##sys#explicit-library-modules)
 	       (let* ((info (##sys#extension-information id 'require-extension))
 		      (nr (assq 'import-only info))
@@ -1259,7 +1310,7 @@
 			      `(##sys#load-library ',id #f) )
 			  '(##core#undefined))
 		      impid #f))
-		  #t) ) )
+		  #t id) ) )
 	      (else
 	       (let ((info (##sys#extension-information id 'require-extension)))
 		 (cond (info
@@ -1280,14 +1331,14 @@
 						(cond (rr (cdr rr))
 						      (else (list id)) ) ) ) ) ) ) )
 			    impid #f)
-			   #t) ) )
+			   #t id) ) )
 		       (else
 			(add-req id #f)
 			(values
 			 (impform
 			  `(##sys#require ',id) 
 			  impid #f)
-			 #f)))))))
+			 #f id)))))))
       (cond ((and (pair? id) (symbol? (car id)))
 	     (case (car id)
 	       ((srfi)
@@ -1296,11 +1347,11 @@
 			`(##core#begin
 			  ,@(map (lambda (n)
 				   (let ((rid (srfi-id n)))
-				     (let-values (((exp f2) (doit rid rid)))
+				     (let-values (((exp f2 _) (doit rid rid)))
 				       (set! f (or f f2))
 				       exp)))
 				 (cdr id)))))
-		  (values exp f)))
+		  (values exp f id)))	;XXX `id' not fully correct
 	       ((rename except only prefix)
 		(let follow ((id2 id))
 		  (if (and (pair? id2) (pair? (cdr id2)))
@@ -1333,116 +1384,66 @@
 
 ;;; Environments:
 
-(define ##sys#r4rs-environment (make-vector environment-table-size '()))
-(define ##sys#r5rs-environment #f)
-(define ##sys#interaction-environment (##sys#make-structure 'environment #f #t))
+(define interaction-environment
+  (let ((e (##sys#make-structure 'environment 'interaction-environment #f #f)))
+    (lambda () e)))
 
-(define (##sys#environment? obj)
-  (and (##sys#structure? obj 'environment) (fx= 3 (##sys#size obj))) )
+(define-record-printer (environment e p)
+  (##sys#print "#<environment " #f p)
+  (##sys#print (##sys#slot e 1) #f p)
+  (##sys#write-char-0 #\> p))
 
-(define ##sys#copy-env-table
-  (lambda (e mff mf . args)
-    (let ([syms (and (pair? args) (car args))])
-      (let* ([s (##sys#size e)]
-             [e2 (##sys#make-vector s '())] )
-       (do ([i 0 (fx+ i 1)])
-           ((fx>= i s) e2)
-         (##sys#setslot 
-          e2 i
-          (let copy ([b (##sys#slot e i)])
-            (if (null? b)
-                '()
-                (let ([bi (##sys#slot b 0)])
-                  (let ([sym (##sys#slot bi 0)])
-                    (if (or (not syms) (memq sym syms))
-                      (cons (vector
-                              sym
-                              (##sys#slot bi 1)
-                              (if mff mf (##sys#slot bi 2)))
-                            (copy (##sys#slot b 1)))
-                      (copy (##sys#slot b 1)) ) ) ) ) ) ) ) ) ) ) )
+(define scheme-report-environment)
+(define null-environment)
 
-(define ##sys#environment-symbols
-  (lambda (env . args)
-    (##sys#check-structure env 'environment)
-    (let ([pred (and (pair? args) (car args))])
-      (let ([envtbl (##sys#slot env 1)])
-        (if envtbl
-            ;then "real" environment
-          (let ([envtblsiz (vector-length envtbl)])
-            (do ([i 0 (fx+ i 1)]
-                 [syms
-                   '()
-                   (let loop ([bucket (vector-ref envtbl i)] [syms syms])
-                     (if (null? bucket)
-                       syms
-                       (let ([sym (vector-ref (car bucket) 0)])
-                         (if (or (not pred) (pred sym))
-                           (loop (cdr bucket) (cons sym syms))
-                           (loop (cdr bucket) syms) ) ) ) )])
-	        ((fx>= i envtblsiz) syms) ) )
-	    ;else interaction-environment
-	  (let ([syms '()])
-	    (##sys#walk-namespace
-	      (lambda (sym)
-	        (when (or (not pred) (pred sym))
-	          (set! syms (cons sym syms)) ) ) )
-	    syms ) ) ) ) ) )
-
-(define (interaction-environment) ##sys#interaction-environment)
-
-(define scheme-report-environment
-  (lambda (n . mutable)
-    (##sys#check-exact n 'scheme-report-environment)
-    (let ([mf (and (pair? mutable) (car mutable))])
+(let* ((r4s (module-environment 'r4rs 'scheme-report-environment/4))
+       (r5s (module-environment 'scheme 'scheme-report-environment/5))
+       (r4n (module-environment 'r4rs-null 'null-environment/4))
+       (r5n (module-environment 'r5rs-null 'null-environment/5)))
+  (define (strip se)
+    (foldr
+     (lambda (s r)
+       (if (memq (car s)
+		 '(import 
+		    require-extension 
+		    require-library 
+		    begin-for-syntax
+		    export 
+		    module
+		    cond-expand
+		    syntax
+		    reexport 
+		    import-for-syntax))
+	   r
+	   (cons s r)))
+     '()
+     se))
+  ;; Strip non-std syntax from SEs
+  (##sys#setslot r4s 2 (strip (##sys#slot r4s 2)))
+  (##sys#setslot r4n 2 (strip (##sys#slot r4n 2)))
+  (##sys#setslot r5s 2 (strip (##sys#slot r5s 2)))
+  (##sys#setslot r5n 2 (strip (##sys#slot r5n 2)))
+  (set! scheme-report-environment
+    (lambda (n)
+      (##sys#check-exact n 'scheme-report-environment)
       (case n
-	[(4) (##sys#make-structure 'environment (##sys#copy-env-table ##sys#r4rs-environment #t mf) mf)]
-	[(5) (##sys#make-structure 'environment (##sys#copy-env-table ##sys#r5rs-environment #t mf) mf)]
-	[else (##sys#error 'scheme-report-environment "no support for version" n)] ) ) ) )
-
-(define null-environment
-  (let ([make-vector make-vector])
-    (lambda (n . mutable)
+	((4) r4s)
+	((5) r5s)
+	(else 
+	 (##sys#error 
+	  'scheme-report-environment
+	  "unsupported scheme report environment version" n)) ) ) )
+  (set! null-environment
+    (lambda (n)
       (##sys#check-exact n 'null-environment)
-      (when (or (fx< n 4) (fx> n 5))
-	(##sys#error 'null-environment "no support for version" n) )
-      (##sys#make-structure
-       'environment
-       (make-vector environment-table-size '())
-       (and (pair? mutable) (car mutable)) ) ) ) )
-
-(let ()
-  (define (initb ht) 
-    (lambda (b)
-      (let ([loc (##sys#hash-table-location ht b #t)])
-        (##sys#setslot loc 1 (##sys#slot b 0)) ) ) )
-  (for-each 
-   (initb ##sys#r4rs-environment)
-   '(not boolean? eq? eqv? equal? pair? cons car cdr caar cadr cdar cddr caaar caadr cadar caddr cdaar
-     cdadr cddar cdddr caaaar caaadr caadar caaddr cadaar cadadr cadddr cdaaar cdaadr cdadar cdaddr
-     cddaar cddadr cdddar cddddr set-car! set-cdr! null? list? list length list-tail list-ref
-     append reverse memq memv member assq assv assoc symbol? symbol->string string->symbol
-     number? integer? exact? real? complex? inexact? rational? zero? odd? even? positive? negative?
-     max min + - * / = > < >= <= quotient remainder modulo gcd lcm abs floor ceiling truncate round
-     exact->inexact inexact->exact exp log expt sqrt sin cos tan asin acos atan number->string
-     string->number char? char=? char>? char<? char>=? char<=? char-ci=? char-ci<? char-ci>?
-     char-ci>=? char-ci<=? char-alphabetic? char-whitespace? char-numeric? char-upper-case?
-     char-lower-case? char-upcase char-downcase char->integer integer->char string? string=?
-     string>? string<? string>=? string<=? string-ci=? string-ci<? string-ci>? string-ci>=? string-ci<=?
-     make-string string-length string-ref string-set! string-append string-copy string->list 
-     list->string substring string-fill! vector? make-vector vector-ref vector-set! string vector
-     vector-length vector->list list->vector vector-fill! procedure? map for-each apply force 
-     call-with-current-continuation input-port? output-port? current-input-port current-output-port
-     call-with-input-file call-with-output-file open-input-file open-output-file close-input-port
-     close-output-port load read eof-object? read-char peek-char
-     write display write-char newline with-input-from-file with-output-to-file ##sys#call-with-values
-     ##sys#values ##sys#dynamic-wind ##sys#void
-     ##sys#list->vector ##sys#list ##sys#append ##sys#cons ##sys#make-promise) )
-  (set! ##sys#r5rs-environment (##sys#copy-env-table ##sys#r4rs-environment #t #t))
-  (for-each
-   (initb ##sys#r5rs-environment)
-   '(dynamic-wind values call-with-values eval scheme-report-environment null-environment interaction-environment) ) )
-
+      (case n
+	((4) r4n)
+	((5) r5n)
+	(else
+	 (##sys#error
+	  'null-environment 
+	  "unsupported null environment version" n) )))))
+            
 
 ;;; Find included file:
 
@@ -1453,8 +1454,7 @@
 (define ##sys#resolve-include-filename
   (let ((string-append string-append) )
     (define (exists? fname)
-      (let ([info (##sys#file-info fname)])
-	(and info (not (eq? 1 (##sys#slot info 4)))) ) )
+      (##sys#file-exists? fname #t #f #f))
     (lambda (fname prefer-source #!optional repo)
       (define (test2 fname lst)
 	(if (null? lst)
@@ -1487,7 +1487,6 @@
 
 ;;;; Read-Eval-Print loop:
 
-(define ##sys#repl-eval-hook #f)
 (define ##sys#repl-print-length-limit #f)
 (define ##sys#repl-read-hook #f)
 (define ##sys#repl-recent-call-chain #f) ; used in csi for ,c command
@@ -1511,7 +1510,7 @@
 	(read read)
 	(call-with-current-continuation call-with-current-continuation)
 	(string-append string-append))
-    (lambda ()
+    (lambda (#!optional (evaluator eval))
 
       (define (write-err xs)
 	(for-each (cut ##sys#repl-print-hook <> ##sys#standard-error) xs) )
@@ -1532,6 +1531,7 @@
 	    (ehandler (##sys#error-handler))
 	    (rhandler (##sys#reset-handler)) 
 	    (lv #f)
+	    (qh ##sys#quit-hook)
 	    (uie ##sys#unbound-in-eval) )
 
 	(define (saveports)
@@ -1544,82 +1544,86 @@
 	  (set! ##sys#standard-output stdout)
 	  (set! ##sys#standard-error stderr) )
 
-	(##sys#dynamic-wind
-	 (lambda ()
-	   (set! lv (load-verbose))
-	   (load-verbose #t)
-	   (##sys#error-handler
-	    (lambda (msg . args)
-	      (resetports)
-	      (##sys#print "\nError" #f ##sys#standard-error)
-	      (when msg
-		(##sys#print ": " #f ##sys#standard-error)
-		(##sys#print msg #f ##sys#standard-error) )
-	      (if (and (pair? args) (null? (cdr args)))
-		  (begin
-		    (##sys#print ": " #f ##sys#standard-error)
-		    (write-err args) )
-		  (begin
-		    (##sys#write-char-0 #\newline ##sys#standard-error)
-		    (write-err args) ) )
-	      (set! ##sys#repl-recent-call-chain
-		(or (and-let* ((lexn ##sys#last-exception) ;XXX not really right
-			       ((##sys#structure? lexn 'condition))
-			       (a (member '(exn . call-chain) (##sys#slot lexn 2))))
-		      (let ((ct (cadr a)))
-			(##sys#really-print-call-chain
-			 ##sys#standard-error ct
-			 "\n\tCall history:\n")
-			ct))
-		    (print-call-chain ##sys#standard-error)))
-	      (flush-output ##sys#standard-error) ) ) )
-	 (lambda ()
-	   (let loop ()
-	     (saveports)
-	     (call-with-current-continuation
-	      (lambda (c)
-		(##sys#reset-handler
-		 (lambda ()
-		   (set! ##sys#read-error-with-line-number #f)
-		   (set! ##sys#enable-qualifiers #t)
-		   (resetports)
-		   (c #f) ) ) ) )
-	     (##sys#read-prompt-hook)
-	     (let ([exp ((or ##sys#repl-read-hook read))])
-	       (unless (eof-object? exp)
-		 (when (char=? #\newline (##sys#peek-char-0 ##sys#standard-input))
-		   (##sys#read-char-0 ##sys#standard-input) )
-		 (##sys#clear-trace-buffer)
-		 (set! ##sys#unbound-in-eval '())
-		 (receive result ((or ##sys#repl-eval-hook eval) exp)
-		   (when (and ##sys#warnings-enabled (pair? ##sys#unbound-in-eval))
-		     (let loop ((vars ##sys#unbound-in-eval) (u '()))
-		       (cond ((null? vars)
-			      (when (pair? u)
-				(##sys#notice
-				 "the following toplevel variables are referenced but unbound:\n")
-				(for-each 
-				 (lambda (v)
-				   (##sys#print "  " #f ##sys#standard-error)
-				   (##sys#print (car v) #t ##sys#standard-error)
-				   (when (cdr v)
-				     (##sys#print " (in " #f ##sys#standard-error)
-				     (##sys#print (cdr v) #t ##sys#standard-error) 
-				     (##sys#write-char-0 #\) ##sys#standard-error) )
-				   (##sys#write-char-0 #\newline ##sys#standard-error) )
-				 u)
-				(##sys#flush-output ##sys#standard-error)))
-			     ((or (memq (caar vars) u) 
-				  (##sys#symbol-has-toplevel-binding? (caar vars)) )
-			      (loop (cdr vars) u) )
-			     (else (loop (cdr vars) (cons (car vars) u))) ) 9 ) )
-		   (write-results result) 
-		   (loop) ) ) ) ) )
-	 (lambda ()
-	   (load-verbose lv)
-	   (set! ##sys#unbound-in-eval uie)
-	   (##sys#error-handler ehandler)
-	   (##sys#reset-handler rhandler) ) ) ) ) ) )
+	(call-with-current-continuation
+	 (lambda (k)
+	   (##sys#dynamic-wind
+	    (lambda ()
+	      (set! lv (load-verbose))
+	      (set! ##sys#quit-hook (lambda (result) (k result)))
+	      (load-verbose #t)
+	      (##sys#error-handler
+	       (lambda (msg . args)
+		 (resetports)
+		 (##sys#print "\nError" #f ##sys#standard-error)
+		 (when msg
+		   (##sys#print ": " #f ##sys#standard-error)
+		   (##sys#print msg #f ##sys#standard-error) )
+		 (if (and (pair? args) (null? (cdr args)))
+		     (begin
+		       (##sys#print ": " #f ##sys#standard-error)
+		       (write-err args) )
+		     (begin
+		       (##sys#write-char-0 #\newline ##sys#standard-error)
+		       (write-err args) ) )
+		 (set! ##sys#repl-recent-call-chain
+		   (or (and-let* ((lexn ##sys#last-exception) ;XXX not really right
+				  ((##sys#structure? lexn 'condition))
+				  (a (member '(exn . call-chain) (##sys#slot lexn 2))))
+			 (let ((ct (cadr a)))
+			   (##sys#really-print-call-chain
+			    ##sys#standard-error ct
+			    "\n\tCall history:\n")
+			   ct))
+		       (print-call-chain ##sys#standard-error)))
+		 (flush-output ##sys#standard-error) ) ) )
+	    (lambda ()
+	      (let loop ()
+		(saveports)
+		(call-with-current-continuation
+		 (lambda (c)
+		   (##sys#reset-handler
+		    (lambda ()
+		      (set! ##sys#read-error-with-line-number #f)
+		      (set! ##sys#enable-qualifiers #t)
+		      (resetports)
+		      (c #f) ) ) ) )
+		(##sys#read-prompt-hook)
+		(let ([exp ((or ##sys#repl-read-hook read))])
+		  (unless (eof-object? exp)
+		    (when (char=? #\newline (##sys#peek-char-0 ##sys#standard-input))
+		      (##sys#read-char-0 ##sys#standard-input) )
+		    (##sys#clear-trace-buffer)
+		    (set! ##sys#unbound-in-eval '())
+		    (receive result (evaluator exp)
+		      (when (and ##sys#warnings-enabled (pair? ##sys#unbound-in-eval))
+			(let loop ((vars ##sys#unbound-in-eval) (u '()))
+			  (cond ((null? vars)
+				 (when (pair? u)
+				   (##sys#notice
+				    "the following toplevel variables are referenced but unbound:\n")
+				   (for-each 
+				    (lambda (v)
+				      (##sys#print "  " #f ##sys#standard-error)
+				      (##sys#print (car v) #t ##sys#standard-error)
+				      (when (cdr v)
+					(##sys#print " (in " #f ##sys#standard-error)
+					(##sys#print (cdr v) #t ##sys#standard-error) 
+					(##sys#write-char-0 #\) ##sys#standard-error) )
+				      (##sys#write-char-0 #\newline ##sys#standard-error) )
+				    u)
+				   (##sys#flush-output ##sys#standard-error)))
+				((or (memq (caar vars) u) 
+				     (##sys#symbol-has-toplevel-binding? (caar vars)) )
+				 (loop (cdr vars) u) )
+				(else (loop (cdr vars) (cons (car vars) u))) ) 9 ) )
+		      (write-results result) 
+		      (loop) ) ) ) ) )
+	    (lambda ()
+	      (load-verbose lv)
+	      (set! ##sys#quit-hook qh)
+	      (set! ##sys#unbound-in-eval uie)
+	      (##sys#error-handler ehandler)
+	      (##sys#reset-handler rhandler) ) ) ) ) ) ) ))
 
 
 ;;; SRFI-10:
