@@ -1,6 +1,6 @@
 ;;; extras.scm - Optional non-standard extensions
 ;
-; Copyright (c) 2008-2011, The Chicken Team
+; Copyright (c) 2008-2012, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
@@ -41,14 +41,13 @@
 
 (define read-file
   (let ([read read]
-	[reverse reverse] 
 	[call-with-input-file call-with-input-file] )
     (lambda (#!optional (port ##sys#standard-input) (reader read) max)
       (define (slurp port)
 	(do ((x (reader port) (reader port))
 	     (i 0 (fx+ i 1))
 	     (xs '() (cons x xs)) )
-	    ((or (eof-object? x) (and max (fx>= i max))) (reverse xs)) ) )
+	    ((or (eof-object? x) (and max (fx>= i max))) (##sys#fast-reverse xs)) ) )
       (if (port? port)
 	  (slurp port)
 	  (call-with-input-file port slurp) ) ) ) )
@@ -84,7 +83,7 @@
       (let* ([parg (pair? args)]
 	     [p (if parg (car args) ##sys#standard-input)]
 	     [limit (and parg (pair? (cdr args)) (cadr args))])
-	(##sys#check-port p 'read-line)
+	(##sys#check-input-port p #t 'read-line)
 	(cond ((##sys#slot (##sys#slot p 2) 8) => (lambda (rl) (rl p limit)))
 	      (else
 	       (let* ((buffer-len (if limit limit 256))
@@ -122,26 +121,26 @@
 	(let loop ((lns '())
 		   (n (or max 1000000000)) ) ; this is silly
 	  (if (eq? n 0)
-	      (reverse lns)
+	      (##sys#fast-reverse lns)
 	      (let ((ln (read-line port)))
 		(if (eof-object? ln)
-		    (reverse lns)
+		    (##sys#fast-reverse lns)
 		    (loop (cons ln lns) (fx- n 1)) ) ) ) ) )
       (if (string? port)
 	  (call-with-input-file port doread)
 	  (begin
-	    (##sys#check-port port 'read-lines)
+	    (##sys#check-input-port port #t 'read-lines)
 	    (doread port) ) ) ) ) )
 
 (define write-line
   (lambda (str . port)
-    (let ((p (if (##core#inline "C_eqp" port '())
-		 ##sys#standard-output
-		 (##sys#slot port 0) ) ) )
-      (##sys#check-port p 'write-line)
+    (let* ((p (if (##core#inline "C_eqp" port '())
+                  ##sys#standard-output
+                  (##sys#slot port 0) ) ))
+      (##sys#check-output-port p #t 'write-line)
       (##sys#check-string str 'write-line)
-      (display str p)
-      (newline p) ) ) )
+      ((##sys#slot (##sys#slot p 2) 3) p str) ; write-string method
+      (##sys#write-char-0 #\newline p))))
 
 
 ;;; Extended I/O 
@@ -175,7 +174,7 @@
 			 (else (fx+ n2 m))) )))))))
 
 (define (read-string! n dest #!optional (port ##sys#standard-input) (start 0))
-  (##sys#check-port port 'read-string!)
+  (##sys#check-input-port port #t 'read-string!)
   (##sys#check-string dest 'read-string!)
   (when n
     (##sys#check-exact n 'read-string!)
@@ -188,7 +187,7 @@
 
 (define ##sys#read-string/port
   (lambda (n p)
-    (##sys#check-port p 'read-string)
+    (##sys#check-input-port p #t 'read-string)
     (cond (n (##sys#check-exact n 'read-string)
 	     (let* ((str (##sys#make-string n))
 		    (n2 (##sys#read-string! n str p 0)) )
@@ -218,7 +217,7 @@
 ;; string-, process- and tcp ports.
 
 (define (read-buffered #!optional (port ##sys#standard-input))
-  (##sys#check-port port 'read-buffered)
+  (##sys#check-input-port port #t 'read-buffered)
   (let ((rb (##sys#slot (##sys#slot port 2) 9))) ; read-buffered method
     (if rb
 	(rb port)
@@ -230,7 +229,7 @@
 (define read-token
   (lambda (pred . port)
     (let ([port (optional port ##sys#standard-input)])
-      (##sys#check-port port 'read-token)
+      (##sys#check-input-port port #t 'read-token)
       (let ([out (open-output-string)])
 	(let loop ()
 	  (let ([c (##sys#peek-char-0 port)])
@@ -244,19 +243,19 @@
   (lambda (s . more)
     (##sys#check-string s 'write-string)
     (let-optionals more ([n #f] [port ##sys#standard-output])
-      (##sys#check-port port 'write-string)
+      (##sys#check-output-port port #t 'write-string)
       (when n (##sys#check-exact n 'write-string))
-      (display 
+      ((##sys#slot (##sys#slot port 2) 3) ; write-string
+       port
        (if (and n (fx< n (##sys#size s)))
 	   (##sys#substring s 0 n)
-	   s)
-       port) ) ) )
+	   s)))))
 
 
 ;;; Binary I/O
 
 (define (read-byte #!optional (port ##sys#standard-input))
-  (##sys#check-port port 'read-byte)
+  (##sys#check-input-port port #t 'read-byte)
   (let ((x (##sys#read-char-0 port)))
     (if (eof-object? x)
 	x
@@ -264,7 +263,7 @@
 
 (define (write-byte byte #!optional (port ##sys#standard-output))
   (##sys#check-exact byte 'write-byte)
-  (##sys#check-port port 'write-byte)
+  (##sys#check-output-port port #t 'write-byte)
   (##sys#write-char-0 (integer->char byte) port) )
 
 
@@ -381,15 +380,19 @@
 	       (out (get-output-string o) col) ) )
 	    ((port? obj) (out (string-append "#<port " (##sys#slot obj 3) ">") col))
 	    ((##core#inline "C_bytevectorp" obj)
-	     (if (##core#inline "C_permanentp" obj)
-		 (out "#<static blob of size" col)
-		 (out "#<blob of size " col) )
-	     (out (number->string (##core#inline "C_block_size" obj)) col)
-	     (out ">" col) )
+	     (out "#${" col)
+	     (let ((len (##sys#size obj)))
+	       (do ((i 0 (fx+ i 1)))
+		   ((fx>= i len))
+		 (let ((b (##sys#byte obj i)))
+		   (when (fx< b 16)
+		     (out "0" col))
+		   (out (##sys#number->string b 16) col)))
+	       (out "}" col)))
 	    ((##core#inline "C_lambdainfop" obj)
 	     (out "#<lambda info " col)
 	     (out (##sys#lambda-info->string obj) col)
-	     (out "#>" col) )
+	     (out ">" col) )
 	    (else (out "#<unprintable object>" col)) ) )
 
     (define (pp obj col)
@@ -532,7 +535,7 @@
       (define (pp-do expr col extra)
 	(pp-general expr col extra #f pp-expr-list pp-expr-list pp-expr))
 
-					; define formatting style (change these to suit your style)
+      ;; define formatting style (change these to suit your style)
 
       (define indent-general 2)
 
@@ -575,7 +578,7 @@
 
 (define fprintf0
   (lambda (loc port msg args)
-    (when port (##sys#check-port port loc))
+    (when port (##sys#check-output-port port #t loc))
     (let ((out (if (and port (##sys#tty-port? port))
 		   port
 		   (open-output-string))))
