@@ -1,6 +1,6 @@
 ;;;; lolevel.scm - Low-level routines for CHICKEN
 ;
-; Copyright (c) 2008-2011, The Chicken Team
+; Copyright (c) 2008-2012, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
@@ -222,9 +222,9 @@ EOF
   (##sys#check-special ptr 'pointer->address)
   (##sys#pointer->address ptr) )
 
-(define null-pointer ##sys#null-pointer)
+(define null-pointer ##sys#null-pointer) ; DEPRECATED
 
-(define (null-pointer? ptr)
+(define (null-pointer? ptr)		; DEPRECATED
   (##sys#check-special ptr 'null-pointer?)
   (##core#inline "C_null_pointerp" ptr))
 
@@ -245,8 +245,6 @@ EOF
   (foreign-lambda* nonnull-c-pointer ([c-pointer ptr] [integer off])
     "C_return((unsigned char *)ptr + off);") )
 
-(define pointer-offset pointer+)	; DEPRECATED
-
 (define align-to-word
   (let ([align (foreign-lambda integer "C_align" integer)])
     (lambda (x)
@@ -486,8 +484,6 @@ EOF
 
 ;;; Evict objects into static memory:
 
-(define-constant evict-table-size 301)
-
 (define (object-evicted? x) (##core#inline "C_permanentp" x))
 
 (define (object-evict x . allocator)
@@ -565,7 +561,9 @@ EOF
 		 (do ([i (if (##core#inline "C_specialp" x) 1 0) (fx+ i 1)])
 		     [(fx>= i n)]
 		   (release (##sys#slot x i))) )
-	       (free (##sys#address->pointer (##core#inline_allocate ("C_block_address" 4) x))) ) ] ) ) ) )
+	       (free 
+		(##sys#address->pointer
+		 (##core#inline_allocate ("C_block_address" 4) x))) ) ] ) ) ) )
 
 (define (object-size x)
   (let ([tab (make-hash-table eq?)])
@@ -616,15 +614,17 @@ EOF
   (##sys#check-become-alist alst 'object-become!)
   (##sys#become! alst) )
 
-(define (mutate-procedure old proc)
-  (##sys#check-closure old 'mutate-procedure)
-  (##sys#check-closure proc 'mutate-procedure)
+(define (mutate-procedure! old proc)
+  (##sys#check-closure old 'mutate-procedure!)
+  (##sys#check-closure proc 'mutate-procedure!)
   (let* ([n (##sys#size old)]
 	 [words (##core#inline "C_words" n)]
 	 [new (##core#inline "C_copy_block" old (##sys#make-vector words))] )
     (##sys#become! (list (cons old (proc new))))
     new ) )
 
+(define mutate-procedure mutate-procedure!) ; DEPRECATED
+
 
 ;;; pointer vectors
 
