@@ -1,6 +1,6 @@
 ;;;; optimizer.scm - The CHICKEN Scheme compiler (optimizations)
 ;
-; Copyright (c) 2008-2011, The Chicken Team
+; Copyright (c) 2008-2012, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
@@ -74,7 +74,9 @@
 	  [(if ##core#cond ##core#switch)
 	   (scan (first subs) e)
 	   (touch)
-	   (scan-each (cdr subs) e)]
+	   (scan (first subs) e)
+	   (set! previous '())
+	   (scan (second subs) e)]
 
 	  [(let)
 	   (scan-each (butlast subs) e)
@@ -82,7 +84,7 @@
 
 	  [(lambda ##core#lambda) #f]
 
-	  [(##core#call ##core#callunit) (touch)]
+	  [(##core#call) (touch)]
 
 	  [(set!)
 	   (let ((var (first params))
@@ -142,6 +144,9 @@
     (define (node-value n) (first (node-parameters n)))
     (define (touch) (set! dirty #t))
 
+    (define (invalidate-gae! gae)
+      (for-each (cut set-cdr! <> #f) gae))
+
     (define (simplify n)
       (or (and-let* ([entry (##sys#hash-table-ref simplifications (node-class n))])
 	    (any (lambda (s)
@@ -159,23 +164,23 @@
 		 entry) )
 	  n) )
 
-    (define (walk n fids)
+    (define (walk n fids gae)
       (if (memq n broken-constant-nodes)
 	  n
 	  (simplify
 	   (let* ((odirty dirty)
-		  (n1 (walk1 n fids))
+		  (n1 (walk1 n fids gae))
 		  (subs (node-subexpressions n1)) )
 	     (case (node-class n1)
 
 	       ((if)			; (This can be done by the simplifier...)
 		(cond ((constant-node? (car subs))
-		       (set! removed-ifs (+ removed-ifs 1))
+		       (set! removed-ifs (add1 removed-ifs))
 		       (touch)
 		       (walk (if (node-value (car subs))
 				 (cadr subs)
 				 (caddr subs) )
-			     fids) )
+			     fids gae) )
 		      (else n1) ) )
 
 	       ((##core#call)
@@ -187,7 +192,7 @@
 			  (constant-form-eval
 			   var
 			   (cddr subs)
-			   (lambda (ok form result)
+			   (lambda (ok form result msg)
 			     (cond ((not ok)
 				    (unless odirty (set! dirty #f))
 				    (set! broken-constant-nodes
@@ -199,41 +204,64 @@
 				    (let ((n2 (qnode result)))
 				      (make-node
 				       '##core#call
-				       '(#t)
+				       (list #t)
 				       (list (cadr subs) n2) ) ) ) )))
 			  n1) )
 		    n1) )
 
 	       (else n1) ) ) ) ) )
 
-    (define (walk1 n fids)
+    (define (replace-var var)
+      (cond ((test var 'replacable) =>
+             (lambda (rvar)
+               (let ((final-var (replace-var rvar)))
+                 ;; Store intermediate vars to avoid recurring same chain again
+                 (put! db var 'replacable final-var)
+                 final-var)))
+            (else var)))
+    
+    (define (walk1 n fids gae)
       (let ((subs (node-subexpressions n))
 	    (params (node-parameters n)) 
 	    (class (node-class n)) )
 	(case class
 
 	  ((##core#variable)
-	   (let replace ((var (first params)))
-	     (cond ((test var 'replacable) => replace)
-		   ((test var 'collapsable)
+	   (let ((var (replace-var (first params))))
+	     (cond ((test var 'collapsable)
 		    (touch)
 		    (debugging 'o "substituted constant variable" var)
 		    (qnode (car (node-parameters (test var 'value)))) )
-		   (else
-		    (if (not (eq? var (first params)))
-			(begin
-			  (touch)
-			  (set! replaced-vars (+ replaced-vars 1)) ) )
-		    (varnode var) ) ) ) )
+		   ((not (eq? var (first params)))
+		    (touch)
+		    (set! replaced-vars (+ replaced-vars 1))
+		    (varnode var))
+		   ((assq var gae) =>
+		    (lambda (a)
+		      (let ((gvar (cdr a)))
+			(cond ((and gvar
+				    (not (eq? 'no (variable-mark gvar '##compiler#inline))))
+			       (debugging 'o "propagated global variable" var gvar)
+			       (varnode gvar))
+			      (else (varnode var))))))
+		   (else (varnode var)))))
 
 	  ((let)
-	   (let ([var (first params)])
-	     (cond [(or (test var 'removable)
-			(and (test var 'contractable) (not (test var 'replacing))) )
+	   (let ((var (first params)))
+	     (cond ((or (test var 'removable)
+			(and (test var 'contractable) 
+			     (not (test var 'replacing))))
 		    (touch)
 		    (set! removed-lets (add1 removed-lets))
-		    (walk (second subs) fids) ]
-		   [else (make-node 'let params (map (cut walk <> fids) subs))] ) ) )
+		    (walk (second subs) fids gae) )
+		   (else
+		    (let ((gae (if (and (eq? '##core#variable (node-class (first subs)))
+					(test (first (node-parameters (first subs)))
+					      'global))
+				   (alist-cons var (first (node-parameters (first subs)))
+					       gae)
+				   gae)))
+		      (make-node 'let params (map (cut walk <> fids gae) subs))) ) ) ))
 
 	  ((##core#lambda)
 	   (let ((llist (third params))
@@ -249,11 +277,12 @@
 			  '##core#lambda
 			  (list (first params) (second params)
 				(cond [(and rest (test id 'explicit-rest))
-				       (debugging 'o "merged explicitly consed rest parameter" rest)
+				       (debugging 
+					'o "merged explicitly consed rest parameter" rest)
 				       (build-lambda-list used (add1 argc) #f) ]
 				      [else (build-lambda-list used argc rest)] )
 				(fourth params) )
-			  (list (walk (first subs) (cons id fids))) ) ) ) ) ]
+			  (list (walk (first subs) (cons id fids) '())) ) ) ) ) ]
 		   [(test id 'explicit-rest)
 		    (decompose-lambda-list
 		     llist
@@ -266,58 +295,65 @@
 			      (second params)
 			      (build-lambda-list vars (add1 argc) #f)
 			      (fourth params) )
-			(list (walk (first subs) (cons id fids))) ) ) ) ]
-		   [else (walk-generic n class params subs (cons id fids))] ) ) )
+			(list (walk (first subs) (cons id fids) '())) ) ) ) ]
+		   [else (walk-generic n class params subs (cons id fids) '() #f)] ) ) )
+
+	  ((##core#direct_lambda)
+	   (walk-generic n class params subs fids '() #f))
 
 	  ((##core#call)
-	   (let* ([fun (car subs)]
-		  [funclass (node-class fun)] )
+	   (let* ((fun (car subs))
+		  (funclass (node-class fun)))
 	     (case funclass
 	       [(##core#variable)
 		;; Call to named procedure:
-		(let* ([var (first (node-parameters fun))]
-		       [lval (and (not (test var 'unknown)) 
+		(let* ((var (first (node-parameters fun)))
+		       (info (call-info params var))
+		       (lval (and (not (test var 'unknown)) 
 				  (or (test var 'value)
-				      (test var 'local-value)))]
-		       [args (cdr subs)] )
-		  (cond ((test var 'contractable)
+				      (test var 'local-value))))
+		       (args (cdr subs)) )
+		  (cond ((and (test var 'contractable)
+			      (not (test var 'replacing))
+			      ;; inlinable procedure has changed
+			      (not (test (first (node-parameters lval)) 'inline-target)))
+			 ;; only called once
 			 (let* ([lparams (node-parameters lval)]
 				[llist (third lparams)] )
 			   (check-signature var args llist)
-			   (debugging 'o "contracted procedure" var)
+			   (debugging 'o "contracted procedure" info)
 			   (touch)
 			   (for-each (cut put! db <> 'inline-target #t) fids)
 			   (walk
 			    (inline-lambda-bindings
 			     llist args (first (node-subexpressions lval)) #f db
 			     void)
-			    fids) ) )
-			((variable-mark var '##compiler#pure) =>
-			 (lambda (pb)
-			   (or (and-let* ((k (car args))
-					  ((or (eq? #t pb) 
-					       (let ((im (variable-mark var '##compiler#intrinsic)))
-						 (or (eq? im 'internal) (eq? im pb)))))
-					  ((eq? '##core#variable (node-class k)))
-					  (kvar (first (node-parameters k)))
-					  (lval (and (not (test kvar 'unknown)) (test kvar 'value))) 
-					  ((eq? '##core#lambda (node-class lval)))
-					  (llist (third (node-parameters lval)))
-					  ((or (test (car llist) 'unused)
-					       (and (not (test (car llist) 'references))
-						    (not (test (car llist) 'assigned)))))
-					  ((not (any (cut expression-has-side-effects? <> db) (cdr args) ))))
-				 (let ((info (and (pair? (cdr params)) (second params))))
-				   (debugging 
-				    'o
-				    "removed call to pure procedure with unused result"
-				    (or (source-info->string info) var)))
-				 (make-node
-				  '##core#call '(#t)
-				  (list k (make-node '##core#undefined '() '())) ) ) 
-			       (walk-generic n class params subs fids)) ) )
+			    fids gae) ) )
+			((variable-mark var '##compiler#pure)
+			 ;; callee is side-effect free
+			 (or (and-let* ((k (car args))
+					((eq? '##core#variable (node-class k)))
+					(kvar (first (node-parameters k)))
+					(lval (and (not (test kvar 'unknown)) 
+						   (test kvar 'value))) 
+					((eq? '##core#lambda (node-class lval)))
+					(llist (third (node-parameters lval)))
+					((or (test (car llist) 'unused)
+					     (and (not (test (car llist) 'references))
+						  (not (test (car llist) 'assigned)))))
+					((not (any (cut expression-has-side-effects? <> db)
+						   (cdr args) ))))
+			       (debugging 
+				'o
+				"removed call to pure procedure with unused result"
+				info)
+			       (make-node
+				'##core#call (list #t)
+				(list k (make-node '##core#undefined '() '())) ) ) 
+			     (walk-generic n class params subs fids gae #f)) )
 			((and lval
 			      (eq? '##core#lambda (node-class lval)))
+			 ;; callee is a lambda
 			 (let* ([lparams (node-parameters lval)]
 				[llist (third lparams)] )
 			   (decompose-lambda-list
@@ -339,43 +375,47 @@
 					(if external
 					    "global inlining" 	
 					    "inlining")
-					var ifid (fourth lparams))
+					info ifid (fourth lparams))
 				       (for-each (cut put! db <> 'inline-target #t) fids)
 				       (check-signature var args llist)
-				       (debugging 'o "inlining procedure" var)
+				       (debugging 'o "inlining procedure" info)
 				       (call/cc
 					(lambda (return)
 					  (define (cfk cvar)
 					    (debugging 
 					     'i
 					     "not inlining procedure because it refers to contractable"
-					     var cvar)
-					    (return (walk-generic n class params subs fids)))
+					     info cvar)
+					    (return 
+					     (walk-generic n class params subs fids gae #t)))
 					  (let ((n2 (inline-lambda-bindings
 						     llist args (first (node-subexpressions lval))
 						     #t db cfk)))
 					    (touch)
-					    (walk n2 fids)))))
+					    (walk n2 fids gae)))))
 				      ((test ifid 'has-unused-parameters)
 				       (if (< (length args) argc) ; Expression was already optimized (should this happen?)
-					   (walk-generic n class params subs fids)
+					   (walk-generic n class params subs fids gae #t)
 					   (let loop ((vars vars) (argc argc) (args args) (used '()))
 					     (cond [(or (null? vars) (zero? argc))
 						    (touch)
-						    (make-node
-						     '##core#call
-						     params
-						     (map (cut walk <> fids) (cons fun (append-reverse used args))) ) ]
+						    (let ((args
+							   (map (cut walk <> fids gae)
+								(cons 
+								 fun
+								 (append-reverse used args))) ) )
+						      (invalidate-gae! gae)
+						      (make-node '##core#call params args))]
 						   [(test (car vars) 'unused)
 						    (touch)
 						    (debugging
 						     'o "removed unused parameter to known procedure" 
-						     (car vars) var)
+						     (car vars) info)
 						    (if (expression-has-side-effects? (car args) db)
 							(make-node
 							 'let
 							 (list (gensym 't))
-							 (list (walk (car args) fids)
+							 (list (walk (car args) fids gae)
 							       (loop (cdr vars) (sub1 argc) (cdr args) used) ) )
 							(loop (cdr vars) (sub1 argc) (cdr args) used) ) ]
 						   [else (loop (cdr vars)
@@ -386,14 +426,14 @@
 					    (not (memq n rest-consers)) ) ; make sure we haven't inlined rest-list already
 				       (let ([n (llist-length llist)])
 					 (if (< (length args) n)
-					     (walk-generic n class params subs fids)
+					     (walk-generic n class params subs fids gae #t) 
 					     (begin
-					       (debugging 'o "consed rest parameter at call site" var n)
+					       (debugging 'o "consed rest parameter at call site" info n)
 					       (let-values ([(args rargs) (split-at args n)])
 						 (let ([n2 (make-node
 							    '##core#call
 							    params
-							    (map (cut walk <> fids)
+							    (map (cut walk <> fids gae)
 								 (cons fun
 								       (append 
 									args
@@ -405,14 +445,30 @@
 									      (list "C_a_i_list" (* 3 (length rargs)))
 									      rargs) ) ) ) ) ) ) ] )
 						   (set! rest-consers (cons n2 rest-consers))
+						   (invalidate-gae! gae)
 						   n2) ) ) ) ) )
-				      (else (walk-generic n class params subs fids)) ) ) ) ) ) )
-			(else (walk-generic n class params subs fids)) ) ) ]
+				      (else (walk-generic n class params subs fids gae #t)) ) ) ) ) ) )
+			((and lval
+			      (eq? '##core#variable (node-class lval))
+			      (intrinsic? (first (node-parameters lval))))
+			 ;; callee is intrinsic
+			 (debugging 'i "inlining call to intrinsic alias" 
+				    info (first (node-parameters lval)))
+			 (walk
+			  (make-node
+			   '##core#call
+			   params
+			   (cons lval (cdr subs)))
+			  fids gae))
+			(else (walk-generic n class params subs fids gae #t)) ) ) ]
 	       [(##core#lambda)
 		(if (first params)
-		    (walk-generic n class params subs fids)
-		    (make-node '##core#call (cons #t (cdr params)) (map (cut walk <> fids) subs)) ) ]
-	       [else (walk-generic n class params subs fids)] ) ) )
+		    (walk-generic n class params subs fids gae #f)
+		    (let ((n2 (make-node '##core#call (cons #t (cdr params))
+					 (map (cut walk <> fids gae) subs)) ))
+		      (invalidate-gae! gae)
+		      n2))]
+	       [else (walk-generic n class params subs fids gae #t)] ) ) )
 
 	  ((set!)
 	   (let ([var (first params)])
@@ -424,39 +480,61 @@
 		   ((test var 'replacable)
 		    (touch)
 		    (make-node '##core#undefined '() '()) )
-		   [(and (or (not (test var 'global))
+		   ((and (or (not (test var 'global))
 			     (not (variable-visible? var)))
 			 (not (test var 'inline-transient))
 			 (not (test var 'references)) 
 			 (not (expression-has-side-effects? (first subs) db)) )
 		    (touch)
 		    (debugging 'o "removed side-effect free assignment to unused variable" var)
-		    (make-node '##core#undefined '() '()) ]
-		   [else (make-node 'set! params (list (walk (car subs) fids)))] ) ) )
+		    (make-node '##core#undefined '() '()) )
+		   (else
+		    (let ((n2 (make-node 'set! params (list (walk (car subs) fids gae)))))
+		      (for-each
+		       (if (test var 'global)
+			   (lambda (a)
+			     (when (eq? var (cdr a)) ; assignment to alias?
+			       (set-cdr! a #f)))
+			   (lambda (a)
+			     (when (eq? var (car a))
+			       (set-cdr! a #f))))
+		       gae)
+		      n2)))))
 
-	  (else (walk-generic n class params subs fids)) ) ) )
+	  (else (walk-generic n class params subs fids gae #f)) ) ) )
     
-    (define (walk-generic n class params subs fids)
-      (let ((subs2 (map (cut walk <> fids) subs)))
-	(if (every eq? subs subs2)
-	    n
-	    (make-node class params subs2) ) ) )
+    (define (walk-generic n class params subs fids gae invgae)
+      (let lp ((same? #t)
+               (subs subs)
+               (subs2 '()))
+        (cond ((null? subs)
+               (when invgae (invalidate-gae! gae))
+               ;; Create new node if walk made changes, otherwise original node
+               (if same? n (make-node class params (reverse subs2))))
+              (else
+               (let ((sub2 (walk (car subs) fids gae)))
+                 (lp (and same? (eq? sub2 (car subs)))
+                     (cdr subs) (cons sub2 subs2)))) ) ))
 
     (if (perform-pre-optimization! node db)
 	(values node #t)
 	(begin
 	  (debugging 'p "traversal phase...")
 	  (set! simplified-ops '())
-	  (let ((node2 (walk node '())))
+	  (let ((node2 (walk node '() '())))
 	    (when (pair? simplified-classes) (debugging 'o "simplifications" simplified-classes))
-	    (when (and (pair? simplified-ops) (debugging 'o "  call simplifications:"))
-	      (for-each
-	       (lambda (p)
-		 (print* #\tab (car p))
-		 (if (> (cdr p) 1)
-		     (print #\tab (cdr p))
-		     (newline) ) )
-	       simplified-ops) )
+	    (when (pair? simplified-ops)
+	      (with-debugging-output
+	       'o
+	       (lambda ()
+		 (print "  call simplifications:")
+		 (for-each
+		  (lambda (p)
+		    (print* "    " (car p))
+		    (if (> (cdr p) 1)
+			(print #\tab (cdr p))
+			(newline) ) )
+		  simplified-ops) ) ) )
 	    (when (> replaced-vars 0) (debugging 'o "replaced variables" replaced-vars))
 	    (when (> removed-lets 0) (debugging 'o "removed binding forms" removed-lets))
 	    (when (> removed-ifs 0) (debugging 'o "removed conditional forms" removed-ifs))
@@ -858,10 +936,10 @@
 		     (and (eq? '##core#variable (node-class arg1))
 			  (eq? '##core#variable (node-class arg2))
 			  (equal? (node-parameters arg1) (node-parameters arg2))
-			  (make-node '##core#call '(#t) (list cont (qnode #t))) ) ) )
+			  (make-node '##core#call (list #t) (list cont (qnode #t))) ) ) )
 	      (and inline-substitutions-enabled
 		   (make-node
-		    '##core#call '(#t) 
+		    '##core#call (list #t) 
 		    (list cont (make-node '##core#inline (list (second classargs)) callargs)) ) ) ) ) )
 
     ;; (<op> ...) -> (##core#inline <iop> ...)
@@ -873,19 +951,21 @@
 	  (or (third classargs) unsafe)
 	  (let ((arg1 (first callargs)))
 	    (make-node
-	     '##core#call '(#t)
+	     '##core#call (list #t)
 	     (list 
 	      cont
 	      (make-node '##core#inline (list (second classargs)) callargs) ) ) ) ) )
 
     ;; (<op> ...) -> <var>
-    ((3) ; classargs = (<var>)
+    ((3) ; classargs = (<var> <argc>)
+     ;; - <argc> may be #f
      (and inline-substitutions-enabled
 	  (intrinsic? name)
+	  (or (not (second classargs)) (= (length callargs) (second classargs)))
 	  (fold-right
 	   (lambda (val body)
-	     (make-node 'let (list (gensym "t")) (list val body)) )
-	   (make-node '##core#call '(#t) (list cont (varnode (first classargs))))
+	     (make-node 'let (list (gensym)) (list val body)) )
+	   (make-node '##core#call (list #t) (list cont (varnode (first classargs))))
 	   callargs)))
 
     ;; (<op> a b) -> (<primitiveop> a (quote <i>) b)
@@ -909,7 +989,7 @@
 	  (= 1 (length callargs))
 	  (let ((ntype (third classargs)))
 	    (or (not ntype) (eq? ntype number-type)) )
-	  (make-node '##core#call '(#t)
+	  (make-node '##core#call (list #t)
 		     (list cont
 			   (make-node '##core#inline (list (first classargs))
 				      (list (first callargs)
@@ -921,7 +1001,7 @@
 	   inline-substitutions-enabled
 	   (= 1 (length callargs))
 	   (intrinsic? name)
-	   (make-node '##core#call '(#t)
+	   (make-node '##core#call (list #t)
 		      (list cont
 			    (make-node '##core#inline (list (first classargs))
 				       (list (make-node '##core#inline (list (second classargs))
@@ -933,7 +1013,7 @@
 	  inline-substitutions-enabled
 	  (= (length callargs) (first classargs))
 	  (intrinsic? name)
-	  (make-node '##core#call '(#t)
+	  (make-node '##core#call (list #t)
 		     (list cont
 			   (make-node '##core#inline (list (second classargs))
 				      (append callargs
@@ -951,7 +1031,7 @@
      (and inline-substitutions-enabled
 	  (intrinsic? name)
 	  (if (< (length callargs) 2)
-	      (make-node '##core#call '(#t) (list cont (qnode #t)))
+	      (make-node '##core#call (list #t) (list cont (qnode #t)))
 	      (and (or (and unsafe (not (eq? number-type 'generic)))
 		       (and (eq? number-type 'fixnum) (third classargs))
 		       (and (eq? number-type 'flonum) (fourth classargs)) )
@@ -960,7 +1040,7 @@
 		     (fold-right
 		      (lambda (x n y) (make-node 'let (list n) (list x y)))
 		      (make-node
-		       '##core#call '(#t)
+		       '##core#call (list #t)
 		       (list 
 			cont
 			(let ([op (list
@@ -1011,7 +1091,7 @@
 	  (let ((n (length callargs)))
 	    (and (<= n (third classargs))
 		 (case n
-		   ((1) (make-node '##core#call '(#t) (cons cont callargs)))
+		   ((1) (make-node '##core#call (list #t) (cons cont callargs)))
 		   (else (make-node '##core#call (list #t (first classargs))
 				    (cons* (varnode (first classargs))
 					   cont callargs) ) ) ) ) ) ) )
@@ -1034,7 +1114,7 @@
 	  (eq? number-type (first classargs))
 	  (or (fourth classargs) unsafe)
 	  (make-node
-	   '##core#call '(#t)
+	   '##core#call (list #t)
 	   (list cont
 		 (make-node
 		  '##core#inline
@@ -1052,7 +1132,7 @@
 		 (make-node '##core#call (list #t (third classargs))
 			    (cons* (varnode (third classargs)) cont callargs) ) )
 		((eq? number-type (second classargs))
-		 (make-node '##core#call '(#t) (cons cont callargs)) )
+		 (make-node '##core#call (list #t) (cons cont callargs)) )
 		(else #f) ) ) )
 
     ;; (<alloc-op> ...) -> (##core#inline_allocate (<aiop> <words>) ...)
@@ -1064,8 +1144,6 @@
      ;;   number of arguments plus 1.
      ;; - if <counted> is given and true and <argc> is between 1-8, append "<count>"
      ;;   to the name of the inline routine.
-     ;; - if <safe> is 'specialized and `unsafe-specialized-arithmetic' is declared,
-     ;;   then assume it is safe
      (let ((argc (first classargs))
 	   (rargc (length callargs))
 	   (safe (third classargs))
@@ -1074,12 +1152,9 @@
        (and inline-substitutions-enabled
 	    (or (not argc) (= rargc argc))
 	    (intrinsic? name)
-	    (or unsafe
-		(if (eq? safe 'specialized)
-		    unchecked-specialized-arithmetic
-		    safe))
+	    (or unsafe safe)
 	    (make-node
-	     '##core#call '(#t)
+	     '##core#call (list #t)
 	     (list cont 
 		   (make-node
 		    '##core#inline_allocate
@@ -1097,7 +1172,7 @@
 	  (= (length callargs) (first classargs))
 	  (intrinsic? name)
 	  (make-node
-	   '##core#call '(#t)
+	   '##core#call (list #t)
 	   (list cont
 		 (make-node '##core#inline
 			    (list (if (and unsafe (pair? (cddr classargs)))
@@ -1110,7 +1185,7 @@
      (and inline-substitutions-enabled
 	  (null? callargs)
 	  (intrinsic? name)
-	  (make-node '##core#call '(#t) (list cont (qnode (first classargs))) ) ) )
+	  (make-node '##core#call (list #t) (list cont (qnode (first classargs))) ) ) )
 
     ;; (<op>) -> <id>
     ;; (<op> <x>) -> <x>
@@ -1128,12 +1203,12 @@
 		     (and (eq? 'quote (node-class x))
 			  (eq? id (first (node-parameters x))) ) ) 
 		   callargs) ] )
-	    (cond [(null? callargs) (make-node '##core#call '(#t) (list cont (qnode id)))]
+	    (cond [(null? callargs) (make-node '##core#call (list #t) (list cont (qnode id)))]
 		  [(null? (cdr callargs))
-		   (make-node '##core#call '(#t) (list cont (first callargs))) ]
+		   (make-node '##core#call (list #t) (list cont (first callargs))) ]
 		  [(or (fourth classargs) (eq? number-type 'fixnum))
 		   (make-node
-		    '##core#call '(#t)
+		    '##core#call (list #t)
 		    (list
 		     cont
 		     (fold-inner
@@ -1150,7 +1225,7 @@
 	    (= n (first classargs))
 	    (intrinsic? name)
 	    (make-node
-	     '##core#call '(#t)
+	     '##core#call (list #t)
 	     (list cont
 		   (make-node 
 		    '##core#inline (list (second classargs))
@@ -1177,12 +1252,12 @@
 		     (and (eq? 'quote (node-class x))
 			  (eq? id (first (node-parameters x))) ) ) 
 		   callargs) ] )
-	    (cond [(null? callargs) (make-node '##core#call '(#t) (list cont (qnode id)))]
+	    (cond [(null? callargs) (make-node '##core#call (list #t) (list cont (qnode id)))]
 		  [(null? (cdr callargs))
-		   (make-node '##core#call '(#t) (list cont (first callargs))) ]
+		   (make-node '##core#call (list #t) (list cont (first callargs))) ]
 		  [else
 		   (make-node
-		    '##core#call '(#t)
+		    '##core#call (list #t)
 		    (list
 		     cont
 		     (fold-inner
@@ -1203,7 +1278,7 @@
 	    (intrinsic? name)
 	    (or (third classargs) unsafe)
 	    (make-node
-	     '##core#call '(#t)
+	     '##core#call (list #t)
 	     (list cont 
 		   (if (eq? number-type 'fixnum)
 		       (make-node
@@ -1482,3 +1557,208 @@
     (debugging 'p "direct leaf routine optimization pass...")
     (walk #f node #f)
     dirty) )
+
+
+;;; turn groups of local procedures into dispatch loop ("clustering")
+;
+; This turns (in bodies)
+;
+;   :
+;   (define (a x) (b x))
+;   (define (b y) (a y))
+;   (a z)))
+;
+; into something similar to
+;
+;   (letrec ((<dispatch>
+;              (lambda (<a1> <i>)
+;                (case <i>
+;                  ((1) (let ((x <a1>)) (<dispatch> x 2)))
+;                  ((2) (let ((y <a1>)) (<dispatch> y 1)))
+;                  (else (<dispatch> z 1))))))
+;     (<dispatch> #f 0))
+
+(define (determine-loop-and-dispatch node db)
+  (let ((groups '())
+	(outer #f)
+	(group '()))
+
+    (define (close)			; "close" group of local definitions
+      (when (pair? group)
+	(when (> (length group) 1)
+	  (set! groups (alist-cons outer group groups)))
+	(set! group '())
+	(set! outer #f)))
+
+    (define (user-lambda? n)
+      (and (eq? '##core#lambda (node-class n))
+	   (list? (third (node-parameters n))))) ; no rest argument allowed
+
+    (define (walk n e)
+      (let ((subs (node-subexpressions n))
+	    (params (node-parameters n)) 
+	    (class (node-class n)) )
+	(case class
+	  ((let)
+	   (let ((var (first params))
+		 (val (first subs))
+		 (body (second subs)))
+	     (cond ((and (not outer) 
+			 (eq? '##core#undefined (node-class val)))
+		    ;; find outermost "(let ((VAR (##core#undefined))) ...)"
+		    (set! outer n)
+		    (walk body (cons var e)))
+		   ((and outer
+			 (eq? 'set! (node-class val))
+			 (let ((sval (first (node-subexpressions val)))
+			       (svar (first (node-parameters val))))
+			   ;;XXX should we also accept "##core#direct_lambda" ?
+			   (and (eq? '##core#lambda (node-class sval))
+				(= (length (or (get db svar 'references) '()))
+				   (length (or (get db svar 'call-sites) '())))
+				(memq svar e)
+				(user-lambda? sval))))
+		    ;; "(set! VAR (lambda ...))" - add to group
+		    (set! group (cons val group))
+		    (walk body (cons var e)))
+		   (else
+		    ;; other "let" binding, close group (if any)
+		    (close)
+		    (walk val e)
+		    (walk body (cons var e))))))
+	  ((##core#lambda ##core#direct_lambda)
+	   (decompose-lambda-list
+	    (third params)
+	    (lambda (vars argc rest)
+	      ;; walk recursively, with cleared cluster state
+	      (fluid-let ((group '())
+			  (outer #f))
+		(walk (first subs) vars)))))
+	  (else
+	   ;; other form, close group (if any)
+	   (close)
+	   (for-each (cut walk <> e) subs)))))
+
+    (debugging 'p "collecting clusters ...")
+
+    ;; walk once and gather groups
+    (walk node '())
+
+    ;; process found clusters
+    (for-each
+     (lambda (g)
+       (let* ((outer (car g))
+	      (group (cdr g))
+	      (dname (gensym 'dispatch))
+	      (i (gensym 'i))
+	      (n 1)
+	      (bodies
+	       (map (lambda (assign)
+		      ;; collect information and replace assignment
+		      ;; with "(##core#undefined)"
+		      (let* ((name (first (node-parameters assign)))
+			     (proc (first (node-subexpressions assign)))
+			     (pparams (node-parameters proc))
+			     (llist (third pparams))
+			     (aliases (map gensym llist)))
+			(decompose-lambda-list
+			 llist
+			 (lambda (vars argc rest)
+			   (let ((body (first (node-subexpressions proc)))
+				 (m n))
+			     (set! n (add1 n))
+			     (copy-node!
+			      (make-node '##core#undefined '() '())
+			      assign)
+			     (list name m llist body))))))
+		    group))
+	      (k (gensym 'k))
+	      (maxargs (apply max (map (o length third) bodies)))
+	      (dllist (append
+		       (list-tabulate maxargs (lambda _ (gensym 'a)))
+		       (list i))))
+
+	 (debugging 'x "clustering" (map first bodies)) ;XXX
+
+	 ;; first descend into "(let ((_ (##core#undefined))) ...)" forms
+	 ;; to make them visible everywhere
+
+	 (let descend ((outer outer))
+	   ;;(print "outer: " (node-parameters outer))
+	   (let ((body (second (node-subexpressions outer))))
+	     (if (and (eq? 'let (node-class body))
+		      (let ((val (first (node-subexpressions body))))
+			(eq? '##core#undefined (node-class val))))
+		 (descend body)
+		 ;; wrap cluster into dispatch procedure
+		 (copy-node!
+		  (make-node
+		   'let
+		   (list dname)
+		   (list
+		    (make-node '##core#undefined '() '())
+		    (make-node
+		     'let (list (gensym))
+		     (list
+		      (make-node 
+		       'set! (list dname)
+		       (list
+			(make-node
+			 '##core#lambda
+			 (list (gensym 'f_) #t dllist 0)
+			 (list
+			  ;; dispatch to cluster member or main body
+			  (make-node
+			   '##core#switch
+			   (list (sub1 n))
+			   (append
+			    (list (varnode i))
+			    (append-map
+			     (lambda (b)
+			       (list (qnode (second b))
+				     (let loop ((args dllist)
+						(vars (third b)))
+				       (if (null? vars)
+					   (fourth b)
+					   (make-node
+					    'let (list (car vars))
+					    (list (varnode (car args))
+						  (loop (cdr args) (cdr vars))))))))
+			     bodies)
+			    (cdr (node-subexpressions outer))))))))
+		      ;; call to enter dispatch loop - the current continuation is
+		      ;; not used, so the first parameter is passed as "#f" (it is
+		      ;; a tail call)
+		      (make-node
+		       '##core#call '(#t)
+		       (cons* (varnode dname)
+			      (append
+			       (list-tabulate maxargs (lambda _ (qnode #f)))
+			       (list (qnode 0)))))))))
+		  outer))))
+
+	 ;; modify call-sites to invoke dispatch loop instead
+	 (for-each
+	  (lambda (b)
+	    (let ((sites (get db (car b) 'call-sites)))
+	      (for-each
+	       (lambda (site)
+		 (let* ((callnode (cdr site))
+			(args (cdr (node-subexpressions callnode))))
+		   (copy-node!
+		    (make-node
+		     '##core#call (node-parameters callnode)
+		     (cons* (varnode dname)
+			    (append
+			     args
+			     (list-tabulate
+			      (- maxargs (length args))
+			      (lambda _ (qnode #f)))
+			     (list (qnode (second b))))))
+		    callnode)))
+	       sites)))
+	  bodies)))
+
+     groups)
+    (values node (pair? groups))))
+
