@@ -1,6 +1,6 @@
 ;;;; posixunix.scm - Miscellaneous file- and process-handling routines
 ;
-; Copyright (c) 2008-2011, The Chicken Team
+; Copyright (c) 2008-2012, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
@@ -142,6 +142,7 @@ static C_TLS char C_hostbuf[ 256 ];
 static C_TLS struct stat C_statbuf;
 
 #define C_mkdir(str)        C_fix(mkdir(C_c_string(str), S_IRWXU | S_IRWXG | S_IRWXO))
+#define C_fchdir(fd)        C_fix(fchdir(C_unfix(fd)))
 #define C_chdir(str)        C_fix(chdir(C_c_string(str)))
 #define C_rmdir(str)        C_fix(rmdir(C_c_string(str)))
 
@@ -156,8 +157,6 @@ static C_TLS struct stat C_statbuf;
 #define open_text_output_pipe(a, n, name)    open_binary_output_pipe(a, n, name)
 #define close_pipe(p)                        C_fix(pclose(C_port_file(p)))
 
-#define C_set_file_ptr(port, ptr)  (C_set_block_item(port, 0, (C_block_item(ptr, 0))), C_SCHEME_UNDEFINED)
-
 #define C_fork              fork
 #define C_waitpid(id, o)    C_fix(waitpid(C_unfix(id), &C_wait_status, C_unfix(o)))
 #define C_getppid           getppid
@@ -190,10 +189,6 @@ static C_TLS struct stat C_statbuf;
 #define C_truncate(f, n)    C_fix(truncate((char *)C_data_pointer(f), C_num_to_int(n)))
 #define C_ftruncate(f, n)   C_fix(ftruncate(C_unfix(f), C_num_to_int(n)))
 #define C_uname             C_fix(uname(&C_utsname))
-#define C_fdopen(a, n, fd, m) C_mpointer(a, fdopen(C_unfix(fd), C_c_string(m)))
-#define C_C_fileno(p)       C_fix(fileno(C_port_file(p)))
-#define C_dup(x)            C_fix(dup(C_unfix(x)))
-#define C_dup2(x, y)        C_fix(dup2(C_unfix(x), C_unfix(y)))
 #define C_alarm             alarm
 #define C_setvbuf(p, m, s)  C_fix(setvbuf(C_port_file(p), NULL, C_unfix(m), C_unfix(s)))
 #define C_test_access(fn, m)     C_fix(access((char *)C_data_pointer(fn), C_unfix(m)))
@@ -229,6 +224,7 @@ static void C_fcall C_set_arg_string(char **where, int i, char *a, int len) {
     ptr = (char *)C_malloc(len + 1);
     C_memcpy(ptr, a, len);
     ptr[ len ] = '\0';
+    /* Can't barf() here, so the NUL byte check happens in Scheme */
   }
   else ptr = NULL;
   where[ i ] = ptr;
@@ -473,6 +469,26 @@ static int set_file_mtime(char *filename, C_word tm)
   return utime(filename, &tb);
 }
 
+static C_word C_i_fifo_p(C_word name) 
+{
+  struct stat buf;
+  int res;
+
+  res = stat(C_c_string(name), &buf);
+
+  if(res != 0) {
+#ifdef __CYGWIN__
+    return C_SCHEME_FALSE;
+#else
+    if(errno == ENOENT) return C_fix(0);
+    else return C_fix(res);
+#endif
+  }
+
+  if((buf.st_mode & S_IFMT) == S_IFIFO) return C_SCHEME_TRUE;
+  else return C_SCHEME_FALSE;
+}
+
 EOF
 ) )
 
@@ -768,6 +784,12 @@ EOF
 	(posix-error #:file-error 'change-directory "cannot change current directory" name) )
       name)))
 
+(define (change-directory* fd)
+  (##sys#check-exact fd 'change-directory*) 
+  (unless (fx= 0 (##core#inline "C_fchdir" fd)) 
+    (posix-error #:file-error 'change-directory* "cannot change current directory" fd) )
+  fd)
+
 
 ;;; Pipes:
 
@@ -804,11 +826,18 @@ EOF
 	   (else (badmode m)) ) ) ) ) )
   (set! close-input-pipe
     (lambda (port)
-      (##sys#check-port port 'close-input-pipe)
+      (##sys#check-input-port port #t 'close-input-pipe)
       (let ((r (##core#inline "close_pipe" port)))
-	(when (eq? -1 r) (posix-error #:file-error 'close-input/output-pipe "error while closing pipe" port))
+	(when (eq? -1 r)
+	  (posix-error #:file-error 'close-input-pipe "error while closing pipe" port))
 	r) ) )
-  (set! close-output-pipe close-input-pipe) )
+  (set! close-output-pipe
+    (lambda (port)
+      (##sys#check-output-port port #t 'close-output-pipe)
+      (let ((r (##core#inline "close_pipe" port)))
+	(when (eq? -1 r) 
+	  (posix-error #:file-error 'close-output-pipe "error while closing pipe" port))
+	r) ) ))
 
 (define call-with-input-pipe
   (lambda (cmd proc . mode)
@@ -830,24 +859,20 @@ EOF
 
 (define with-input-from-pipe
   (lambda (cmd thunk . mode)
-    (let ([old ##sys#standard-input]
-	  [p (apply open-input-pipe cmd mode)] )
-      (set! ##sys#standard-input p)
-      (##sys#call-with-values thunk
-			      (lambda results
-				(close-input-pipe p)
-				(set! ##sys#standard-input old)
-				(apply values results) ) ) ) ) )
+    (let ([p (apply open-input-pipe cmd mode)])
+      (fluid-let ((##sys#standard-input p))
+	(##sys#call-with-values thunk
+				(lambda results
+				  (close-input-pipe p)
+				  (apply values results) ) ) ) ) ) )
 (define with-output-to-pipe
   (lambda (cmd thunk . mode)
-    (let ([old ##sys#standard-output]
-	  [p (apply open-output-pipe cmd mode)] )
-      (set! ##sys#standard-output p)
-      (##sys#call-with-values thunk
-			      (lambda results
-				(close-output-pipe p)
-				(set! ##sys#standard-output old)
-				(apply values results) ) ) ) ) )
+    (let ([p (apply open-output-pipe cmd mode)])
+      (fluid-let ((##sys#standard-output p))
+	(##sys#call-with-values thunk
+				(lambda results
+				  (close-output-pipe p)
+				  (apply values results) ) ) ) ) ) )
 
 (define-foreign-variable _pipefd0 int "C_pipefds[ 0 ]")
 (define-foreign-variable _pipefd1 int "C_pipefds[ 1 ]")
@@ -922,26 +947,6 @@ EOF
     signal/tstp signal/pipe signal/xcpu signal/xfsz signal/usr1 signal/usr2
     signal/winch))
 
-(let ([oldhook ##sys#interrupt-hook]
-      [sigvector (make-vector 256 #f)] )
-  (set! signal-handler
-    (lambda (sig)
-      (##sys#check-exact sig 'signal-handler)
-      (##sys#slot sigvector sig) ) )
-  (set! set-signal-handler!
-    (lambda (sig proc)
-      (##sys#check-exact sig 'set-signal-handler!)
-      (##core#inline "C_establish_signal_handler" sig (and proc sig))
-      (vector-set! sigvector sig proc) ) )
-  (set! ##sys#interrupt-hook
-    (lambda (reason state)
-      (let ([h (##sys#slot sigvector reason)])
-        (if h
-            (begin
-              (h reason)
-              (##sys#context-switch state) )
-            (oldhook reason state) ) ) ) ) )
-
 (define set-signal-mask!
   (lambda (sigs)
     (##sys#check-list sigs 'set-signal-mask!)
@@ -977,12 +982,6 @@ EOF
   (when (fx< (##core#inline "C_sigprocmask_unblock" 0) 0)
       (posix-error #:process-error 'signal-unmask! "cannot unblock signal") )  )
 
-;;; Set SIGINT handler:
-
-(set-signal-handler!
- signal/int
- (lambda (n) (##sys#user-interrupt-hook)) )
-
 
 ;;; Getting system-, group- and user-information:
 
@@ -1285,65 +1284,6 @@ EOF
       (posix-error #:file-error 'hard-link "could not create hard link" old new) ) ) ) )
 
 
-;;; Using file-descriptors:
-
-(define-foreign-variable _stdin_fileno int "STDIN_FILENO")
-(define-foreign-variable _stdout_fileno int "STDOUT_FILENO")
-(define-foreign-variable _stderr_fileno int "STDERR_FILENO")
-
-(define fileno/stdin _stdin_fileno)
-(define fileno/stdout _stdout_fileno)
-(define fileno/stderr _stderr_fileno)
-
-(let ()
-  (define (mode inp m loc)
-    (##sys#make-c-string
-     (cond [(pair? m)
-            (let ([m (car m)])
-              (case m
-                [(###append) (if (not inp) "a" (##sys#error "invalid mode for input file" m))]
-                [else (##sys#error "invalid mode argument" m)] ) ) ]
-           [inp "r"]
-           [else "w"] )
-     loc) )
-  (define (check loc fd inp r)
-    (if (##sys#null-pointer? r)
-        (posix-error #:file-error loc "cannot open file" fd)
-        (let ([port (##sys#make-port inp ##sys#stream-port-class "(fdport)" 'stream)])
-          (##core#inline "C_set_file_ptr" port r)
-          port) ) )
-  (set! open-input-file*
-    (lambda (fd . m)
-      (##sys#check-exact fd 'open-input-file*)
-      (check 'open-input-file* fd #t (##core#inline_allocate ("C_fdopen" 2) fd (mode #t m 'open-input-file*))) ) )
-  (set! open-output-file*
-    (lambda (fd . m)
-      (##sys#check-exact fd 'open-output-file*)
-      (check 'open-output-file* fd #f (##core#inline_allocate ("C_fdopen" 2) fd (mode #f m 'open-output-file*)) ) ) ) )
-
-(define port->fileno
-  (lambda (port)
-    (##sys#check-port port 'port->fileno)
-    (cond [(eq? 'socket (##sys#slot port 7)) (##sys#tcp-port->fileno port)]
-          [(not (zero? (##sys#peek-unsigned-integer port 0)))
-           (let ([fd (##core#inline "C_C_fileno" port)])
-             (when (fx< fd 0)
-               (posix-error #:file-error 'port->fileno "cannot access file-descriptor of port" port) )
-             fd) ]
-          [else (posix-error #:type-error 'port->fileno "port has no attached file" port)] ) ) )
-
-(define duplicate-fileno
-  (lambda (old . new)
-    (##sys#check-exact old duplicate-fileno)
-    (let ([fd (if (null? new)
-                  (##core#inline "C_dup" old)
-                  (let ([n (car new)])
-                    (##sys#check-exact n 'duplicate-fileno)
-                    (##core#inline "C_dup2" old n) ) ) ] )
-      (when (fx< fd 0)
-        (posix-error #:file-error 'duplicate-fileno "cannot duplicate file-descriptor" old) )
-      fd) ) )
-
 (define ##sys#custom-input-port
   (lambda (loc nam fd #!optional (nonblocking? #f) (bufi 1) (on-close void) (more? #f))
     (when nonblocking? (##sys#file-nonblocking! fd) )
@@ -1355,7 +1295,8 @@ EOF
 	     (lambda ()
 	       (let ((res (##sys#file-select-one fd)))
 		 (if (fx= -1 res)
-		     (if (fx= _errno _ewouldblock)
+		     (if (or (fx= _errno _ewouldblock)
+			     (rx= _errno _eagain))
 			 #f
 			 (posix-error #:file-error loc "cannot select" fd nam))
 		     (fx= 1 res))))]
@@ -1369,13 +1310,15 @@ EOF
 	       (when (fx>= bufpos buflen)
 		 (let loop ()
 		   (let ([cnt (##core#inline "C_read" fd buf bufsiz)])
-		     (cond [(fx= cnt -1)
-			    (if (fx= _errno _ewouldblock)
-				(begin
-				  (##sys#thread-block-for-i/o! ##sys#current-thread fd #:input)
-				  (##sys#thread-yield!)
-				  (loop) )
-				(posix-error #:file-error loc "cannot read" fd nam) )]
+		     (cond ((fx= cnt -1)
+			    (select _errno
+			      ((_ewouldblock _eagain)
+			       (##sys#thread-block-for-i/o! ##sys#current-thread fd #:input)
+			       (##sys#thread-yield!)
+			       (loop) )
+			      ((_eintr)
+			       (##sys#dispatch-interrupt loop))
+			      (else (posix-error #:file-error loc "cannot read" fd nam) )))
 			   [(and more? (fx= cnt 0))
 					; When "more" keep trying, otherwise read once more
 					; to guard against race conditions
@@ -1385,7 +1328,8 @@ EOF
 				  (loop) )
 				(let ([cnt (##core#inline "C_read" fd buf bufsiz)])
 				  (when (fx= cnt -1)
-				    (if (fx= _errno _ewouldblock)
+				    (if (or (fx= _errno _ewouldblock)
+					    (fx= _errno _eagain))
 					(set! cnt 0)
 					(posix-error #:file-error loc "cannot read" fd nam) ) )
 				  (set! buflen cnt)
@@ -1476,18 +1420,21 @@ EOF
 (define ##sys#custom-output-port
   (lambda (loc nam fd #!optional (nonblocking? #f) (bufi 0) (on-close void))
     (when nonblocking? (##sys#file-nonblocking! fd) )
-    (letrec (
-	     [poke
+    (letrec ([poke
 	      (lambda (str len)
-		(let ([cnt (##core#inline "C_write" fd str len)])
-		  (cond [(fx= -1 cnt)
-			 (if (fx= _errno _ewouldblock)
-			     (begin
-			       (##sys#thread-yield!)
-			       (poke str len) )
-			     (posix-error loc #:file-error "cannot write" fd nam) ) ]
-			[(fx< cnt len)
-			 (poke (##sys#substring str cnt len) (fx- len cnt)) ] ) ) )]
+		(let loop ()
+		  (let ([cnt (##core#inline "C_write" fd str len)])
+		    (cond ((fx= -1 cnt)
+			   (select _errno
+			     ((_ewouldblock _eagain)
+			      (##sys#thread-yield!)
+			      (poke str len) )
+			     ((_eintr)
+			      (##sys#dispatch-interrupt loop))
+			     (else
+			      (posix-error loc #:file-error "cannot write" fd nam) ) ) )
+			  ((fx< cnt len)
+			   (poke (##sys#substring str cnt len) (fx- len cnt)) ) ) ) ))]
 	     [store
 	      (let ([bufsiz (if (fixnum? bufi) bufi (##sys#size bufi))])
 		(if (fx= 0 bufsiz)
@@ -1511,8 +1458,7 @@ EOF
 				     (set! bufpos (fx+ bufpos len))] ) )
 			    (when (fx< 0 bufpos)
 			      (poke buf bufpos) ) ) ) ) ) )])
-      (letrec (
-	       [this-port
+      (letrec ([this-port
 		(make-output-port
 		 (lambda (str)		; write-string
 		   (store str) )
@@ -1597,10 +1543,16 @@ EOF
 (define fifo?
   (lambda (filename)
     (##sys#check-string filename 'fifo?)
-    (let ([v (##sys#file-info (##sys#expand-home-path filename))])
-      (if v
-          (fx= 3 (##sys#slot v 4))
-          (posix-error #:file-error 'fifo? "file does not exist" filename) ) ) ) )
+    (case (##core#inline 
+	   "C_i_fifo_p"
+	   (##sys#make-c-string (##sys#expand-home-path filename) 'fifo?))
+      ((#t) #t)
+      ((#f) #f)
+      ((0) (##sys#signal-hook #:file-error 'fifo? "file does not exist" filename) )
+      (else
+       (posix-error 
+	#:file-error 'fifo?
+	"system error while trying to access file" filename) ) ) ) )
 
 
 ;;; Environment access:
@@ -1745,9 +1697,9 @@ EOF
     (##sys#check-port port 'set-buffering-mode!)
     (let ([size (if (pair? size) (car size) _bufsiz)]
 	  [mode (case mode
-		  [(###full) _iofbf]
-		  [(###line) _iolbf]
-		  [(###none) _ionbf]
+		  [(#:full) _iofbf]
+		  [(#:line) _iolbf]
+		  [(#:none) _ionbf]
 		  [else (##sys#error 'set-buffering-mode! "invalid buffering-mode" mode port)] ) ] )
       (##sys#check-exact size 'set-buffering-mode!)
       (when (fx< (if (eq? 'stream (##sys#slot port 7))
@@ -1757,12 +1709,12 @@ EOF
 	(##sys#error 'set-buffering-mode! "cannot set buffering mode" port mode size) ) ) ) )
 
 (define (terminal-port? port)
-  (##sys#check-port port 'terminal-port?)
+  (##sys#check-open-port port 'terminal-port?)
   (let ([fp (##sys#peek-unsigned-integer port 0)])
     (and (not (eq? 0 fp)) (##core#inline "C_tty_portp" port) ) ) )
 
 (define (##sys#terminal-check caller port)
-  (##sys#check-port port caller)
+  (##sys#check-open-port port caller)
   (unless (and (eq? 'stream (##sys#slot port 7))
 	       (##core#inline "C_tty_portp" port))
 	  (##sys#error caller "port is not connected to a terminal" port)))
@@ -1814,9 +1766,12 @@ EOF
             [else pid] ) ) ) ) )
 
 (define process-execute
-  (let ([setarg (foreign-lambda void "C_set_exec_arg" int scheme-pointer int)]
+  ;; NOTE: We use c-string here instead of scheme-object.
+  ;; Because set_exec_* make a copy, this implies a double copy.
+  ;; At least it's secure, we can worry about performance later, if at all
+  (let ([setarg (foreign-lambda void "C_set_exec_arg" int c-string int)]
         [freeargs (foreign-lambda void "C_free_exec_args")]
-        [setenv (foreign-lambda void "C_set_exec_env" int scheme-pointer int)]
+        [setenv (foreign-lambda void "C_set_exec_env" int c-string int)]
         [freeenv (foreign-lambda void "C_free_exec_env")]
         [pathname-strip-directory pathname-strip-directory] )
     (lambda (filename #!optional (arglist '()) envlist)
@@ -1862,17 +1817,6 @@ EOF
               (##core#inline "C_WTERMSIG" _wait-status)]
             [else (##core#inline "C_WSTOPSIG" _wait-status)] ) ) ) )
 
-(define process-wait
-  (lambda args
-    (let-optionals* args ([pid #f] [nohang #f])
-      (let ([pid (or pid -1)])
-        (##sys#check-exact pid 'process-wait)
-        (receive [epid enorm ecode] (##sys#process-wait pid nohang)
-          (if (fx= epid -1)
-              (posix-error #:process-error 'process-wait "waiting for child process failed" pid)
-              (values epid enorm ecode) ) ) ) ) ) )
-
-(define current-process-id (foreign-lambda int "C_getpid"))
 (define parent-process-id (foreign-lambda int "C_getppid"))
 
 (define sleep (foreign-lambda int "C_sleep" int))
@@ -1999,8 +1943,8 @@ EOF
 (define process)
 (define process*)
 
-(let ([%process
-        (lambda (loc err? cmd args env)
+(let ((%process
+        (lambda (loc err? cmd args env k)
           (let ([chkstrlst
                  (lambda (lst)
                    (##sys#check-list lst loc)
@@ -2012,16 +1956,19 @@ EOF
                   (set! args (##sys#shell-command-arguments cmd))
                   (set! cmd (##sys#shell-command)) ) )
             (when env (chkstrlst env))
-            (receive [in out pid err] (##sys#process loc cmd args env #t #t err?)
-              (if err?
-                  (values in out pid err)
-                  (values in out pid) ) ) ) )] )
+            (##sys#call-with-values 
+	     (lambda () (##sys#process loc cmd args env #t #t err?))
+	     k)))))
   (set! process
     (lambda (cmd #!optional args env)
-      (%process 'process #f cmd args env) ))
+      (%process 
+       'process #f cmd args env
+       (lambda (i o p e) (values i o p)))))
   (set! process*
     (lambda (cmd #!optional args env)
-      (%process 'process* #t cmd args env) )) )
+      (%process
+       'process* #t cmd args env
+       values))))
 
 
 ;;; chroot:
