@@ -1,6 +1,6 @@
 ;;;; posixwin.scm - Miscellaneous file- and process-handling routines, available on Windows
 ;
-; Copyright (c) 2008-2011, The Chicken Team
+; Copyright (c) 2008-2012, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
@@ -34,6 +34,7 @@
 ; set-signal-mask!  signal-mask	 signal-masked?	 signal-mask!  signal-unmask!
 ; user-information group-information  get-groups  set-groups!  initialize-groups
 ; errno/wouldblock
+; change-directory*
 ; change-file-owner
 ; current-user-id  current-group-id  current-effective-user-id	current-effective-group-id
 ; current-effective-user-name
@@ -222,14 +223,7 @@ readdir(DIR * dir)
 #define open_text_output_pipe(a, n, name)    open_binary_output_pipe(a, n, name)
 #define close_pipe(p)			     C_fix(_pclose(C_port_file(p)))
 
-#define C_set_file_ptr(port, ptr)  (C_set_block_item(port, 0, (C_block_item(ptr, 0))), C_SCHEME_UNDEFINED)
-
-#define C_getpid	    getpid
 #define C_chmod(fn, m)	    C_fix(chmod(C_data_pointer(fn), C_unfix(m)))
-#define C_fdopen(a, n, fd, m) C_mpointer(a, fdopen(C_unfix(fd), C_c_string(m)))
-#define C_C_fileno(p)	    C_fix(fileno(C_port_file(p)))
-#define C_dup(x)	    C_fix(dup(C_unfix(x)))
-#define C_dup2(x, y)	    C_fix(dup2(C_unfix(x), C_unfix(y)))
 #define C_setvbuf(p, m, s)  C_fix(setvbuf(C_port_file(p), NULL, C_unfix(m), C_unfix(s)))
 #define C_test_access(fn, m)	    C_fix(access((char *)C_data_pointer(fn), C_unfix(m)))
 #define C_pipe(d, m)	    C_fix(_pipe(C_pipefds, PIPE_BUF, C_unfix(m)))
@@ -268,6 +262,7 @@ C_set_arg_string(char **where, int i, char *dat, int len)
 	ptr = (char *)C_malloc(len + 1);
 	C_memcpy(ptr, dat, len);
 	ptr[ len ] = '\0';
+        /* Can't barf() here, so the NUL byte check happens in Scheme */
     }
     else
 	ptr = NULL;
@@ -836,6 +831,7 @@ C_process(const char * app, const char * cmdlin, const char ** env,
 		pb += strlen(*p) + 1;
 	    }
 	    *pb = '\0';
+            /* This _should_ already have been checked for embedded NUL bytes */
 	}
 	else
 	    success = FALSE;
@@ -1088,15 +1084,8 @@ EOF
     (##sys#signal-hook #:file-error 'create-directory
 		       "cannot create directory" name)))
 
-(define-inline (create-directory-check name)
-  (if (file-exists? name)
-      (let ((i   (##sys#file-info name)))
-	(and i
-	     (fx= 1 (##sys#slot i 4))))
-      #f))
-
 (define-inline (create-directory-helper-silent name)
-  (unless (create-directory-check name)
+  (unless (##sys#file-exists? name #f #t #f)
     (create-directory-helper name)))
 
 (define-inline (create-directory-helper-parents name)
@@ -1162,13 +1151,20 @@ EOF
 	   (else (badmode m)) ) ) ) ) )
   (set! close-input-pipe
     (lambda (port)
-      (##sys#check-port port 'close-input-pipe)
+      (##sys#check-input-port port #t 'close-input-pipe)
       (let ((r (##core#inline "close_pipe" port)))
 	(##sys#update-errno)
 	(when (eq? -1 r)
 	  (##sys#signal-hook #:file-error 'close-input-pipe "error while closing pipe" port) )
 	r)))
-  (set! close-output-pipe close-input-pipe) )
+  (set! close-output-pipe
+    (lambda (port)
+      (##sys#check-output-port port #t 'close-output-pipe)
+      (let ((r (##core#inline "close_pipe" port)))
+	(##sys#update-errno)
+	(when (eq? -1 r)
+	  (##sys#signal-hook #:file-error 'close-output-pipe "error while closing pipe" port) )
+	r))))
 
 (define call-with-input-pipe
   (lambda (cmd proc . mode)
@@ -1190,27 +1186,23 @@ EOF
 
 (define with-input-from-pipe
   (lambda (cmd thunk . mode)
-    (let ([old ##sys#standard-input]
-	  [p (apply open-input-pipe cmd mode)] )
-      (set! ##sys#standard-input p)
-      (##sys#call-with-values
-       thunk
-       (lambda results
-	 (close-input-pipe p)
-	 (set! ##sys#standard-input old)
-	 (apply values results) ) ) ) ) )
+    (let ([p (apply open-input-pipe cmd mode)])
+      (fluid-let ((##sys#standard-input p))
+	(##sys#call-with-values
+	 thunk
+	 (lambda results
+	   (close-input-pipe p)
+	   (apply values results) ) ) ) ) ) )
 
 (define with-output-to-pipe
   (lambda (cmd thunk . mode)
-    (let ([old ##sys#standard-output]
-	  [p (apply open-output-pipe cmd mode)] )
-      (set! ##sys#standard-output p)
-      (##sys#call-with-values
-       thunk
-       (lambda results
-	 (close-output-pipe p)
-	 (set! ##sys#standard-output old)
-	 (apply values results) ) ) ) ) )
+    (let ([p (apply open-output-pipe cmd mode)])
+      (fluid-let ((##sys#standard-output p))
+	(##sys#call-with-values
+	 thunk
+	 (lambda results
+	   (close-output-pipe p)
+	   (apply values results) ) ) ) ) ) )
 
 
 ;;; Pipe primitive:
@@ -1268,25 +1260,6 @@ EOF
     signal/term signal/int signal/fpe signal/ill
     signal/segv signal/abrt signal/break))
 
-(let ([oldhook ##sys#interrupt-hook]
-      [sigvector (make-vector 256 #f)] )
-  (set! signal-handler
-    (lambda (sig)
-      (##sys#check-exact sig 'signal-handler)
-      (##sys#slot sigvector sig) ) )
-  (set! set-signal-handler!
-    (lambda (sig proc)
-      (##sys#check-exact sig 'set-signal-handler!)
-      (##core#inline "C_establish_signal_handler" sig (and proc sig))
-      (vector-set! sigvector sig proc) ) )
-  (set! ##sys#interrupt-hook
-    (lambda (reason state)
-      (let ([h (##sys#slot sigvector reason)])
-	(if h
-	    (begin
-	      (h reason)
-	      (##sys#context-switch state) )
-	    (oldhook reason state) ) ) ) ) )
 
 ;;; More errno codes:
 
@@ -1395,7 +1368,7 @@ EOF
 
 (define port->fileno
   (lambda (port)
-    (##sys#check-port port 'port->fileno)
+    (##sys#check-open-port port 'port->fileno)
     (if (not (zero? (##sys#peek-unsigned-integer port 0)))
 	(let ([fd (##core#inline "C_C_fileno" port)])
 	  (when (fx< fd 0)
@@ -1461,7 +1434,7 @@ EOF
       (ex0 (if (pair? code) (car code) 0)) ) ) )
 
 (define (terminal-port? port)
-  (##sys#check-port port 'terminal-port?)
+  (##sys#check-open-port port 'terminal-port?)
   (let ([fp (##sys#peek-unsigned-integer port 0)])
     (and (not (eq? 0 fp)) (##core#inline "C_tty_portp" port) ) ) )
 
@@ -1477,7 +1450,7 @@ EOF
 
 (define set-buffering-mode!
   (lambda (port mode . size)
-    (##sys#check-port port 'set-buffering-mode!)
+    (##sys#check-open-port port 'set-buffering-mode!)
     (let ([size (if (pair? size) (car size) _bufsiz)]
 	  [mode (case mode
 		  [(###full) _iofbf]
@@ -1524,7 +1497,7 @@ EOF
 		      [else (loop (fx+ i 1))]))))])
 	  (let loop ([ilst lst] [olst '()])
 	    (if (null? ilst)
-		(reverse olst)
+		(##sys#fast-reverse olst)
 		(let ([str (car ilst)])
 		  (loop
 		   (cdr ilst)
@@ -1533,8 +1506,11 @@ EOF
 		    olst)) ) ) ) ) ) ) )
 
 (define $exec-setup
-  (let ([setarg (foreign-lambda void "C_set_exec_arg" int scheme-pointer int)]
-	[setenv (foreign-lambda void "C_set_exec_env" int scheme-pointer int)]
+  ;; NOTE: We use c-string here instead of scheme-object.
+  ;; Because set_exec_* make a copy, this implies a double copy.
+  ;; At least it's secure, we can worry about performance later, if at all
+  (let ([setarg (foreign-lambda void "C_set_exec_arg" int c-string int)]
+	[setenv (foreign-lambda void "C_set_exec_env" int c-string int)]
 	[build-exec-argvec
 	  (lambda (loc lst argvec-setter idx)
 	    (if lst
@@ -1574,8 +1550,6 @@ EOF
     ($exec-teardown 'process-spawn "cannot spawn process" filename
       (if envlst (##core#inline "C_spawnvpe" mode prg) (##core#inline "C_spawnvp" mode prg))) ) )
 
-(define current-process-id (foreign-lambda int "C_getpid"))
-
 (define-foreign-variable _shlcmd c-string "C_shlcmd")
 
 (define (##sys#shell-command)
@@ -1615,6 +1589,7 @@ EOF
 ; where stdin-input-port?, etc. is a port or #f, indicating no port created.
 
 (define ##sys#process
+  ;; XXX TODO: When environment is implemented, check for embedded NUL bytes!
   (let ([c-process
 	  (foreign-lambda bool "C_process" c-string c-string c-pointer
 	    (c-pointer int) (c-pointer int) (c-pointer int) (c-pointer int) int)])
@@ -1675,21 +1650,7 @@ EOF
     (values pid #t _exstatus)
     (values -1 #f #f) ) )
 
-(define process-wait
-  (lambda (pid . args)
-    (let-optionals* args ([nohang #f])
-      (##sys#check-exact pid 'process-wait)
-      (receive [epid enorm ecode] (##sys#process-wait pid nohang)
-	(if (fx= epid -1)
-	  (begin
-	    (##sys#update-errno)
-	    (##sys#signal-hook #:process-error 'process-wait "waiting for child process failed" pid) )
-	  (values epid enorm ecode) ) ) ) ) )
-
-(define sleep
-  (lambda (t)
-    (##core#inline "C_sleep" t)
-    0) )
+(define sleep (foreign-lambda int "C_sleep" int))
 
 (define-foreign-variable _hostname c-string "C_hostname")
 (define-foreign-variable _osver c-string "C_osver")
@@ -1731,6 +1692,7 @@ EOF
      (define (?name . _)
        (error '?name (##core#immutable '"this function is not available on this platform")) ) ] ) )
 
+(define-unimplemented change-directory*)
 (define-unimplemented change-file-owner)
 (define-unimplemented create-fifo)
 (define-unimplemented create-session)
