@@ -1,6 +1,6 @@
 ; scheduler.scm - Basic scheduler for multithreading
 ;
-; Copyright (c) 2008-2011, The Chicken Team
+; Copyright (c) 2008-2012, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
@@ -163,7 +163,7 @@ EOF
 	(let ([nt (remove-from-ready-queue)])
 	  (cond [(not nt) 
 		 (if (and (null? ##sys#timeout-list) (null? ##sys#fd-list))
-		     (##sys#halt "deadlock")
+		     (panic "deadlock")
 		     (loop1) ) ]
 		[(eq? (##sys#slot nt 3) 'ready) (switch nt)]
 		[else (loop2)] ) ) ) ) ) )
@@ -309,35 +309,24 @@ EOF
   (##sys#setislot t 4 #f)
   (##sys#add-to-ready-queue t) )
 
-(define ##sys#default-exception-handler
-  (let ([print-error-message print-error-message]
-	[display display]
-	[print-call-chain print-call-chain]
-	[open-output-string open-output-string]
-	[get-output-string get-output-string] )
-    (lambda (arg)
-      (let ([ct ##sys#current-thread])
-	(dbg "exception: " ct " -> " 
-	     (if (##sys#structure? arg 'condition) (##sys#slot arg 2) arg))
-	(cond [(foreign-value "C_abort_on_thread_exceptions" bool)
-	       (let* ([pt ##sys#primordial-thread]
-		      [ptx (##sys#slot pt 1)] )
-		 (##sys#setslot 
-		  pt 1 
-		  (lambda ()
-		    (##sys#signal arg)
-		    (ptx) ) )
-		 (##sys#thread-unblock! pt) ) ]
-	      [##sys#warnings-enabled
-	       (let ([o (open-output-string)])
-		 (display "Warning (" o)
-		 (display ct o)
-		 (display ")" o)
-		 (print-error-message arg ##sys#standard-error (get-output-string o))
-		 (print-call-chain ##sys#standard-error 0 ct) ) ] )
-	(##sys#setslot ct 7 arg)
-	(##sys#thread-kill! ct 'terminated)
-	(##sys#schedule) ) ) ) )
+(define (##sys#default-exception-handler arg)
+  (let ([ct ##sys#current-thread])
+    (dbg "exception: " ct " -> " 
+	 (if (##sys#structure? arg 'condition) (##sys#slot arg 2) arg))
+    (cond ((foreign-value "C_abort_on_thread_exceptions" bool)
+	   (let* ([pt ##sys#primordial-thread]
+		  [ptx (##sys#slot pt 1)] )
+	     (##sys#setslot 
+	      pt 1 
+	      (lambda ()
+		(##sys#signal arg)
+		(ptx) ) )
+	     (##sys#thread-unblock! pt) ) )
+	  (else
+	   (##sys#show-exception-warning arg "in thread" ct)))
+    (##sys#setslot ct 7 arg)
+    (##sys#thread-kill! ct 'terminated)
+    (##sys#schedule) ) )
 
 
 ;;; `select()'-based blocking:
@@ -366,7 +355,7 @@ EOF
 (define fdset-clear
   (foreign-lambda* void ()
     "FD_ZERO(&C_fdset_input);"
-    "FD_ZERO(&C_fdset_output);") )
+    "FD_ZERO(&C_fdset_output);"))
 
 (define fdset-input-set
   (foreign-lambda* void ([int fd])
@@ -434,7 +423,7 @@ EOF
 		     (let* ([a (car lst)]
 			    [fd (car a)]
 			    [inf (##core#inline "C_fd_test_input" fd)]
-			    [outf (##core#inline "C_fd_test_output" fd)] )
+			    [outf (##core#inline "C_fd_test_output" fd)])
 		       (dbg "fd " fd " state: input=" inf ", output=" outf)
 		       (if (or inf outf)
 			   (let loop2 ((threads (cdr a)) (keep '()))
@@ -455,13 +444,9 @@ EOF
 					    (##sys#remove-from-timeout-list t))
 					  (##sys#thread-basic-unblock! t) 
 					  (loop2 (cdr threads) keep))
-					 ((or (not (eq? fd (car p)))
-					      ;; thread on fd-list has incorrect
-					      ;; file-descriptor registered. 
-					      ;; We just assume this is the right one and
-					      ;; unblock.
-					      ;; XXX Needs to be investigated...
-					      (fdset-test inf outf (cdr p)))
+					 ((not (eq? fd (car p)))
+					  (panic "thread is registered for I/O on unknown file-descriptor"))
+					 ((fdset-test inf outf (cdr p))
 					  (when (##sys#slot t 4) ; also blocked for timeout?
 					    (##sys#remove-from-timeout-list t))
 					  (##sys#thread-basic-unblock! t) 
