@@ -1,6 +1,6 @@
 ;;;; setup-api.scm - build + installation API for eggs
 ;
-; Copyright (c) 2008-2011, The Chicken Team
+; Copyright (c) 2008-2012, The Chicken Team
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -39,7 +39,7 @@
     ((run execute)
      compile
      standard-extension
-     make make/proc
+     make make/proc			; DEPRECATED
      host-extension
      install-extension install-program install-script
      setup-verbose-mode setup-install-mode deployment-mode
@@ -290,6 +290,9 @@
 
 ;;; "make" functionality
 
+;;; DEPRECATED
+;;; vvv
+
 (define (make:find-matching-line str spec)
   (let ((match? (lambda (s) (string=? s str))))
     (let loop ((lines spec))
@@ -381,8 +384,8 @@
 					 ((string? reason)
 					  (string-append " because " reason " changed"))
 					 (else
-					  (string-append (sprintf " just because (reason: ~a date: ~a)" 
-								  reason date)))) ) )
+					  (sprintf " just because (reason: ~a date: ~a)" 
+					    reason date)))) )
 			      (handle-exceptions exn
 				  (begin
 				    (printf "make: Failed to make ~a: ~a~%"
@@ -443,6 +446,9 @@
 		'('())
 		(cddr form)))))))
 
+;;;^^^
+;;; DEPRECATED
+
 
 ;;; Processing setup scripts
 
@@ -495,8 +501,17 @@
 		      to-path
 		      (make-pathname prefix to-path) )
 		  to-path))))
-    (ensure-directory to)
-    (run (,*copy-command* ,(shellpath from) ,(shellpath to)))
+    (let walk ((from from) (to to))
+      (cond ((directory? from)
+	     (for-each
+	      (lambda (f)
+		(walk (make-pathname from f) (make-pathname to f)))
+	      (directory from)))
+	    (else
+	     (ensure-directory to)
+	     (run (,*copy-command* 
+		   ,(shellpath from)
+		   ,(shellpath to))))))
     to))
 
 (define (path-prefix? pref path)
@@ -544,9 +559,17 @@
 	    "unknown"))))
 
 (define (supply-version info version)
-  (if (assq 'version info)
-      info
-      (cons `(version ,(what-version version)) info)))
+  (cond ((assq 'version info) => 
+	 (lambda (a)
+	   (cons 
+	    `(egg-name ,(extension-name))
+	    info)))
+	(else
+	 (let ((v (what-version version)))
+	   (cons*
+	    `(version ,v)
+	    `(egg-name ,(extension-name))
+	    info)))))
 
 
 ;;; Convenience function
@@ -601,7 +624,7 @@
   (when (setup-install-mode)
     (let* ((files (check-filelist (if (list? files) files (list files))))
 	   (pre (installation-prefix))
-	   (ppath (ensure-directory (make-pathname pre "bin")))
+	   (ppath (ensure-directory (make-pathname pre "bin") #t))
 	   (files (if *windows*
                       (map (lambda (f)
                              (if (list? f) 
@@ -623,7 +646,7 @@
   (when (setup-install-mode)
     (let* ((files (check-filelist (if (list? files) files (list files))))
 	   (pre (installation-prefix))
-	   (ppath (ensure-directory (make-pathname pre "bin")))
+	   (ppath (ensure-directory (make-pathname pre "bin") #t))
 	   (pfiles (map (lambda (f)
 			  (let ((from (if (pair? f) (car f) f))
 				(to (make-dest-pathname ppath f)) )
@@ -650,11 +673,11 @@
 			  (sprintf "lib/chicken/~a" (##sys#fudge 42)))
 			 (repository-path)))) ; otherwise use repo-path
 	       (repository-path))) )
-    (ensure-directory p)
+    (ensure-directory p #t)
     p) )
 
-(define (ensure-directory path)
-  (and-let* ((dir (pathname-directory path)))
+(define (ensure-directory path #!optional full)
+  (and-let* ((dir (if full path (pathname-directory path))))
     (if (file-exists? dir)
 	(unless (directory? dir)
 	  (error "cannot create directory: a file with the same name already exists") )
@@ -763,8 +786,7 @@
                    [ensure-string (lambda (x) (if (or (not x) (null? x)) "" (->string x)))])
                (list (ensure-string nam) (ensure-string ver)) ) ]
             [else
-             (warning "invalid extension-name-and-version" x)
-             (extension-name-and-version) ] ) ) ) )
+             (error "invalid extension-name-and-version" x)]))))
 
 (define (extension-name)
   (car (extension-name-and-version)) )
