@@ -1,6 +1,6 @@
 ;;;; srfi-4.scm - Homogeneous numeric vectors
 ;
-; Copyright (c) 2008-2011, The Chicken Team
+; Copyright (c) 2008-2012, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
@@ -217,7 +217,7 @@ EOF
      (##sys#check-structure x 'u32vector 'u32vector-ref)
      (let ((len (##core#inline "C_u_i_u32vector_length" x)))
        (check-range i 0 len 'u32vector-ref)
-       (##core#inline_allocate ("C_a_i_u32vector_ref" 4) x i)))
+       (##core#inline_allocate ("C_a_u_i_u32vector_ref" 4) x i)))
    u32vector-set!
    "(u32vector-ref v i)"))
 
@@ -227,7 +227,7 @@ EOF
      (##sys#check-structure x 's32vector 's32vector-ref)
      (let ((len (##core#inline "C_u_i_s32vector_length" x)))
        (check-range i 0 len 's32vector-ref)
-       (##core#inline_allocate ("C_a_i_s32vector_ref" 4) x i)))
+       (##core#inline_allocate ("C_a_u_i_s32vector_ref" 4) x i)))
    s32vector-set!
    "(s32vector-ref v i)"))
 
@@ -237,7 +237,7 @@ EOF
      (##sys#check-structure x 'f32vector 'f32vector-ref)
      (let ((len (##core#inline "C_u_i_f32vector_length" x)))
        (check-range i 0 len 'f32vector-ref)
-       (##core#inline_allocate ("C_a_i_f32vector_ref" 4) x i)))
+       (##core#inline_allocate ("C_a_u_i_f32vector_ref" 4) x i)))
    f32vector-set!
    "(f32vector-ref v i)"))
 
@@ -245,9 +245,9 @@ EOF
   (getter-with-setter
    (lambda (x i)
      (##sys#check-structure x 'f64vector 'f64vector-ref)
-     (let ((len (##core#inline "C_u_i_8vector_length" x)))
+     (let ((len (##core#inline "C_u_i_f64vector_length" x)))
        (check-range i 0 len 'f64vector-ref)
-       (##core#inline_allocate ("C_a_i_f64vector_ref" 4) x i)))
+       (##core#inline_allocate ("C_a_u_i_f64vector_ref" 4) x i)))
    f64vector-set!
    "(f64vector-ref v i)"))
 
@@ -394,33 +394,35 @@ EOF
 
 ;;; Creating vectors from a list:
 
-(define-syntax (list->XXXvector x r c)
-  (let* ((tag (##sys#strip-syntax (cadr x)))
-	 (tagstr (symbol->string tag))
-	 (name (string->symbol (string-append "list->" tagstr)))
-	 (make (string->symbol (string-append "make-" tagstr)))
-	 (set (string->symbol (string-append tagstr "-set!"))))
-    `(define ,name
-       (let ((,make ,make))
-	 (lambda (lst)
-	   (##sys#check-list lst ',tag)
-	   (let* ((n (##core#inline "C_i_length" lst))
-		  (v (,make n)) )
-	     (do ((p lst (##core#inline "C_slot" p 1))
-		  (i 0 (##core#inline "C_fixnum_plus" i 1)) )
-		 ((##core#inline "C_eqp" p '()) v)
-	       (if (and (##core#inline "C_blockp" p) (##core#inline "C_pairp" p))
-		   (,set v i (##core#inline "C_slot" p 0))
-		   (##sys#error-not-a-proper-list lst) ) ) ) )))))
+(define-syntax list->NNNvector 
+  (er-macro-transformer 
+   (lambda (x r c)
+     (let* ((tag (##sys#strip-syntax (cadr x)))
+	    (tagstr (symbol->string tag))
+	    (name (string->symbol (string-append "list->" tagstr)))
+	    (make (string->symbol (string-append "make-" tagstr)))
+	    (set (string->symbol (string-append tagstr "-set!"))))
+       `(define ,name
+	  (let ((,make ,make))
+	    (lambda (lst)
+	      (##sys#check-list lst ',tag)
+	      (let* ((n (##core#inline "C_i_length" lst))
+		     (v (,make n)) )
+		(do ((p lst (##core#inline "C_slot" p 1))
+		     (i 0 (##core#inline "C_fixnum_plus" i 1)) )
+		    ((##core#inline "C_eqp" p '()) v)
+		  (if (and (##core#inline "C_blockp" p) (##core#inline "C_pairp" p))
+		      (,set v i (##core#inline "C_slot" p 0))
+		      (##sys#error-not-a-proper-list lst) ) ) ) )))))))
 
-(list->XXXvector u8vector)
-(list->XXXvector s8vector)
-(list->XXXvector u16vector)
-(list->XXXvector s16vector)
-(list->XXXvector u32vector)
-(list->XXXvector s32vector)
-(list->XXXvector f32vector)
-(list->XXXvector f64vector)
+(list->NNNvector u8vector)
+(list->NNNvector s8vector)
+(list->NNNvector u16vector)
+(list->NNNvector s16vector)
+(list->NNNvector u32vector)
+(list->NNNvector s32vector)
+(list->NNNvector f32vector)
+(list->NNNvector f64vector)
 
 
 ;;; More constructors:
@@ -452,30 +454,32 @@ EOF
 
 ;;; Creating lists from a vector:
 
-(define-syntax (XXXvector->list x r c)
-  (let* ((tag (##sys#strip-syntax (cadr x)))
-	 (alloc? (pair? (cddr x)))
-	 (name (string->symbol (string-append (symbol->string tag) "->list"))))
-    `(define (,name v)
-       (##sys#check-structure v ',tag ',name)
-       (let ((len (##core#inline ,(conc "C_u_i_" tag "_length") v)))
-	 (let loop ((i 0))
-	   (if (fx>= i len)
-	       '()
-	       (cons 
-		,(if alloc?
-		     `(##core#inline_allocate (,(conc "C_a_i_" tag "_ref") 4) v i)
-		     `(##core#inline ,(conc "C_u_i_" tag "_ref") v i))
-		(loop (fx+ i 1)) ) ) ) ) ) ) )
+(define-syntax NNNvector->list
+  (er-macro-transformer
+   (lambda (x r c)
+     (let* ((tag (##sys#strip-syntax (cadr x)))
+	    (alloc? (pair? (cddr x)))
+	    (name (string->symbol (string-append (symbol->string tag) "->list"))))
+       `(define (,name v)
+	  (##sys#check-structure v ',tag ',name)
+	  (let ((len (##core#inline ,(conc "C_u_i_" tag "_length") v)))
+	    (let loop ((i 0))
+	      (if (fx>= i len)
+		  '()
+		  (cons 
+		   ,(if alloc?
+			`(##core#inline_allocate (,(conc "C_a_u_i_" tag "_ref") 4) v i)
+			`(##core#inline ,(conc "C_u_i_" tag "_ref") v i))
+		   (loop (fx+ i 1)) ) ) ) ) ) ) )))
 
-(XXXvector->list u8vector)
-(XXXvector->list s8vector)
-(XXXvector->list u16vector)
-(XXXvector->list s16vector)
-(XXXvector->list u32vector #t)
-(XXXvector->list s32vector #t)
-(XXXvector->list f32vector #t)
-(XXXvector->list f64vector #t)
+(NNNvector->list u8vector)
+(NNNvector->list s8vector)
+(NNNvector->list u16vector)
+(NNNvector->list s16vector)
+(NNNvector->list u32vector #t)
+(NNNvector->list s32vector #t)
+(NNNvector->list f32vector #t)
+(NNNvector->list f64vector #t)
 
 
 ;;; Predicates:
@@ -610,9 +614,9 @@ EOF
 
 ;;; Subvectors:
 
-(declare (hide subvector))
+(declare (hide subnvector))
 
-(define (subvector v t es from to loc)
+(define (subnvector v t es from to loc)
   (##sys#check-structure v t loc)
   (let* ([bv (##sys#slot v 1)]
 	 [len (##sys#size bv)]
@@ -626,19 +630,19 @@ EOF
 	(##core#inline "C_copy_subvector" bv2 bv 0 (fx* from es) size2)
 	v) ) ) )
 
-(define (subu8vector v from to) (subvector v 'u8vector 1 from to 'subu8vector))
-(define (subu16vector v from to) (subvector v 'u16vector 2 from to 'subu16vector))
-(define (subu32vector v from to) (subvector v 'u32vector 4 from to 'subu32vector))
-(define (subs8vector v from to) (subvector v 's8vector 1 from to 'subs8vector))
-(define (subs16vector v from to) (subvector v 's16vector 2 from to 'subs16vector))
-(define (subs32vector v from to) (subvector v 's32vector 4 from to 'subs32vector))
-(define (subf32vector v from to) (subvector v 'f32vector 4 from to 'subf32vector))
-(define (subf64vector v from to) (subvector v 'f64vector 8 from to 'subf64vector))
+(define (subu8vector v from to) (subnvector v 'u8vector 1 from to 'subu8vector))
+(define (subu16vector v from to) (subnvector v 'u16vector 2 from to 'subu16vector))
+(define (subu32vector v from to) (subnvector v 'u32vector 4 from to 'subu32vector))
+(define (subs8vector v from to) (subnvector v 's8vector 1 from to 'subs8vector))
+(define (subs16vector v from to) (subnvector v 's16vector 2 from to 'subs16vector))
+(define (subs32vector v from to) (subnvector v 's32vector 4 from to 'subs32vector))
+(define (subf32vector v from to) (subnvector v 'f32vector 4 from to 'subf32vector))
+(define (subf64vector v from to) (subnvector v 'f64vector 8 from to 'subf64vector))
 
 (define (write-u8vector v #!optional (port ##sys#standard-output) (from 0)
 			(to (u8vector-length v)))
   (##sys#check-structure v 'u8vector 'write-u8vector)
-  (##sys#check-port port 'write-u8vector)
+  (##sys#check-output-port port #t 'write-u8vector)
   (do ((i from (fx+ i 1)))
       ((fx>= i to))
     (##sys#write-char-0 
@@ -646,7 +650,7 @@ EOF
      port) ) )
 
 (define (read-u8vector! n dest #!optional (port ##sys#standard-input) (start 0))
-  (##sys#check-port port 'read-u8vector!)
+  (##sys#check-input-port port #t 'read-u8vector!)
   (##sys#check-exact start 'read-u8vector!)
   (##sys#check-structure dest 'u8vector 'read-u8vector!)
   (let ((dest (##sys#slot dest 1)))
@@ -666,7 +670,7 @@ EOF
 	 (##core#inline "C_substring_copy" str str2 0 n 0)
 	 str2) ) )
     (lambda (#!optional n (p ##sys#standard-input))
-      (##sys#check-port p 'read-u8vector)
+      (##sys#check-input-port p #t 'read-u8vector)
       (cond (n (##sys#check-exact n 'read-u8vector)
 	       (let* ((str (##sys#allocate-vector n #t #f #t))
 		      (n2 (##sys#read-string! n str p 0)) )
