@@ -1,6 +1,6 @@
 ;;;; support.scm - Miscellaneous support code for the CHICKEN compiler
 ;
-; Copyright (c) 2008-2011, The Chicken Team
+; Copyright (c) 2008-2012, The Chicken Team
 ; Copyright (c) 2000-2007, Felix L. Winkelmann
 ; All rights reserved.
 ;
@@ -49,17 +49,56 @@
       (apply error (string-append "[internal compiler error] " (car msg-and-args)) (cdr msg-and-args))
       (error "[internal compiler error]") ) )
 
+(define collected-debugging-output
+  (open-output-string))
+
+(define +logged-debugging-modes+ '(o x S))
+
 (define (debugging mode msg . args)
-  (and (memq mode debugging-chicken)
-       (begin
-	 (printf "~a" msg)
-	 (if (pair? args)
-	     (begin
-	       (display ": ")
-	       (for-each (lambda (x) (printf "~s " (force x))) args) ) )
-	 (newline)
-	 (flush-output)
-	 #t) ) )
+  (define (text)
+    (with-output-to-string
+      (lambda ()
+	(display msg)
+	(when (pair? args)
+	  (display ": ")
+	  (for-each
+	   (lambda (x) (printf "~s " (force x))) 
+	   args) )
+	(newline))))
+  (define (dump txt)
+    (fprintf collected-debugging-output "~a|~a" mode txt))
+  (cond ((memq mode debugging-chicken)
+	 (let ((txt (text)))
+	   (display txt)
+	   (flush-output)
+	   (when (memq mode +logged-debugging-modes+)
+	     (dump txt))
+	   #t))
+	(else
+	 (when (memq mode +logged-debugging-modes+)
+	   (dump (text)))
+	 #f)))
+
+(define (with-debugging-output mode thunk)
+  (define (collect text)
+    (for-each
+     (lambda (ln)
+       (fprintf collected-debugging-output "~a|~a~%"
+	 (if (pair? mode) (car mode) mode)
+	 ln))
+     (string-split text "\n")))
+  (define (test-mode mode set)
+    (if (symbol? mode)
+	(memq mode set)
+	(pair? (lset-intersection eq? mode set))))
+  (cond ((test-mode mode debugging-chicken)
+	 (let ((txt (with-output-to-string thunk)))
+	   (display txt)
+	   (flush-output)
+	   (when (test-mode mode +logged-debugging-modes+)
+	     (collect txt))))
+	((test-mode mode +logged-debugging-modes+)
+	 (collect (with-output-to-string thunk)))))
 
 (define (quit msg . args)
   (let ([out (current-error-port)])
@@ -113,6 +152,12 @@
 	  [(eq? x (car lst)) i]
 	  [else (loop (cdr lst) (add1 i))] ) ) )
 
+(define (posv x lst)
+  (let loop ([lst lst] [i 0])
+    (cond [(null? lst) #f]
+	  [(eqv? x (car lst)) i]
+	  [else (loop (cdr lst) (add1 i))] ) ) )
+
 (define (stringify x)
   (cond ((string? x) x)
 	((symbol? x) (symbol->string x))
@@ -123,6 +168,10 @@
 	((string? x) (string->symbol x))
 	(else (string->symbol (sprintf "~a" x))) ) )
 
+(define (slashify s) (string-translate (->string s) "\\" "/"))
+
+(define (uncommentify s) (string-translate* (->string s) '(("*/" . "*_/"))))
+  
 (define (build-lambda-list vars argc rest)
   (let loop ((vars vars) (n argc))
     (cond ((or (zero? n) (null? vars)) (or rest '()))
@@ -139,7 +188,7 @@
 	  '(#\")
 	  (let* ((c (car chars))
 		 (code (char->integer c)) )
-	    (if (or (< code 32) (>= code 127) (memq c '(#\" #\' #\\ #\?)))
+	    (if (or (< code 32) (>= code 127) (memq c '(#\" #\' #\\ #\? #\*)))
 		(append '(#\\)
 			(cond ((< code 8) '(#\0 #\0))
 			      ((< code 64) '(#\0))
@@ -396,6 +445,7 @@
 		 (inline-target . ilt) (inline-transient . itr)
 		 (undefined . und) (replacing . rpg) (unused . uud) (extended-binding . xtb)
 		 (inline-export . ilx) (hidden-refs . hrf)
+		 (value-ref . vvf)
 		 (customizable . cst) (has-unused-parameters . hup) (boxed-rest . bxr) ) ) 
 	(omit #f))
     (lambda (db)
@@ -462,7 +512,7 @@
   (subexpressions node-subexpressions node-subexpressions-set!)) ; (node...)
 
 (define-record-printer (node n out)
-  (fprintf out "#<node ~a>" (node-class n)))
+  (fprintf out "#<node ~a ~a>" (node-class n) (node-parameters n)))
 
 (define (make-node c p s)
   (##sys#make-structure 'node c p s) ) ; this kludge is for allowing the inlined `make-node'
@@ -471,13 +521,13 @@
 (define (qnode const) (make-node 'quote (list const) '()))
 
 (define (build-node-graph exp)
-  (let ([count 0])
+  (let ((count 0))
     (define (walk x)
       (cond ((symbol? x) (varnode x))
+	    ((node? x) x)
 	    ((not-pair? x) (bomb "bad expression" x))
 	    ((symbol? (car x))
 	     (case (car x)
-	       ((##core#global-ref) (make-node '##core#global-ref (list (cadr x)) '()))
 	       ((if ##core#undefined) (make-node (car x) '() (map walk (cdr x))))
 	       ((quote)
 		(let ((c (cadr x)))
@@ -494,13 +544,39 @@
 		      [body (caddr x)] )
 		  (if (null? bs)
 		      (walk body)
-		      (make-node 'let (unzip1 bs)
-				 (append (map (lambda (b) (walk (cadr b))) (cadr x))
-					 (list (walk body)) ) ) ) ) )
+		      (make-node
+		       'let 
+		       (map (lambda (v) 
+			      ;; for temporaries introduced by specialization
+			      (if (eq? '#:tmp v) (gensym) v)) ; OBSOLETE
+			    (unzip1 bs))
+		       (append (map (lambda (b) (walk (cadr b))) (cadr x))
+			       (list (walk body)) ) ) ) ) )
 	       ((lambda ##core#lambda) 
 		(make-node 'lambda (list (cadr x)) (list (walk (caddr x)))))
+	       ((##core#the)
+		(make-node '##core#the
+			   (list (second x) (third x))
+			   (list (walk (fourth x)))))
+	       ((##core#typecase)
+		;; clause-head is already stripped
+		(let loop ((cls (cdddr x)) (types '()) (exps (list (walk (caddr x)))))
+		  (cond ((null? cls) 	; no "else" clause given
+			 (make-node
+			  '##core#typecase 
+			  (cons (cadr x) (reverse types))
+			  (reverse
+			   (cons (make-node '##core#undefined '() '()) exps))))
+			((eq? 'else (caar cls))
+			 (make-node
+			  '##core#typecase
+			  (cons (cadr x) (reverse (cons '* types)))
+			  (reverse (cons (walk (cadar cls)) exps))))
+			(else (loop (cdr cls)
+				    (cons (caar cls) types)
+				    (cons (walk (cadar cls)) exps))))))
 	       ((##core#primitive)
-		(let ([arg (cadr x)])
+		(let ((arg (cadr x)))
 		  (make-node
 		   (car x)
 		   (list (if (and (pair? arg) (eq? 'quote (car arg))) (cadr arg) arg))
@@ -523,7 +599,7 @@
 					##core#inline_loc_ref ##core#inline_loc_update)
 		(make-node (first x) (second x) (map walk (cddr x))) )
 	       ((##core#app)
-		(make-node '##core#call '(#t) (map walk (cdr x))) )
+		(make-node '##core#call (list #t) (map walk (cdr x))) )
 	       (else
 		(receive (name ln) (get-line-2 x)
 		  (make-node
@@ -538,10 +614,10 @@
 				     (or rn (##sys#symbol->qualified-string name))) )
 			     (##sys#symbol->qualified-string name) ) )
 		   (map walk x) ) ) ) ) )
-	    (else (make-node '##core#call '(#f) (map walk x))) ) )
+	    (else (make-node '##core#call (list #f) (map walk x))) ) )
     (let ([exp2 (walk exp)])
       (when (positive? count)
-	(debugging 'o "eliminated procedure checks" count))
+	(debugging 'o "eliminated procedure checks" count)) ;XXX perhaps throw this out
       exp2) ) )
 
 (define (build-expression-tree node)
@@ -553,8 +629,12 @@
 	((if ##core#box ##core#cond) (cons class (map walk subs)))
 	((##core#closure)
 	 `(##core#closure ,params ,@(map walk subs)) )
-	((##core#variable ##core#global-ref) (car params))
-	((quote) `(quote ,(car params)))
+	((##core#variable) (car params))
+	((quote)
+	 (let ((c (car params)))
+	   (if (or (boolean? c) (string? c) (number? c) (char? c))
+	       c
+	       `(quote ,(car params)))))
 	((let)
 	 `(let ,(map list params (map walk (butlast subs)))
 	    ,(walk (last subs)) ) )
@@ -564,7 +644,20 @@
 		   '##core#lambda)
 	       (third params)
 	       (walk (car subs)) ) )
-	((##core#call) (map walk subs))
+	((##core#the)
+	 `(the ,(first params) ,(walk (first subs))))
+	((##core#typecase)
+	 `(compiler-typecase
+	   ,(walk (first subs))
+	   ,@(let loop ((types (cdr params)) (bodies (cdr subs)))
+	       (if (null? types)
+		   (if (null? bodies)
+		       '()
+		       `((else ,(walk (car bodies)))))
+		   (cons (list (car types) (walk (car bodies)))
+			 (loop (cdr types) (cdr bodies)))))))
+	((##core#call) 
+	 (map walk subs))
 	((##core#callunit) (cons* '##core#callunit (car params) (map walk subs)))
 	((##core#undefined) (list class))
 	((##core#bind) 
@@ -636,6 +729,7 @@
 		  (val1 (walk (first subs) rl))
 		  (a (gensym v))
 		  (rl2 (alist-cons v a rl)) )
+	     (put! db a 'inline-transient #t)
 	     (make-node 
 	      'let (list a)
 	      (list val1 (walk (second subs) rl2)))) ]
@@ -775,7 +869,7 @@
   (let walk ([n node])
     (let ([subs (node-subexpressions n)])
       (case (node-class n)
-	[(##core#variable quote ##core#undefined ##core#proc ##core#global-ref) #f]
+	[(##core#variable quote ##core#undefined ##core#proc) #f]
 	[(##core#lambda) 
 	 (let ([id (first (node-parameters n))])
 	   (find (lambda (fs)
@@ -1152,6 +1246,66 @@
 	   body)])))
 
 
+;;; Translate foreign-type into scrutinizer type:
+
+(define (foreign-type->scrutiny-type t mode) ; MODE = 'arg | 'result
+  (let ((ft (final-foreign-type t)))
+    (case ft
+      ((void) 'undefined)
+      ((char unsigned-char) 'char)
+      ((int unsigned-int short unsigned-short byte unsigned-byte int32 unsigned-int32)
+       'fixnum)
+      ((float double)
+       (case mode
+	 ((arg) 'number)
+	 (else 'float)))
+      ((scheme-pointer nonnull-scheme-pointer) '*)
+      ((blob) 
+       (case mode
+	 ((arg) '(or boolean blob))
+	 (else 'blob)))
+      ((nonnull-blob) 'blob)
+      ((pointer-vector) 
+       (case mode
+	 ((arg) '(or boolean pointer-vector))
+	 (else 'pointer-vector)))
+      ((nonnull-pointer-vector) 'pointer-vector)
+      ((u8vector u16vector s8vector s16vector u32vector s32vector f32vector f64vector)
+       (case mode
+	 ((arg) `(or boolean (struct ,ft)))
+	 (else `(struct ,ft))))
+      ((nonnull-u8vector) '(struct u8vector))
+      ((nonnull-s8vector) '(struct s8vector))
+      ((nonnull-u16vector) '(struct u16vector))
+      ((nonnull-s16vector) '(struct s16vector))
+      ((nonnull-u32vector) '(struct u32vector))
+      ((nonnull-s32vector) '(struct s32vector))
+      ((nonnull-f32vector) '(struct f32vector))
+      ((nonnull-f64vector) '(struct f64vector))
+      ((integer long size_t integer32 unsigned-integer32 integer64 unsigned-integer64
+		unsigned-long) 
+       'number)
+      ((c-pointer)
+       '(or boolean pointer locative))
+      ((nonnull-c-pointer) 'pointer)
+      ((c-string c-string* unsigned-c-string unsigned-c-string*)
+       '(or boolean string))
+      ((c-string-list c-string-list*)
+       '(list-of string))
+      ((nonnull-c-string nonnull-c-string* nonnull-unsigned-c-string*) 'string)
+      ((symbol) 'symbol)
+      (else
+       (cond ((pair? t)
+	      (case (car t)
+		((ref pointer function c-pointer)
+		 '(or boolean pointer locative))
+		((const) (foreign-type->scrutiny-type (cadr t) mode))
+		((enum) 'number)
+		((nonnull-pointer nonnull-c-pointer) 'pointer)
+		(else '*)))
+	     (else '*))))))
+
+
 ;;; Scan expression-node for variable usage:
 
 (define (scan-used-variables node vars)
@@ -1252,6 +1406,9 @@
 (define (set-real-name! name rname)
   (##sys#hash-table-set! real-name-table name rname) )
 
+;; Arbitrary limit to prevent runoff into exponential behavior
+(define real-name-max-depth 20)
+
 (define (real-name var . db)
   (define (resolve n)
     (let ([n2 (##sys#hash-table-ref real-name-table n)])
@@ -1263,15 +1420,20 @@
     (cond [(not rn) (##sys#symbol->qualified-string var)]
 	  [(pair? db)
 	   (let ([db (car db)])
-	     (let loop ([prev (##sys#symbol->qualified-string rn)] 
+	     (let loop ([nesting (list (##sys#symbol->qualified-string rn))]
+                        [depth 0]
 			[container (get db var 'contained-in)] )
-	       (if container
-		   (let ([rc (resolve container)])
-		     (if (eq? rc container)
-			 prev
-			 (loop (sprintf "~A in ~A" prev rc)
-			       (get db container 'contained-in) ) ) )
-		   prev) ) ) ]
+	       (cond
+                ((> depth real-name-max-depth)
+                 (string-intersperse (reverse (cons "..." nesting)) " in "))
+                (container
+                 (let ([rc (resolve container)])
+                   (if (eq? rc container)
+                       (string-intersperse (reverse nesting) " in ")
+                       (loop (cons (symbol->string rc) nesting)
+                             (fx+ depth 1)
+                             (get db container 'contained-in) ) ) ))
+                (else (string-intersperse (reverse nesting) " in "))) ) ) ]
 	  [else (##sys#symbol->qualified-string rn)] ) ) )
 
 (define (real-name2 var db)
@@ -1296,25 +1458,32 @@
       (car info)
       (and info (->string info))))
 
+(define (call-info params var)
+  (or (and-let* ((info (and (pair? (cdr params)) (second params))))
+	(and (list? info)
+	     (let ((ln (car info))
+		   (name (cadr info)))
+	       (conc "(" ln ") " var))))
+      var))
+
 
 ;;; constant folding support:
 
-(define (string-null? x) 
-  (##core#inline "C_i_string_null_p" s))
-
 (define (constant-form-eval op argnodes k)
   (let* ((args (map (lambda (n) (first (node-parameters n))) argnodes))
 	 (form (cons op (map (lambda (arg) `(quote ,arg)) args))))
     (handle-exceptions ex 
 	(begin
-	  (debugging 'o "folding constant expression failed" form)
-	  (k #f form #f))
+	  (k #f form #f (get-condition-property ex 'exn 'message)))
       ;; op must have toplevel binding, result must be single-valued
       (let ((proc (##sys#slot op 0)))
 	(if (procedure? proc)
-	    (let ((result (apply proc args)))
-	      (debugging 'o "folded constant expression" form)
-	      (k #t form result))
+	    (let ((results (receive (apply proc args))))
+	      (cond ((= 1 (length results))
+		     (debugging 'o "folded constant expression" form)
+		     (k #t form (car results) #f))
+		    (else 
+		     (bomb "attempt to constant-fold call to procedure that has multiple results" form))))
 	    (bomb "attempt to constant-fold call to non-procedure" form))))))
 
 
@@ -1428,8 +1597,7 @@
 (define (load-identifier-database name)
   (and-let* ((rp (repository-path))
 	     (dbfile (file-exists? (make-pathname rp name))))
-    (when verbose-mode
-      (printf "loading identifier database ~a ...~%" dbfile))
+    (debugging 'p (sprintf "loading identifier database ~a ...~%" dbfile))
     (for-each
      (lambda (e)
        (let ((id (car e)))
@@ -1510,6 +1678,7 @@ Usage: chicken FILENAME OPTION ...
     -no-lambda-info              omit additional procedure-information
     -scrutinize                  perform local flow analysis for static checks
     -types FILENAME              load additional type database
+    -emit-type-file FILENAME     write type-declaration information into file
 
   Optimization options:
 
@@ -1526,7 +1695,7 @@ Usage: chicken FILENAME OPTION ...
     -inline                      enable inlining
     -inline-limit LIMIT          set inlining threshold
     -inline-global               enable cross-module inlining
-    -unboxing                    use unboxed temporaries if possible
+    -specialize                  perform type-based specialization of primitive calls
     -emit-inline-file FILENAME   generate file with globally inlinable
                                   procedures (implies -inline -local)
     -consult-inline-file FILENAME  explicitly load inline file
@@ -1534,20 +1703,20 @@ Usage: chicken FILENAME OPTION ...
     -no-bound-checks             disable bound variable checks
     -no-procedure-checks         disable procedure call checks
     -no-procedure-checks-for-usual-bindings
-                                 disable procedure call checks only for usual
-                                  bindings
+                                   disable procedure call checks only for usual
+                                   bindings
     -no-procedure-checks-for-toplevel-bindings
                                    disable procedure call checks for toplevel
-                                    bindings
+                                   bindings
+    -strict-types                assume variable do not change their type
+    -clustering                  combine groups of local procedures into dispatch
+                                   loop
 
   Configuration options:
 
     -unit NAME                   compile file as a library unit
     -uses NAME                   declare library unit as used.
     -heap-size NUMBER            specifies heap-size of compiled executable
-    -heap-initial-size NUMBER    specifies heap-size at startup time
-    -heap-growth PERCENTAGE      specifies growth-rate of expanding heap
-    -heap-shrinkage PERCENTAGE   specifies shrink-rate of contracting heap
     -nursery NUMBER  -stack-size NUMBER
                                  specifies nursery size of compiled executable
     -extend FILENAME             load file before compilation commences
@@ -1570,3 +1739,45 @@ Usage: chicken FILENAME OPTION ...
 
 EOF
 ) )
+
+(define (print-debug-options)
+  (display #<<EOF
+
+Available debugging options:
+
+     a          show node-matching during simplification
+     b          show breakdown of time needed for each compiler pass
+     c          print every expression before macro-expansion
+     d          lists all assigned global variables
+     e          show information about specializations
+     h          you already figured that out
+     i          show information about inlining
+     m          show GC statistics during compilation
+     n          print the line-number database 
+     o          show performed optimizations
+     p          display information about what the compiler is currently doing
+     r          show invocation parameters
+     s          show program-size information and other statistics
+     t          show time needed for compilation
+     u          lists all unassigned global variable references
+     x          display information about experimental features
+     D          when printing nodes, use node-tree output
+     I          show inferred type information for unexported globals
+     M          show syntax-/runtime-requirements
+     N          show the real-name mapping table
+     P          show expressions after specialization
+     S          show applications of compiler syntax
+     T          show expressions after converting to node tree
+     1          show source expressions
+     2          show canonicalized expressions
+     3          show expressions converted into CPS
+     4          show database after each analysis pass
+     5          show expressions after each optimization pass
+     6          show expressions after each inlining pass
+     7          show expressions after complete optimization
+     8          show database after final analysis
+     9          show expressions after closure conversion
+
+
+EOF
+))
