@@ -40,9 +40,11 @@
 ;; some basic contrived testing
 
 (define (fac n)
-  (let-syntax ((m1 (lambda (n r c) 
-		     (pp `(M1: ,n))
-		     (list (r 'sub1) (cadr n)))))
+  (let-syntax ((m1 
+		(er-macro-transformer
+		 (lambda (n r c) 
+		   (pp `(M1: ,n))
+		   (list (r 'sub1) (cadr n))))))
     (define (sub1 . _)			; ref. transp.? (should not be used here)
       (error "argh.") )
     #;(print "fac: " n)		  
@@ -196,6 +198,30 @@
       (bar foo))))
 )
 
+;;; strip-syntax cuts across multiple levels of syntax
+;;; reported by Matthew Flatt
+(define-syntax c 
+  (syntax-rules () 
+    [(_) 
+     (let ([x 10]) 
+       (let-syntax ([z (syntax-rules ()
+                         [(_) (quote x)])])
+         (z)))]))
+
+(t "x" (symbol->string (c)))
+
+(define-syntax c2
+   (syntax-rules ()
+     [(_)
+      (let ([x 10])
+        (let-syntax ([z (syntax-rules ()
+                          [(_) (let-syntax ([w (syntax-rules ()
+                                                 [(_) (quote x)])])
+                                 (w))])])
+          (z)))]))
+
+(t "x" (symbol->string (c2)))
+
 ;;; strip-syntax on renamed module identifiers, as well as core identifiers
 (module foo (bar)
   (import chicken scheme)
@@ -341,11 +367,12 @@
 
 
 (define-syntax loop
-  (lambda (x r c)
-    (let ((body (cdr x)))
-      `(,(r 'call/cc)
-	(,(r 'lambda) (exit)
-	 (,(r 'let) ,(r 'f) () ,@body (,(r 'f))))))))
+  (er-macro-transformer
+   (lambda (x r c)
+     (let ((body (cdr x)))
+       `(,(r 'call/cc)
+	 (,(r 'lambda) (exit)
+	  (,(r 'let) ,(r 'f) () ,@body (,(r 'f)))))))))
 
 (let ((n 10))
   (loop
@@ -363,10 +390,11 @@
 (f (while0 #f (print "no.")))
 
 (define-syntax while
-  (lambda (x r c)
-    `(,(r 'loop) 
-      (,(r 'if) (,(r 'not) ,(cadr x)) (exit #f))
-      ,@(cddr x))))
+  (er-macro-transformer
+   (lambda (x r c)
+     `(,(r 'loop) 
+       (,(r 'if) (,(r 'not) ,(cadr x)) (exit #f))
+       ,@(cddr x)))))
 
 (let ((n 10))
   (while (not (zero? n))
@@ -385,8 +413,9 @@
   (syntax-rules ()
     ((_ (name . llist) body ...)
      (define-syntax name
-       (lambda (x r c)
-	 (apply (lambda llist body ...) (strip-syntax (cdr x))))))))
+       (er-macro-transformer
+	(lambda (x r c)
+	  (apply (lambda llist body ...) (strip-syntax (cdr x)))))))))
 
 (define-macro (loop . body)
   (let ((loop (gensym)))
@@ -525,8 +554,9 @@
   (define-syntax s1 (syntax-rules () ((_ x) (list x))))
 
   (define-syntax s2
-    (lambda (x r c)
-      (r `(vector (s1 ,(cadr x)))))) )	; without renaming the local version of `s1'
+    (er-macro-transformer
+     (lambda (x r c)
+       (r `(vector (s1 ,(cadr x))))))) )	; without renaming the local version of `s1'
 					; below will be captured 
 
 (import m1)
@@ -734,6 +764,57 @@
 
 (assert (string=? "hello, XXX" (bar who: "XXX")))
 
+;;; DSSSL keyword arguments in various combinations with optional and rest args
+;;; reported on IRC by R. Winkler
+
+(define (test-optional&rest x y #!optional z #!rest r)
+  (list x y z r))
+
+(assert (equal? '(3 4 5 (6 7)) (test-optional&rest 3 4 5 6 7)))
+
+(define (test-optional&key x y #!optional z #!key i (j 1))
+  (list x y z i: i j: j))
+
+(assert (equal? '(3 4 5 i: 6 j: 7) (test-optional&key 3 4 5 i: 6 j: 7 8)))
+;; Unfortunate but correct (missing optional arg)
+(assert (equal? '(3 4 i: i: #f j: 1) (test-optional&key 3 4 i: 6 j: 7 8)))
+
+(define (test-key&rest x y #!rest r #!key i (j 1))
+  (list x y i: i j: j r))
+
+(assert (equal? '(3 4 i: 5 j: 1 (i: 5 6 7)) (test-key&rest 3 4 i: 5 6 7)))
+(assert (equal? '(3 4 i: 5 j: 6 (i: 5 j: 6 7 8))
+                (test-key&rest 3 4 i: 5 j: 6 7 8)))
+
+(define (test-optional-key&rest x y #!optional z #!rest r #!key i (j 1))
+  (list x y z i: i j: j r))
+
+(assert (equal? '(3 4 5 i: 6 j: 7 (i: 6 j: 7 8))
+                (test-optional-key&rest 3 4 5 i: 6 j: 7 8)))
+
+;;; Miscellaneous DSSSL tests
+
+;; DSSSL annotations may each appear only once
+(f (eval '(lambda (x #!optional o1 #!optional o2) 'foo)))
+(f (eval '(lambda (x #!rest r1 #!rest r1) 'foo)))
+(f (eval '(lambda (x #!key k1 #!key k2) 'foo)))
+
+;; DSSSL annotations must occur in order (optional, rest, key)
+(f (eval '(lambda (x #!rest r1 #!optional o1) 'foo)))
+(f (eval '(lambda (x #!key k1 #!optional o1) 'foo)))
+(f (eval '(lambda (x #!key r1 #!rest k1) 'foo)))
+
+;; #!rest limited to 1 arg
+(f (eval '(lambda (x #!rest r1 r2) 'foo)))
+
+;; list arguments invalid for required and rest parameters
+(f (eval '(lambda ((x 0) #!rest r1) 'foo)))
+(f (eval '(lambda (x #!rest (r1 0)) 'foo)))
+
+;; Dotted list syntax can be mixed in
+(t '(1 2 3 4 (5 6))
+   ((lambda (x y #!optional o1 o2 . r) (list x y o1 o2 r))
+    1 2 3 4 5 6))
 
 ;;; import not seen, if explicitly exported and renamed:
 
@@ -914,3 +995,63 @@
   (set-cdr! a a)
   `(1 ,@a))
 
+
+;; ##sys#alias-global-hook, when invoked via eval/meta, did resolve identifiers
+;; used during evaluation of an expander body in the wrong environment and mapped
+;; an identifier to something imported for the runtime environment
+
+(module foonumbers (+)
+  (import (except scheme +) chicken)
+  (define (+ . _) (error "failed.")))
+
+(import foonumbers)
+
+(define-syntax foo
+  (er-macro-transformer
+   (lambda (x r c)
+     `(print ,(+ (cadr x) 1)))))
+
+(foo 3)
+
+
+;; #578: "use" with import-specifier has no effect for internal modules on csi's top-level
+
+(use (prefix srfi-1 list-))
+take
+
+
+;; #805: case-lambda is unhygienic (as well as ensure, see 4706afb4 and bc5cc698)
+(module case-lambda-and-ensure-hygiene ()
+  (import (prefix chicken c/) (prefix scheme s/))
+  (c/case-lambda ((a) a))
+  (c/ensure s/even? 2))
+
+
+;; #816: compiler-syntax should obey hygiene in its rewrites
+(module foo ()
+  (import (prefix (only scheme map lambda list) ~))
+  (~map (~lambda (y) y) (~list 1)))
+
+;; #852: renamed macros should not be returned as first-class
+;;       objects in the interpreter
+(module renamed-macros (renamed-macro-not-firstclassed)
+  (import chicken scheme)
+  (define-syntax renamed-macro-not-firstclassed
+    (er-macro-transformer
+     (lambda (e r c)
+       `(,(r 'list) ,(r 'define))))))
+
+(f (eval '(begin (import renamed-macros) (renamed-macro-not-firstclassed))))
+
+;; #893: implicitly renamed variables shouldn't be resolved to core
+;;       builtins (#%xyz), but go through a level of indirection, so
+;;       strip-syntax can still access the original symbol.
+(module rename-builtins
+ (strip-syntax-on-*)
+ (import chicken scheme)
+ (define-syntax strip-syntax-on-*
+   (ir-macro-transformer
+    (lambda (e r c) '(quote *)))))
+
+(import rename-builtins)
+(assert (eq? '* (strip-syntax-on-*)))
\ No newline at end of file
